import { Sequelize, Transaction } from '@sequelize/core'

import MYSQL from 'mysql2/promise'
import MariaDB from 'mariadb'
import MSSQL from 'tedious'
import SQLITE from 'sqlite3'

import { database } from '../../config/index.js'

const instance = new Sequelize(database.name, database.username, database.password, {
    host: database.host,
    port: database.port,
    dialect: database.dialect,
    dialectOptions: {
        'mysql': {},
        'sqlite': {},
        'mariadb': {},
        'mssql': { encrypt: true }
    }[database.dialect],
    define: {
        charset: 'utf8',
        collate: 'utf8mb3_general_ci',
        timestamps: true
    },
    isolationLevel: Transaction.ISOLATION_LEVELS.REPEATABLE_READ,
    logging: !!database.logging,
    pool: {
        max: 10,
        min: 0,
        idle: 10000
    },
})

instance.addHook('beforeConnect', async (config) => {
    // Створити базу даних при першому запуску якщо такої не існує
    var connection = null
    switch (database.dialect) {
        case 'mysql':
        case 'mariadb':
            await MYSQL.createConnection({ host: database.host, port: database.port, user: database.username, password: database.password }).then(connection => {
                return new Promise(async (resolve, reject) => {
                    try {
                        await connection.query(`CREATE DATABASE IF NOT EXISTS ${database.name} CHARACTER SET utf8 COLLATE utf8_general_ci;`)
                        // Вимкнути примусово повне групування вибірки при агрегуванні
                        const result = await connection.query('SELECT @@sql_mode')
                        if (result && result[0][0]['@@sql_mode'].includes('ONLY_FULL_GROUP_BY')) {
                            await connection.query(`SET SESSION sql_mode="${result[0][0]['@@sql_mode'].replace('ONLY_FULL_GROUP_BY', '').replace(',,', ',')}"`)
                        }
                        resolve(true)
                    } catch (error) {
                        reject(error)
                    }
                }).catch(error => {
                    console.log(`[${new Date().toISOString()}] `, 'MYSQL помилка створення бази даних: ', error)
                }).finally(() => {
                    connection.destroy()
                })
            }).catch(error => {
                console.log(`[${new Date().toISOString()}] `, 'MYSQL помилка підключення до субд: ', error)
            })
            return
        case 'mssql':
            return new Promise((resolve, reject) => {
                try {
                    connection = new MSSQL.Connection({
                        server: database.host,
                        options: {
                            trustServerCertificate: true
                        },
                        authentication: {
                            type: 'default',
                            options: {
                                userName: database.username,
                                password: database.password,
                            }
                        }
                    });
                    const query = (sql) => new Promise((resolve, reject) => connection.execSql(new MSSQL.Request(sql, (error) => error ? reject(error) : resolve())))
                    connection.on('connect', async function (error) {
                        try {
                            if (error) throw error
                            await query(`IF NOT EXISTS ( SELECT name FROM sys.databases WHERE name = N'${database.name}') CREATE DATABASE [${database.name}];`)
                            await query(`IF SERVERPROPERTY('ProductVersion') > '12' ALTER DATABASE [${database.name}] SET QUERY_STORE=ON;`)
                            connection.close()
                            resolve()
                        } catch (error) {
                            console.log(`[${new Date().toISOString()}] `, 'MSSQL помилка створення бази даних: ', error)
                            connection.close()
                            reject(error)
                        }
                    });
                    connection.connect();
                } catch (error) {
                    console.log(`[${new Date().toISOString()}] `, 'MSSQL помилка створення бази даних: ', error)
                }
            })
        case 'sqlite':
            connection = new SQLITE.Database(database.host, SQLITE.OPEN_READWRITE | SQLITE.OPEN_CREATE, (error) => {
                if (error) console.log(`[${new Date().toISOString()}] `, 'SQLITE помилка створення бази даних: ', error)
            });
            connection.close();
            return
    }
    return
})


export default instance