
import { DataTypes } from "@sequelize/core"
import instance from "../instance.js"

export class Column {

    static async add(db, transaction, table, name, { type = DataTypes.STRING, allowNull = true, defaultValue = undefined, unique = undefined, references = null, throwIfExists = false }) {
        try {
            await db.addColumn(table, name, {
                type,
                allowNull,
                ...(unique !== undefined ? { unique } : {}),
                ...(defaultValue !== undefined ? { defaultValue } : {}),
                ...(references ? { references } : {}),
            }, { transaction });
        } catch (error) {
            switch (error?.parent?.code) {
                case 'ER_DUP_FIELDNAME':
                    if (throwIfExists) throw error
                    break;
                default: throw error
            }
        }
    }
    static async change(db, transaction, table, name, { type = DataTypes.STRING, allowNull = true, defaultValue = undefined, unique = undefined, references = null }) {
        await db.changeColumn(table, name, {
            type,
            allowNull,
            ...(unique !== undefined ? { unique } : {}),
            ...(defaultValue !== undefined ? { defaultValue } : {}),
            ...(references ? { references } : {})
        }, { transaction });
    }

    static async rename(db, transaction, table, name, new_name) {
        await db.renameColumn(table, name, new_name, { transaction })
    }

    static async remove(db, transaction, table, name) {
        try {
            await db.removeColumn(table, name, { transaction })
        } catch (error) {
            switch (error?.original?.code) {
                case 'ER_CANT_DROP_FIELD_OR_KEY': return
                default: throw error
            }
        }
    }
}

export class Index {

    static async add(db, transaction, table, { fields, name, type, unique = false, throwIfExists = false }) {
        try {
            await db.addIndex(table, fields, { name, type, unique, transaction, })
        } catch (error) {
            switch (error?.parent?.code) {
                case 'ER_DUP_KEYNAME':
                    if (throwIfExists) throw error
                    break;
                default: throw error
            }
        }
    }

    static async remove(db, transaction, table, name) {
        try {
            await db.removeConstraint(table, name, { transaction })
        } catch (error) {
            switch (error?.name) {
                case 'SequelizeUnknownConstraintError': return
                default: throw error
            }
        }
    }

}

export class Constraint {

    static async addPrimary(db, transaction, table_name, { fields, name = "PRIMARY" }) {
        await db.addConstraint(table_name, { fields, name, type: 'primary key', transaction });
    }

    static async addForeign(db, transaction, table_name, { fields, name, references: { table, field }, onUpdate = 'cascade', onDelete = 'restrict' }) {
        await db.addConstraint(table_name, { fields, name: name || `foreign_key_to_${table}_${field}`, references: { table, field }, type: 'foreign key', onUpdate, onDelete, transaction });
    }

    static async addUnique(db, transaction, table_name, { fields, name = "UNIQUE", references: { table, field }, onUpdate = 'cascade', onDelete = 'restrict' }) {
        await db.addConstraint(table_name, { fields, name, type: 'unique', transaction });
    }

    static async remove(db, transaction, table, name) {
        try {
            await db.removeConstraint(table, name, { transaction })
        } catch (error) {
            switch (error?.name) {
                case 'SequelizeUnknownConstraintError': return
                default: throw error
            }
        }
    }
}