import fs from 'fs'
import { Transaction, Op } from '@sequelize/core'
import Database from '../database/index.js'
import Transform from './transform.js'
import Notifications from '../lib/Notifications.js'
import config from '../../config/index.js'
import structure from './scheme.js'
import ArrivalHelper from '../helpers/ArrivalHelper.js'
import ProductPriceRegisterValuesHelper from '../database/schemes/registers/product/product_price_values_helper.js'


export default class Exchange {

    static GetPackageSize(data) {
        return data && Object.keys(data).length ? Object.keys(data).map(key => data[key] && Array.isArray(data[key]) ? data[key].length || null : 0).reduce((prev, next) => prev + next) : 0
    }

    static GetPackageMemory(data) {
        return typeof data === 'string' ? (data.length * 2 / 1024 / 1024).toFixed(1) : (JSON.stringify(payload).length * 2 / 1024 / 1024).toFixed(1)
    }


    static async PrepareExchangePayload({ receive_from }) {
        const transaction = await Database.instance.transaction({ isolationLevel: Transaction.ISOLATION_LEVELS.READ_COMMITTED })
        try {
            const query = { where: receive_from ? { updatedAt: { [Op.gte]: receive_from } } : {}, transaction }

            const Shifts = await Database.Documents.Shifts.findAll({
                include: [
                    { model: Database.Documents.ShiftsTransactions, as: 'open_transaction', where: { type: 'SHIFT_OPEN' }, attributes: { exclude: ['subdivision_id'] }, },
                    { model: Database.Documents.ShiftsTransactions, as: 'close_transaction', where: { type: 'SHIFT_CLOSE' }, attributes: { exclude: ['subdivision_id'] }, }
                ],
                ...query,
            }).then(shifts => {
                return shifts.map(shift => {
                    return {
                        UUID: shift.id,
                        Number: shift.number,
                        UserID: shift.staff_id,
                        WorkplaceID: shift.workplace_id,
                        DepartmentID: shift.subdivision_id,
                        OpenDate: shift.open_transaction.createdAt.toISOString(),
                        CloseDate: shift.close_transaction.createdAt.toISOString(),
                        UpdatedAt: shift.updatedAt
                    }
                })
            })
            //await Database.Documents.Shifts.update({ uploading_mark: 1 }, { where: { status: 'CLOSED', uploading_mark: 0 }, transaction })

            const ShiftReports = await Database.Documents.Report.findAll({
                include: [{ model: Database.Documents.ReportGoods, as: 'goods' }],
                ...query
            }).then(reports => {
                return reports.map(report => {
                    return {
                        UUID: report.id,
                        Number: report.number,
                        Presentation: report.presentation,
                        ShiftID: report.shift_id,
                        UserID: report.staff_id,
                        WarehouseID: report.warehouse_id,
                        WorkplaceID: report.workplace_id,
                        DepartmentID: report.subdivision_id,
                        Date: report.createdAt.toISOString(),
                        UpdatedAt: report.updatedAt.toISOString(),
                        Goods: report.goods.map(good => {
                            return {
                                GoodID: good.product_id,
                                SeriesID: good.series_id,
                                PackingID: good.packing_id,
                                TaxID: good.tax_id,
                                Quantity: good.amount,
                                Price: good.price,
                                UpdatedAt: good.updatedAt
                            }
                        })
                    }
                })
            })

            const GoodsReturn = await Database.Documents.Return.findAll({
                include: [{ model: Database.Documents.ReturnGoods, as: 'items' }],
                ...query
            }).then(returns => {
                return returns.map(document => {
                    return {
                        "UUID": document.id,
                        "Number": document.number.toString(),
                        "DeletionMark": false,
                        "Comment": document.name,
                        "Date": document.createdAt,
                        "WarehouseID": document.warehouse_id,
                        "PriceTypeID": document.price_type_id,
                        "CounterpartyID": document.counterparty_id,
                        "PriceIncludesVAT": false,
                        "Goods": document.goods.map(good => {
                            return {
                                "ProductID": good.product_id,
                                "SeriesID": good.series_id,
                                "PackageID": good.packing_id,
                                "Quantity": good.amount,
                                "Amount": good.amount * good.price
                            }
                        })
                    }
                })
            })
            await transaction.commit()
            return { Shifts, ShiftReports, GoodsReturn }
        } catch (error) {
            await transaction.rollback()
            throw error
        }
    }

    static async SaveExchangeResponse(data) {
        const transaction = await Database.instance.transaction({ isolationLevel: Transaction.ISOLATION_LEVELS.READ_COMMITTED })

        try {

            const {
                skip_counterparties_ids,
                skip_counterparties,
                skip_price_types_ids,
                skip_warehouses_ids,
                skip_warehouses,
                skip_categories_ids,
                skip_products_ids,
                skip_prices,
                skip_stocks
            } = config.exchange
            const {
                subdivisions,
                taxes,
                products,
                products_barcodes,
                products_kinds,
                products_characteristic,
                products_packings,
                products_categories,
                products_producers,
                products_active_substances,
                series,
                price_types,
                warehouses,
                counterparties,
                arrivals,
                prices,
                stocks
            } = data

            for (let i = 0; subdivisions && i < subdivisions.length; i++) {
                await Database.Subdivisions.upsert(subdivisions[i], { transaction })
            }
            for (let i = 0; !skip_counterparties && counterparties && i < counterparties.length; i++) {
                if (skip_counterparties_ids?.includes(counterparties[i].id)) continue
                await Database.Counterparties.upsert(counterparties[i], { transaction })
            }
            for (let i = 0; price_types && i < price_types.length; i++) {
                if (skip_price_types_ids?.includes(price_types[i].id)) continue
                await Database.PriceTypes.upsert(price_types[i], { transaction })
            }
            for (let i = 0; !skip_warehouses && warehouses && i < warehouses.length; i++) {
                if (skip_warehouses_ids?.includes(warehouses[i].id)) continue
                await Database.Warehouses.upsert(warehouses[i], { transaction })
            }
            for (let i = 0; taxes && i < taxes.length; i++) {
                await Database.Taxes.upsert(taxes[i], { transaction })
            }
            for (let i = 0; products_categories && i < products_categories.length; i++) {
                if (skip_categories_ids?.includes(products_categories[i].id)) continue
                await Database.ProductCategories.upsert(products_categories[i], { transaction })
            }
            for (let i = 0; products_kinds && i < products_kinds.length; i++) {
                await Database.ProductKinds.upsert(products_kinds[i], { transaction })
            }
            for (let i = 0; products_characteristic && i < products_characteristic.length; i++) {
                await Database.ProductCharacteristic.upsert(products_characteristic[i], { transaction })
            }
            for (let i = 0; products_producers && i < products_producers.length; i++) {
                await Database.ProductProducers.upsert(products_producers[i], { transaction })
            }
            for (let i = 0; products_active_substances && i < products_active_substances.length; i++) {
                await Database.ProductActiveSubstances.upsert(products_active_substances[i], { transaction })
            }
            for (let i = 0; products && i < products.length; i++) {
                if (skip_categories_ids?.includes(products[i].category_id)) continue
                if (skip_products_ids?.includes(products[i].id)) continue
                await Database.Products.upsert(products[i], { transaction })
            }
            // Зберігає лише ті пакування для яких є товар
            const existed_products_ids = await Database.Products.findAll({ attributes: ['id'], transaction }).then(rows => rows.map(it => it.id))
            for (let i = 0; products_packings && i < products_packings.length; i++) {
                if (!existed_products_ids.includes(products_packings[i].product_id)) {
                    console.error(`Відсутній товар для збереження пакування: \n${JSON.stringify(products_packings[i], null, 2)}`);
                    continue;
                }
                await Database.ProductPacking.upsert(products_packings[i], { transaction })
            }


            for (let i = 0; series && i < series.length; i++) {
                await Database.Series.upsert(series[i], { transaction })
            }
            for (let i = 0; products_barcodes && i < products_barcodes.length; i++) {
                await Database.Barcodes.upsert(products_barcodes[i], { transaction })
            }

            const warehouses_local = await Database.Warehouses.findAll({ transaction })
            for (let i = 0; arrivals && i < arrivals.length; i++) {
                const warehouse = warehouses_local.find(it => it.id === arrivals[i].warehouse_id)
                if (!warehouse) throw Error(`Не вдалось знайти склад для прихідного ордера ${arrivals[i].id}`)
                arrivals[i]['subdivision_id'] = warehouse.subdivision_id
                const items = arrivals[i].goods.map(good => {
                    good['subdivision_id'] = warehouse.subdivision_id
                    return good
                })
                delete arrivals[i].goods
                await Database.Documents.Arrival.upsert(arrivals[i], { transaction })
                await Database.Documents.ArrivalGoods.update({ deletion_mark: true }, { where: { document_id: arrivals[i].id }, transaction })
                for (let j = 0; items && j < items.length; j++) {
                    await Database.Documents.ArrivalGoods.upsert(items[j], { transaction })
                }
                await ArrivalHelper.ApplyProductsStockMoves({ basis_id: arrivals[i].id, warehouse_id: arrivals[i].warehouse_id, period: arrivals[i].createdAt, items }, transaction)
                await ArrivalHelper.ApplyProductsPriceMoves({ basis_id: arrivals[i].id, price_type_id: arrivals[i].price_type_id, period: arrivals[i].createdAt, items }, transaction)
            }


            for (let i = 0; !skip_stocks && stocks && i < stocks.length; i++) {
                await Database.Registers.ProductStock.upsert(stocks[i], { transaction })
            }

            const changed = []
            const existed_price_type_ids = await Database.PriceTypes.findAll({ attributes: ['id'], transaction }).then(rows => rows.map(it => it.id))
            for (let i = 0; !skip_prices && prices && i < prices.length; i++) {
                const { price_type_id, product_id, packing_id, series_id } = prices[i]
                if (!existed_price_type_ids.includes(price_type_id)) continue
                if (!existed_products_ids.includes(product_id)) continue
                if (!changed.find(it => it.product_id == product_id && it.series_id == series_id && it.packing_id == packing_id && it.price_type_id == price_type_id)) {
                    changed.push({ product_id, series_id, packing_id, price_type_id })
                }
                await Database.Registers.ProductPrice.upsert(prices[i], { transaction })
            }
            if (changed?.length) {
                await Promise.all(changed.map(({ product_id, series_id, packing_id, price_type_id }) => (
                    ProductPriceRegisterValuesHelper.recalculate({ where: { product_id, series_id, packing_id, price_type_id }, transaction })
                )))
            }

            await transaction.commit()
        } catch (error) {
            await transaction.rollback()
            throw error
        }
    }

    static async send(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { message_number, receive_from } = Transform.parse({ data: req.query, structure })
            if (!receive_from) throw Error(`Відсутній GET параметер "${structure.receive_from}" призначений для визначення часового діапазону вибірки даних`)
            console.log(`[${new Date().toISOString()}] `, `Отримано запит на обмін: вивантаження даних від ${receive_from}`)
            const payload = await Exchange.PrepareExchangePayload({ receive_from })
            const body = JSON.stringify(Transform.prepare(payload))
            console.log(`[${new Date().toISOString()}] `, `Пакет для надсилання (${Exchange.GetPackageSize(payload)} рядків) (${Exchange.GetPackageMemory(body)} мб)`)
            res.status(200).send(body)
            console.log(`[${new Date().toISOString()}] `, `Обмін ${message_number} виконано успішно`)
        } catch (error) {
            console.log(`[${new Date().toISOString()}] `, `Помилка проведення обміну : `, error.message)
            const error_id = await Database.onCatchError(error, { name: 'Помилка проведення обміну', subdivision_id: config.subdivision.id })
            await Notifications.create({ type: 'exchange_error', text: `Помилка обміну при надсиланні даних`, error_id })
            res.status(500).send(JSON.stringify(Transform.prepare({}, { status: false, message: error.message })))
        }
    }

    static async receive(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const data = Transform.parse({ data: req.body, structure })
            const json = JSON.stringify(req.body)
            if (!fs.existsSync(`${config.app.path}/uploads/exchanges`)) fs.mkdirSync(`${config.app.path}/uploads/exchanges`, { recursive: true })
            fs.writeFile(`${config.app.path}/uploads/exchanges/${data.message_number}.json`, json, () => { })
            if (!data.message_number) throw Error(`Відсутнє поле "${structure.message_number}" в тілі запиту призначене для ідентифікації пакету даних`)
            console.log(`[${new Date().toISOString()}] `, `Отримано запит на обмін: завантаження пакету №${data.message_number}`)
            console.log(`[${new Date().toISOString()}] `, `Пакет для збереження (${Exchange.GetPackageSize(data)} рядків) (${Exchange.GetPackageMemory(json)} мб)`)
            await Exchange.SaveExchangeResponse(data)
            res.status(200).send(Transform.prepare({}))
            console.log(`[${new Date().toISOString()}] `, `Обмін ${data.message_number} виконано успішно`, '\n')
        } catch (error) {
            console.log(`[${new Date().toISOString()}] `, `Помилка проведення обміну : `, error.message)
            const error_id = await Database.onCatchError(error, { name: 'Помилка обміну при отриманні даних', subdivision_id: config.subdivision.id })
            await Notifications.create({ type: 'exchange_error', text: `Помилка обміну при отриманні даних`, error_id })
            res.status(500).send(JSON.stringify(Transform.prepare({}, { status: false, message: error.message })))
        }
    }
}