var fetch from 'node-fetch');

const delay = (timeout) => new Promise((resolve) => setTimeout(resolve, timeout))


export default class Scarb {

    constructor({ url, license, login, password }) {
        this.url = url
        this.license = license
        this.login = login
        this.password = password
        this.token = ''
    }


    request(path, data, timeout = 30000) {
        return new Promise(async (resolve, reject) => {
            try {
                setTimeout(() => { reject({ message: "Timeout (" + timeout + "ms)", timeout: true }) }, timeout)
                const response = await fetch(this.url + path, {
                    method: 'POST',
                    body: typeof data === 'string' ? data : JSON.stringify(data),
                    headers: {
                        'API-Key': this.license,
                        'Authorization': `Bearer ${this.token}`
                    }
                })
                const body = await response.json()
                if (response.status === 200) {
                    resolve(body)
                } else {
                    reject(body.message)
                }
            } catch (error) {
                reject(error.message)
            }
        })
    }


    async signin(email, password) {
        const result = await this.request('/account/login', { email: email ? email : this.email, password: password ? password : this.password })
        this.token = result.token
    }

    async prescription(id) {
        const { data } = await this.request('/medication/get-data', { id })
        const prescription = data.medicationRequest
        const address = prescription.division.addresses.length > 0 ? prescription.division.addresses[0] : null
        return {
            "id": prescription.id,
            "category": prescription.category,
            "created_at": prescription.created_at,
            "dispense_valid_from": prescription.dispense_valid_from,
            "dispense_valid_to": prescription.dispense_valid_to,
            "is_blocked": prescription.is_blocked,
            "started_at": prescription.started_at,
            "ended_at": prescription.ended_at,
            "status": prescription.status,
            "division": {
                "address": address ? `${address.area} ${address.street} ${address.building} ${address.zip || ''}` : '',
                "email": prescription.division.email,
                "name": prescription.division.name,
                "phones": prescription.division.phones.map(it => it.number),
                "type": prescription.division.type,
                "working_hours": prescription.division.working_hours
            },
            "dosage_instruction": prescription.dosage_instruction,
            "employee": prescription.employee,

            "intent": prescription.intent,
            "legal_entity": prescription.legal_entity,
            "medical_program": {
                "id": prescription.medical_program.id,
                "name": prescription.medical_program.name,
            },
            "medication_info": {
                "dosage": prescription.medication_info.dosage,
                "form": prescription.medication_info.form,
                "medication_id": prescription.medication_info.medication_id,
                "medication_name": prescription.medication_info.medication_name,
                "medication_qty": prescription.medication_info.medication_qty,
            },
            "medication_remaining_qty": prescription.medication_remaining_qty,
            "person": {
                "id": prescription.person.id,
                "name": prescription.person.short_name,
                "age": prescription.person.age,
            },
            "request_number": prescription.request_number,
            "participants": prescription.participants.map(participant => {
                return ({
                    "id": participant.id,
                    "medication_id": participant.medication_id,
                    "medication_name": participant.medication_name,
                    "manufacturer": {
                        "country": participant.manufacturer.country,
                        "name": participant.manufacturer.name
                    },
                    "registry_number": participant.registry_number,
                    "consumer_price": participant.consumer_price,
                    "wholesale_price": participant.wholesale_price,
                    "estimated_payment_amount": participant.estimated_payment_amount,
                    "reimbursement_amount": participant.reimbursement_amount,
                    "reimbursement_daily_dosage": participant.reimbursement_daily_dosage,
                    "start_date": participant.start_date,
                    "end_date": participant.end_date,
                    "form": participant.form,
                    "container_dosage": {
                        "denumerator_unit": participant.container_dosage.denumerator_unit,
                        "denumerator_value": participant.container_dosage.denumerator_value,
                        "numerator_unit": participant.container_dosage.numerator_unit,
                        "numerator_value": participant.container_dosage.numerator_value
                    },
                    "package_qty": participant.package_qty,
                    "package_min_qty": participant.package_min_qty,
                })
            })
        }
    }

    async init(payload) {
        const { data } = await this.request('/medication/get-data', payload)
        return data
    }

    async confirm(id, signedJson) {
        const { data } = await this.request('/medication/process-dispense', { id, signedJson })
        return data
    }

    async cancel(id) {
        const { data } = await this.request('/medication/reject-dispense', { id })
        return data
    }


    async get(id) {
        const { data } = await this.request('/medication/get-dispense-data', { id })
        return data
    }




}