import QrImage from 'qr-image'

export default class ReceiptGenerator {

    static Align = class {
        static Left = "LEFT"
        static Center = "CENTER"
        static Right = "RIGHT"
    }

    constructor({ width = "36", font_size = "13", font_weight = "700" }) {
        this.width = parseInt(width)
        this.font_size = parseInt(font_size)
        this.font_weight = parseInt(font_weight)
        this.content = ""
        this.html = ""
    }

    /**
    * Додає стрічку розділювач у чек із вказаною товщиною
    * @example <caption>Приклад виклику</caption>
    * receipt.separator()       // лінія із символів '-'
    * receipt.separator(true)   // лінія із символів '='
    */
    separator(bold = false) {
        const line = "".padEnd(this.width, bold ? "=" : "-")
        this.content += line + "\n"
        this.html += `<span>${line}</span>\n`
    }

    /**
    * Додає стрічку тексту до чеку із вирівнюванням
    * @example <caption>Приклад виклику</caption>
    * receipt.text('Дата: 12.05.2022')                      // вирівнювання по лівій стороні чеку
    * receipt.text('Час: 12:33', ReceiptGenerator.Align.Right)       // вирівнювання по правій стороні чеку
    * receipt.text('ПН: 1234567890', ReceiptGenerator.Align.Center)  // вирівнювання по лівій стороні чеку
    * 
    *      
    * |                ...                  |
    * |Дата: 12.05.2022                     | <-
    * |                           Час: 12:33| <-
    * |             ПН: 1234567890          | <-
    * |                ...                  |
    */
    text(text, align = ReceiptGenerator.Align.Left) {
        var line = ''
        if (text.length <= this.width) {
            switch (align) {
                case ReceiptGenerator.Align.Right:
                    line = text.padStart(this.width, " ")
                    this.content += line + "\n";
                    this.html += `<span>${line}</span>\n`
                    break;
                case ReceiptGenerator.Align.Center:
                    let content_length = text.length
                    let padding_start = Math.ceil((this.width - content_length) / 2)
                    let padding_end = this.width - content_length - padding_start
                    line = text.padStart(padding_start + content_length, " ").padEnd(padding_start + content_length + padding_end, " ")
                    this.content += line + "\n";
                    this.html += `<span>${line}</span>\n`
                    break;
                case ReceiptGenerator.Align.Left:
                    line = text.padEnd(this.width, " ")
                    this.content += line + "\n";
                    this.html += `<span>${line}</span>\n`
                    break;
            }
        } else {
            var source = text
            while (source.length) {
                if (source.length < this.width) {
                    this.text(source, align)
                    source = ""
                } else {
                    let wrap_position = source.slice(0, this.width).lastIndexOf(" ")
                    let cut_position = wrap_position + 1
                    if (wrap_position === -1) {
                        wrap_position = this.width
                        cut_position = this.width
                    }
                    this.text(source.slice(0, wrap_position), align)
                    source = source.slice(cut_position)
                }
            }
        }
    }

    /**
     * Додає запис до чеку із товаром, кількістю, ціною, сумою та податковою ставкою
     * @example <caption>Приклад виклику</caption>
     * receipt.product({ name: 'Кетчуп "Торчин" Лагідний 400гр.', price: 24.5, quantity:2, total: 49.00, tax: 'A')
     * 
     * 
     * |                ...                 |
     * |Кетчуп "Торчин" Лагідний 400гр.     | <-
     * |2.000 x 24.50               49.00  А| <-
     * |                ...                 |
     */
    product({ name, quantity, price, total, tax }) {
        const str_quantity = typeof quantity === "string" ? quantity : quantity.toFixed(3)
        const str_price = typeof price === "string" ? price : price.toFixed(2)
        const str_total = typeof total === "string" ? total : total.toFixed(2)
        const str_tax = typeof tax === "string" ? tax : tax.toString()
        const str_total_with_tax = `${str_total} ${str_tax}`
        this.text(name)
        this.text(`${str_quantity} x ${str_price}`.padEnd(this.width - str_total_with_tax.length, " ") + str_total_with_tax)
    }



    /**
     * Додає запис до чеку із сумою коштів із певним заголовком 
     * @example <caption>Приклад виклику</caption>
     * receipt.summ({ name: 'Оплата готівкою', total: 49, currency: 'Грн')
     * 
     * |                ...                  |
     * |Оплата готівкою             49.00 Грн| <-
     * |                ...                  |
     */
    summ({ name, total, currency }) {
        const str_total = typeof total === "string" ? total : total.toFixed(2)
        const str_total_with_currency = `${str_total} ${currency}`

        if (name.length + 1 + str_total_with_currency.length <= this.width) {
            this.text(name.padEnd(this.width - str_total_with_currency.length, " ") + str_total_with_currency)
        } else {
            var source = name
            while (source.length) {
                let wrap_position = source.slice(0, this.width).lastIndexOf(" ")
                let cut_position = wrap_position + 1
                if (wrap_position === -1) {
                    wrap_position = Math.min(this.width, source.length)
                    cut_position = Math.min(this.width, source.length)
                }
                let current_line = source.slice(0, wrap_position)
                let last_line = !source.slice(cut_position).length
                let free_space = this.width - current_line.length
                if (last_line) {
                    if (free_space > (1 + str_total_with_currency.length)) {
                        this.content += current_line + str_total_with_currency.padStart(this.width - current_line.length) + "\n";
                        this.html += `<span>${current_line + str_total_with_currency.padStart(this.width - current_line.length)}</span>\n`
                    } else {
                        this.content += current_line.padEnd(this.width, " ") + "\n";
                        this.html += `<span>${current_line.padEnd(this.width, " ")}</span>\n`
                        this.content += str_total_with_currency.padStart(this.width, " ") + "\n"
                        this.html += `<span>${str_total_with_currency.padStart(this.width, " ")}</span>\n`
                    }
                } else {
                    this.content += current_line.padEnd(this.width, " ") + "\n";
                    this.html += `<span>${current_line.padEnd(this.width, " ")}</span>\n`
                }
                source = source.slice(cut_position)
            }
        }
    }

    /**
       * Додає запис із двох частин до чеку із відступом поцентрку
       * @example <caption>Приклад виклику</caption>
       * receipt.justify({ left: '10.06.2022', right: '14:59:01'})
       * 
       * |                ...                  |
       * |10.06.2022                   14:59:01| <-
       * |                ...                  |
       */
    justify({ left, right }) {
        const left_part = typeof left === "string" ? left : left.toString()
        const right_part = typeof right === "string" ? right : right.toString()
        const spacing = this.width - left_part.length - right_part.length
        this.text(left_part.padEnd(left_part.length + spacing, " ") + right_part)
    }

    /**
        * Додає qr код сформований із символів
        * @example <caption>Приклад виклику</caption>
        * receipt.qr({ text: 'https://tax.gov.ua/validate?fiscal=123456789' })
        * 
        * |                ...                  |
        * |10.06.2022                   14:59:01| <-
        * |                ...                  |
        */
    qr({ text, width = this.width }) {
        this.html += QrImage.imageSync(text, { ec_level: 'L', type: 'svg' })
    }

    toText() {
        return this.content
    }
    toHTML() {
        return (
            `
<html class="receipt-html">
    <head>
    <meta charset="UTF-8">
    <style>
        @page{ margin: 0; padding: 0; }
        .receipt-html body{
             padding: 0; margin: 0; width: max-content;
        }
        .receipt-wrapper{ 
            margin: 0; margin: "receipt-wrapper-margin";
            padding: 0; padding: "receipt-wrapper-padding"; 
        }
        .receipt-content{ 
            font-size: ${this.font_size}px; 
            font-weight: ${this.font_weight}; 
            margin: 0; margin: "receipt-margin"; 
            padding: 0; padding: "receipt-padding"; 
            display: flex; flex-direction: column; 
        }
        .receipt-content svg{ 
            min-width: calc(${this.width} * 5px); min-width: calc(${this.width} * (${this.font_size}px / 3)); 
            max-width: calc(${this.width} * 5px); max-width: calc(${this.width} * (${this.font_size}px / 3)); 
            min-height: calc(${this.width} * 5px); min-height: calc(${this.width} * (${this.font_size}px / 3));
            max-height: calc(${this.width} * 5px); max-height: calc(${this.width} * (${this.font_size}px / 3));
            align-self: center; 
        }
    </style>
    </head>
    <body>
        <div class="receipt-wrapper">
            <pre class="receipt-content">${this.html}</pre>
        </div>
    <body>
</html>
`
        )
    }




}