
import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import AquaringTerminalsFactory from '../../lib/AquaringTerminalsFactory/index.js'
import Rest from '../../lib/rest.js'

export default class Acquirings {

    static prepare(data, id) {
        const { subdivision_id, name, vendor, ip, is_enabled = false } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            name,
            vendor,
            ip,
            is_enabled: is_enabled,
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.AcquiringTerminals.count({ where, transaction }))
            const result = await Database.AcquiringTerminals.findAll(Acquirings.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All acquiring terminals', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.AcquiringTerminals.findOne(Acquirings.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get acquiring terminals', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { text } = req.query
            const registers = await Database.AcquiringTerminals.findAll({
                attributes: ['id', 'name', 'is_enabled', 'status', 'error'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(is_master ? [] : [{ subdivision_id }])
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*', },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(registers))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete acquiring terminals', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const result = await Database.AcquiringTerminals.create(Acquirings.prepare({ ...req.body, subdivision_id }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create acquiring terminals', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const exists = await Database.AcquiringTerminals.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            if (req.body.is_enabled) {
                try {
                    var device = AquaringTerminalsFactory.getTerminal(id)
                    if (!device) await AquaringTerminalsFactory.addTerminal(id, req.body.vendor, req.body)
                } catch (error) {
                    console.log(error)
                    throw Error("Неможливо підключитись до терміналу")
                }
            } else {
                AquaringTerminalsFactory.removeTerminal(id)
            }
            await exists.update(Database.prepare(req.body, ['name', 'vendor', 'ip', 'is_enabled']), { transaction })
            const result = await Database.AcquiringTerminals.findOne(Acquirings.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update acquiring terminals', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.AcquiringTerminals.destroy({ where: { id, ...(is_master ? {} : { subdivision_id }) } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete acquiring terminals', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Екваєренговий термінал використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}