import { v4 as uuid } from 'uuid'
import { Op } from '@sequelize/core'
import Database from '../../database/index.js'
import { COLORS } from '../../../constants/index.js'
import Rest from '../../lib/rest.js'
import Uploads from '../../lib/Uploads.js'



export default class Categories {

    static prepare(data, id) {
        const { name, parent, parent_id, color, image, position, is_enabled = true, } = data
        return ({
            id: id ? id : uuid(),
            name: name,
            parent_id: parent || parent_id || null,
            color: color || COLORS[parseInt(Math.random() * COLORS.length) % COLORS.length],
            image: image || null,
            is_enabled
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { include: [['parent_id', 'parent']] },
            where,
            order: [['position', 'ASC']],
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.ProductCategories.count({ where, transaction }))
            const result = await Database.ProductCategories.findAll(Categories.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All products categories', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.ProductCategories.findOne(Categories.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get products categories', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text } = req.query
            const result = await Database.ProductCategories.findAll({
                attributes: ['id', 'name'],
                where: text ? Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`) : {},
                limit: 100,
                offset: 0,
                transaction,
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete products categories', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async create(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            let result = null
            if (Array.isArray(req.body)) {
                result = await Database.ProductCategories.bulkCreate(req.body.map(item => Categories.prepare(item, item.id)), { transaction })
            } else {
                result = await Database.ProductCategories.create(Categories.prepare(req.body, req.body.id), { transaction })
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products categories', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            const exists = await Database.ProductCategories.findOne({ where: { id }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            req.body['parent_id'] = req.body['parent_id'] || req.body['parent']
            await exists.update(Database.prepare(req.body, ['name', 'parent_id', 'color', 'image', 'is_enabled']), { transaction })
            const result = await Database.ProductCategories.findOne(Categories.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update products categories', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        try {
            const { id } = req.params
            await Database.ProductCategories.destroy({ where: { id } })
            res.setHeader('Content-Type', 'application/json')
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete products categories', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Категорія використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async positions(req, res) {
        const transaction = await Database.instance.transaction()
        try {
            const changes = req.body
            for (let i = 0; i < changes.length; i++) {
                const { id, position } = changes[i]
                await Database.ProductCategories.update({ position }, { where: { id }, transaction })
            }
            const result = await Database.ProductCategories.findAll(Categories.options({}, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Positions products categories', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}