import { col, Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

export default class Counterparties {

    static prepare(data, id) {
        const { name, fullname, email, phone, is_client = false, is_supplier = false, deletion_mark = false } = data
        return ({
            id: id ? id : uuid(),
            name: name || fullname,
            fullname: fullname || name,
            email,
            phone,
            is_client,
            is_supplier,
            deletion_mark
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['deletion_mark'] },
            include: [
                {
                    model: Database.CounterpartiesProducts, as: 'items', attributes: ['name'], include: [
                        { model: Database.Products, as: 'product', attributes: ['id', 'name'] },
                        { model: Database.ProductPacking, as: 'packing', attributes: ['id', 'name'] }
                    ]
                }
            ],
            where: { [Op.and]: [where, { deletion_mark: false }] },
            order: [
                ['createdAt', 'DESC'],
                ['items', 'position', 'ASC']
            ],
            ...params
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req, {})
            const pagination = Rest.pagination(req, res, await Database.Counterparties.count({ where, transaction }))
            const result = await Database.Counterparties.findAll(Counterparties.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримати перелік контрагентів', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.Counterparties.findOne(Counterparties.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримати контрагента за ідентифікатором', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text } = req.query
            const where = Rest.where(req)
            delete where.text
            const counterparties = await Database.Counterparties.findAll({
                attributes: ['id', 'name'],
                where: {
                    [Op.and]: [
                        text ? Database.instance.literal(`MATCH(fullname, name, email, phone) AGAINST(:text IN BOOLEAN MODE)`) : {},
                        where,
                        { deletion_mark: false }
                    ]
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(counterparties))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримання контрагентів для автозаповнення', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { items } = req.body
            const created = await Database.Counterparties.create(Counterparties.prepare(req.body, req.body.id), { transaction })
            const counterparty_id = created.id
            if (items && items.length) {
                const records = items.map(({ product_id, product, packing_id, packing, name }, index) => {
                    return { counterparty_id, product_id: product?.id || product_id, packing_id: packing?.id || packing_id, name, position: index }
                })
                await Database.CounterpartiesProducts.bulkCreate(records, { transaction })
            }
            const result = await Database.Counterparties.findOne(Counterparties.options({ id: counterparty_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Створення контрагента', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const { items, ...payload } = req.body
            const exists = await Database.Counterparties.findOne({ where: { id }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(Database.prepare(payload, ['name', 'fullname', 'email', 'phone', 'is_client', 'is_supplier']), { transaction })
            if (items) {
                await Database.CounterpartiesProducts.destroy({ where: { counterparty_id: id }, transaction })
                if (items.length) {
                    const records = items.map(({ product_id, product, packing_id, packing, name }, index) => {
                        return { counterparty_id: id, product_id: product?.id || product_id, packing_id: packing?.id || packing_id, name, position: index }
                    })
                    await Database.CounterpartiesProducts.bulkCreate(records, { transaction })
                }
            }
            const result = await Database.Counterparties.findOne(Counterparties.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Оновлення контрагента', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.Counterparties.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Видалення контрагента', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Контрагент використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async products(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.CounterpartiesProducts.findAll({
                attributes: ['product_id', 'packing_id'],
                include: [
                    { model: Database.Products, as: 'product', attributes: ['id', 'name'] },
                    { model: Database.ProductPacking, as: 'packing', attributes: ['id', 'name'] }
                ],
                where: { counterparty_id: id },
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримати контрагента за ідентифікатором', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


}