
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

import FiscalRegistersFactory from '../../lib/FiscalRegistersFactory/index.js'
import { Op } from '@sequelize/core'
export default class FiscalRegisters {


    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            where,
            ...params,
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.FiscalRegisters.count({ where, transaction }))
            const result = await Database.FiscalRegisters.findAll(FiscalRegisters.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All fiscal registers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.FiscalRegisters.findOne(FiscalRegisters.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get fiscal registers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { text } = req.query
            const registers = await Database.FiscalRegisters.findAll({
                attributes: ['id', 'name', 'is_enabled', 'status', 'error'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(is_master ? [] : [{ subdivision_id }])
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*', subdivision_id },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(registers))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete fiscal registers', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }

        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const payload = Database.prepare(req.body, ['name', 'is_enabled', 'is_physical', 'vendor', 'number', 'host', 'port', 'license', 'login', 'password'], [], { id: uuid(), subdivision_id })
            const created = await Database.FiscalRegisters.create(payload, { transaction })
            const result = await Database.FiscalRegisters.findOne(FiscalRegisters.options({ id: created.id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create fiscal registers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const exist = await Database.FiscalRegisters.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exist) throw Error('Помилка редагування. Не знайдено обʼєкт редагування, або у користувача недостатньо прав.')
            exist.update(Database.prepare(req.body, ['name', 'is_enabled', 'is_physical', 'vendor', 'number', 'host', 'port', 'license', 'login', 'password']), { transaction })
            const register = await FiscalRegistersFactory.getRegister(id)
            if (register) {
                FiscalRegistersFactory.removeRegister(id)
                if (req.body.is_enabled) await FiscalRegistersFactory.addRegister(id, req.body)
            } else {
                if (req.body.is_enabled) await FiscalRegistersFactory.addRegister(id, req.body)
            }
            const result = await Database.FiscalRegisters.findOne(FiscalRegisters.options({ id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update fiscal registers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.FiscalRegisters.destroy({ where: { id, subdivision_id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete fiscal registers', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Фіскальний реєстратор використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}