import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import { Op, col, literal } from '@sequelize/core'
import Rest from '../../lib/rest.js'

export default class Packings {

    static prepare(data, id) {
        const { product_id, unit_id, coefficient } = data
        return ({
            id: id ? id : uuid(),
            product_id,
            unit_id,
            coefficient: coefficient || 1,
        })
    }

    static options(where = {}, params = {}) {
        return ({
            attributes: {},
            where,
            order: [['createdAt', 'ASC']],
            ...params
        })
    }


    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.ProductPacking.count({ where, transaction }))
            const result = await Database.ProductPacking.findAll(Packings.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All products packings', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            const result = await Database.ProductPacking.findOne(Packings.options({ id }, { transaction, ...pagination }))
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Get products packings', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text, product_id } = req.query
            const result = await Database.ProductPacking.findAll({
                attributes: ['id', 'name', 'coefficient'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(product_id ? [{ product_id }] : [])
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                transaction,
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete products packings', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            await Database.ProductPacking.create(Packings.prepare(req.body), { transaction })
            const result = await Database.ProductPacking.findOne(Packings.options({ id }, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products packings', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.ProductPacking.update(Packings.prepare(req.body, id), { where: { id } })
            res.status(200).send(JSON.stringify(await Packings.payment({ id })))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Update products packings', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.ProductPacking.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete products packings', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Пакування використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}