import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

export default class PaymentsMethods {

    static prepare(data, id) {
        const { subdivision_id, type, name, is_enabled = false, position, acquiring_id, rounding, rounding_rule } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            type,
            name,
            is_enabled,
            position: position || 0,
            acquiring_id: acquiring_id || null,
            rounding,
            rounding_rule
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.AcquiringTerminals, as: 'acquiring', attributes: ['id', 'name', 'status', 'error'], require: false }
            ],
            where,
            order: [['position', 'ASC']],
            ...params,
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.PaymentMethods.count({ where, transaction }))
            const result = await Database.PaymentMethods.findAll(PaymentsMethods.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All payments methods', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.PaymentMethods.findOne(PaymentsMethods.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get payments methods', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const position = await Database.PaymentMethods.max("position", { where: { subdivision_id }, transaction })
            const created = await Database.PaymentMethods.create(PaymentsMethods.prepare({ ...req.body, subdivision_id, position: position + 1 }, req.body.id), { transaction })
            const result = await Database.PaymentMethods.findOne(PaymentsMethods.options({ id: created.id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create payments methods', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const exists = await Database.PaymentMethods.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(Database.prepare(req.body, ['type', 'name', 'is_enabled', 'position', 'acquiring_id', 'rounding', 'rounding_rule']), { transaction })
            const result = await Database.PaymentMethods.findOne(PaymentsMethods.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update payments methods', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        try {
            res.setHeader('Content-Type', 'application/json')
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.PaymentMethods.destroy({ where: { id, subdivision_id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete payments methods', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Платіжний метод використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async positions(req, res) {
        const transaction = await Database.instance.transaction()
        try {
            const changes = req.body
            for (let i = 0; i < changes.length; i++) {
                const { id, position } = changes[i]
                await Database.PaymentMethods.update({ position }, { where: { id }, transaction })
            }
            const result = await Database.PaymentMethods.findAll(PaymentsMethods.options({}, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Positions payments methods', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}