
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

export default class PrescriptionSuppliers {

    static prepare(data, id) {
        const { name, is_enabled = false, vendor, license, login, password } = data
        return ({
            id: id ? id : uuid(),
            name,
            is_enabled,
            vendor,
            license,
            login,
            password
        })
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.PrescriptionSuppliers.count({ where, transaction }))
            const result = await Database.PrescriptionSuppliers.findAll({ attributes: { exclude: ['subdivision_id'] }, where, transaction, ...pagination })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All prescription suppliers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.PrescriptionSuppliers.findByPk(id, { attributes: { exclude: ['subdivision_id'] }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get prescription suppliers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const created = await Database.PrescriptionSuppliers.create(PrescriptionSuppliers.prepare(req.body, req.body.id), { transaction })
            const result = await Database.PrescriptionSuppliers.findOne({ where: { id: created.id }, attributes: { exclude: ['subdivision_id'] }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create prescription suppliers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            await Database.PrescriptionSuppliers.update(req.body, { where: { id }, transaction })
            const result = await Database.PrescriptionSuppliers.findByPk(id, { attributes: { exclude: ['subdivision_id'] }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update prescription suppliers', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.PrescriptionSuppliers.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete prescription suppliers', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Постачальник рецептів використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}