import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

export default class PriceTypes {

    static prepare(data, id) {
        const { name, type, rounding, rounding_rule } = data
        return ({
            id: id ? id : uuid(),
            name,
            type,
            rounding,
            rounding_rule,
        })
    }

    static options(where = {}, params = {}) {
        return {
            where: { [Op.and]: [{ deletion_mark: false }, where] },
            ...params
        }
    }

    static async getOne(where = {}, params = {}) {
        return await Database.PriceTypes.findOne({ where, ...params })
    }

    static async getAll(where = {}, params = {}) {
        return await Database.PriceTypes.findAll({ where, order: [['createdAt', 'DESC']], ...params })
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.PriceTypes.count({ where, transaction }))
            const result = await Database.PriceTypes.findAll(PriceTypes.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All price types', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.PriceTypes.findOne(PriceTypes.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get price types', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text } = req.query
            const query = Rest.where(req)
            delete query.text
            const result = await Database.PriceTypes.findAll({
                attributes: ['id', 'name', 'rounding', 'rounding_rule'],
                where: {
                    [Op.and]: [
                        { deletion_mark: false },
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        query
                    ]
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete price types', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const created = await Database.PriceTypes.create(PriceTypes.prepare({ ...req.body }, req.body.id), { transaction })
            const result = await Database.PriceTypes.findOne(PriceTypes.options({ id: created.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create price types', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const exists = await Database.PriceTypes.findOne({ where: { id }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(Database.prepare(req.body, ['type', 'name', 'rounding', 'rounding_rule']), { transaction })
            const result = await Database.PriceTypes.findOne(PriceTypes.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update price types', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.PriceTypes.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete price types', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Платіжний метод використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}