import { v4 as uuid } from 'uuid'
import sequelize, { Op, HasMany, QueryTypes, literal, HasOne, BelongsTo } from '@sequelize/core'
import { COLORS } from '../../../constants/index.js'
import Database from '../../database/index.js'
import path from 'path'
import fs from 'fs'
import config from '../../../config/index.js'
import Rest from '../../lib/rest.js'
import SettingsHelper from '../../database/schemes/dictionaries/system/settings_helper.js'
import ProductPriceRegisterValuesHelper from '../../database/schemes/registers/product/product_price_values_helper.js'
import ProductStockRegisterValuesHelper from '../../database/schemes/registers/product/product_stock_values_helper.js'
import moment from 'moment'

export default class Products {

    static prepare(data, id) {
        const {
            name, name_in_receipt, article, uktzed, is_enabled = true,
            tax_id, kind_id, category_id, packing_id,
            producer_id, active_substance_id, substitutes_id,
            is_drug, is_insulin, is_thermolabile,
            need_accounting, need_prescription
        } = data
        return ({
            id: id ? id : uuid(),
            name,
            name_in_receipt,
            article,
            uktzed,
            tax_id: tax_id || null,
            kind_id: kind_id || null,
            category_id: category_id || null,
            packing_id: packing_id || null,
            is_enabled,

            producer_id: producer_id || null,
            active_substance_id: active_substance_id || null,
            substitutes_id: substitutes_id || null,

            is_drug,
            is_insulin,
            is_thermolabile,
            need_accounting,
            need_prescription,
        })
    }


    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['deletion_mark', 'createdAt', 'updatedAt'] },
            include: [
                { model: Database.Taxes, as: 'tax', required: false, attributes: { exclude: ['createdAt', 'updatedAt'] }, },
                { model: Database.ProductKinds, as: 'kind', required: false, attributes: { exclude: ['deletion_mark', 'createdAt', 'updatedAt'] }, },
                { model: Database.ProductCategories, as: 'category', required: false, attributes: ['id', 'parent_id', 'name', 'is_enabled'], },
                { model: Database.ProductProducers, as: 'producer', required: false, attributes: { exclude: ['deletion_mark', 'createdAt', 'updatedAt'] }, },
                { model: Database.ProductActiveSubstances, as: 'active_substance', required: false, attributes: { exclude: ['deletion_mark', 'createdAt', 'updatedAt'] }, },
                {
                    model: Database.ProductPacking, as: 'packing', required: false, attributes: { exclude: ['product_id', 'unit_id', 'deletion_mark', 'created_at', 'updatedAt'] },
                    association: new BelongsTo(Database.Products, Database.ProductPacking, { targetKey: 'id', foreignKey: 'packing_id', as: 'packing' })
                }
            ],
            where,
            ...params,
            order: [...(params.order ? params.order : []), ['id', 'ASC']]
        }
    }

    static async getOne(where = {}, options = {}) {
        return await Database.Products.findOne(Products.options(where, options))
    }

    static async getAll(where = {}, options = {}) {
        return await Database.Products.findAll(Products.options(where, options))
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.Products.count({ where, transaction }))
            const result = await Database.Products.findAll(Products.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All products', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.Products.findOne(Products.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримати товар', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }




    static async substitutes(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { workplace_id } = req.auth
            const { id } = req.params

            const { price_type_id } = await Database.Workplaces.findByPk(workplace_id)

            let products = await Database.instance.query(`
                SELECT
                    products.*,
                    substances.name as active_substance_name,
                    producers.name as producer_name,
                    product_stock.series_id AS 'series_id',
                    IFNULL(product_stock.in_stock, 0) AS 'in_stock',
                    IFNULL(product_stock.to_ship, 0) AS 'to_ship',
                    IFNULL(product_price.price, 0) AS 'price',
                    product_price.packing_id AS 'packing_id',
                    product_price.price_type_id AS 'price_type_id',
                    product_stock.warehouse_id AS 'warehouse_id',
                    product_stock.assignment_id AS 'assignment_id'
                FROM (
                    SELECT
                    products.id,
                    products.name,
                    products.active_substance_id,
                    products.producer_id,
                    products.substitutes_id
                    FROM products
                    WHERE substitutes_id = :substitutes_id
                ) AS products
                LEFT JOIN products_active_substances AS substances ON products.active_substance_id = substances.id
                LEFT JOIN products_producers AS producers ON products.producer_id = producers.id
                INNER JOIN products_stock_register_view AS product_stock ON products.id = product_stock.product_id
                INNER JOIN products_price_register_values AS product_price ON products.id = product_price.product_id AND product_stock.series_id = product_price.series_id AND product_price.price_type_id = :price_type_id
                ORDER BY products.name
                `, {
                replacements: {
                    substitutes_id: id,
                    price_type_id
                },
                type: sequelize.QueryTypes.SELECT,
                nest: true
            });
            await transaction.commit()
            res.status(200).send(JSON.stringify(products))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Subtitute products', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async search(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { text, limit, offset } = req.query

            const where = Rest.where(req, {}, ['text'])

            var products = await Database.Products.findAll({
                attributes: ['id', 'name', 'active_substance_id', 'producer_id', 'substitutes_id', 'category_id'],
                where: {
                    [Op.and]: [
                        ...(text ? [literal('MATCH(name) AGAINST ( :text IN BOOLEAN MODE)')] : []),
                        where
                    ]
                },
                replacements: { text: `*${decodeURIComponent(text)}*` },
                limit: limit ? parseInt(limit) : 100,
                offset: offset ? parseInt(offset) : 0,
                transaction
            })

            if (!products.length && text) {
                products = await Database.Products.findAll({
                    attributes: ['id', 'name', 'active_substance_id', 'producer_id', 'substitutes_id', 'category_id'],
                    where: where,
                    include: [{ model: Database.Barcodes, as: 'barcodes', attributes: [], where: { barcode: { [Op.like]: '%' + text + '%' } } }],
                    transaction
                })
            }

            let products_ids = products.map(it => it.id).filter((it, index, array) => array.indexOf(it) === index)
            let warehouses = await Database.Warehouses.findAll({ where: { subdivision_id }, transaction })
            let warehouses_ids = warehouses.map(it => it.id)
            let price_types_ids = warehouses.map(it => it.price_type_id).filter((it, index, array) => array.indexOf(it) === index)

            const prices = await ProductPriceRegisterValuesHelper.getAll({
                where: {
                    product_id: products_ids,
                    price_type_id: price_types_ids
                },
                order: [['product_id', 'ASC']]
            })

            products_ids = prices.map(it => it.product_id).filter((it, index, array) => array.indexOf(it) === index)
            const stocks = await ProductStockRegisterValuesHelper.getAll({
                where: {
                    product_id: products_ids,
                    warehouse_id: warehouses_ids
                },
                order: [['product_id', 'ASC']]
            })

            const packings = await Database.ProductPacking.findAll({ attributes: ['id', 'name', 'is_weight'], where: { id: { [Op.in]: prices.map(it => it.packing_id) } }, transaction })

            const default_warehouse_id = await SettingsHelper.get('default_warehouse_id')
            const result = []
            for (let product_index = 0; product_index < products.length; product_index++) {
                const product = products[product_index].get()
                let found_price = false
                for (let price_index = 0; price_index < prices.length; price_index++) {
                    if (product.id != prices[price_index].product_id) continue;
                    found_price = true
                    const price_record = prices[price_index].get()
                    const packing_record = packings.find(it => it.id === price_record.packing_id)
                    let found_stock = false
                    for (let stock_index = 0; stock_index < stocks.length; stock_index++) {
                        if (product.id != stocks[stock_index].product_id) continue;
                        if (prices[price_index].series_id != stocks[stock_index].series_id) continue;
                        found_stock = true
                        const stock_record = stocks[stock_index].get()
                        result.push({
                            ...product,
                            series_id: price_record.series_id,
                            packing_id: price_record.packing_id,
                            packing: packing_record?.get(),
                            price_type_id: price_record.price_type_id,
                            warehouse_id: stock_record.warehouse_id,
                            assignment_id: stock_record.assignment_id,
                            price: price_record.price,
                            in_stock: stock_record.in_stock,
                            to_ship: stock_record.to_ship,
                        })
                    }
                    if (!found_stock) {
                        result.push({
                            ...product,
                            series_id: price_record.series_id,
                            packing_id: price_record.packing_id,
                            packing: packing_record?.get(),
                            price_type_id: price_record.price_type_id,
                            warehouse_id: default_warehouse_id,
                            assignment_id: null,
                            price: price_record.price,
                            in_stock: 0,
                            to_ship: 0,
                        })
                    }
                }
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Search products', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async barcode(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { barcode } = req.params

            const product_barcode = await Database.Barcodes.findOne({ where: { barcode }, transaction })
            if (!product_barcode) throw Error(`Товар із штрихкодом ${barcode} не знайдено`)

            const { product_id, packing_id } = product_barcode

            const default_warehouse_id = await SettingsHelper.get('default_warehouse_id')

            const products = await Database.Products.findAll({
                attributes: ['id', 'name', 'active_substance_id', 'producer_id', 'substitutes_id', 'category_id'],
                where: { id: product_id },
                transaction
            })
            // знаходимо усі склади цього підрозділу
            let warehouses = await Database.Warehouses.findAll({ where: { subdivision_id }, transaction })
            // масив ідентифікаторів складів
            let warehouses_ids = warehouses.map(it => it.id)
            // масив ідентифікаторів видів цін для цих складів
            let price_types_ids = warehouses.map(it => it.price_type_id).filter((it, index, array) => array.indexOf(it) === index)
            // знаходимо усі ціни для поточного товару у відповідності до доступних видів цін
            const prices = await ProductPriceRegisterValuesHelper.getAll({
                where: { product_id, packing_id, price_type_id: price_types_ids },
                order: [['product_id', 'ASC']]
            })
            // знаходимо залишки на усіх складах цього підрозділу
            const stocks = await ProductStockRegisterValuesHelper.getAll({
                where: { product_id, warehouse_id: warehouses_ids },
                order: [['product_id', 'ASC']]
            })

            const packings = await Database.ProductPacking.findAll({ attributes: ['id', 'name', 'is_weight'], where: { id: packing_id }, transaction })

            const result = []
            for (let product_index = 0; product_index < products.length; product_index++) {
                const product = products[product_index].get()
                let found_price = false
                for (let price_index = 0; price_index < prices.length; price_index++) {
                    if (product.id != prices[price_index].product_id) continue;
                    found_price = true
                    const price_record = prices[price_index].get()
                    const packing_record = packings.find(it => it.id === price_record.packing_id)
                    let found_stock = false
                    for (let stock_index = 0; stock_index < stocks.length; stock_index++) {
                        if (product.id != stocks[stock_index].product_id) continue;
                        if (prices[price_index].series_id != stocks[stock_index].series_id) continue;
                        found_stock = true
                        const stock_record = stocks[stock_index].get()
                        result.push({
                            ...product,
                            series_id: price_record.series_id,
                            packing_id: price_record.packing_id,
                            packing: packing_record?.get(),
                            price_type_id: price_record.price_type_id,
                            warehouse_id: stock_record.warehouse_id,
                            assignment_id: stock_record.assignment_id,
                            price: price_record.price,
                            in_stock: stock_record.in_stock,
                            to_ship: stock_record.to_ship,
                        })
                    }
                    if (!found_stock) {
                        result.push({
                            ...product,
                            series_id: price_record.series_id,
                            packing_id: price_record.packing_id,
                            packing: packing_record?.get(),
                            price_type_id: price_record.price_type_id,
                            warehouse_id: default_warehouse_id,
                            assignment_id: null,
                            price: price_record.price,
                            in_stock: 0,
                            to_ship: 0,
                        })
                    }
                }
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result.sort((a, b) => a.in_stock < b.in_stock ? 1 : -1)))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Search products', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async find(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            // const { subdivision_id } = req.auth
            // const { product_id, series_id = '', warehouse_id } = req.query
            // if(!product_id && !warehouse_id)
            // const default_warehouse_id = await SettingsHelper.get('default_warehouse_id')

            // const where = Rest.where(req)
            // delete where.text

            // var clause = ""
            // const keys = Object.keys(where)
            // if (text) {
            //     clause = "MATCH(name) AGAINST ( :text IN BOOLEAN MODE)"
            //     if (keys.length) clause += " AND "
            // }
            // if (keys.length) {
            //     clause += keys.filter(it => it != 'text').map(key => {
            //         switch (typeof where[key]) {
            //             case 'object': return ` products.${key} IN (${where[key].map(it => `'${decodeURIComponent(it)}'`).join(',')}) `
            //             case 'string': return ` products.${key} = '${decodeURIComponent(where[key])}' `
            //         }
            //     }).join(' AND ')
            // }

            // const products = await Database.Products.findAll({
            //     attributes: ['id', 'name', 'active_substance_id', 'producer_id', 'substitutes_id', 'category_id'],
            //     where: literal(clause),
            //     replacements: { text: `*${decodeURIComponent(text)}*` },
            //     limit: limit ? parseInt(limit) : 100,
            //     offset: offset ? parseInt(offset) : 0,
            //     transaction
            // })

            // let products_ids = products.map(it => it.id).filter((it, index, array) => array.indexOf(it) === index)
            // let warehouses = await Database.Warehouses.findAll({ where: { subdivision_id }, transaction })
            // let warehouses_ids = warehouses.map(it => it.id)
            // let price_types_ids = warehouses.map(it => it.price_type_id).filter((it, index, array) => array.indexOf(it) === index)

            // const prices = await ProductPriceRegisterValuesHelper.getAll({
            //     where: {
            //         product_id: products_ids,
            //         price_type_id: price_types_ids
            //     },
            //     order: [['product_id', 'ASC']]
            // })

            // products_ids = prices.map(it => it.product_id).filter((it, index, array) => array.indexOf(it) === index)
            // const stocks = await ProductStockRegisterValuesHelper.getAll({
            //     where: {
            //         product_id: products_ids,
            //         warehouse_id: warehouses_ids
            //     },
            //     order: [['product_id', 'ASC']]
            // })

            // const packings = await Database.ProductPacking.findAll({ attributes: ['id', 'name', 'is_weight'], where: { id: { [Op.in]: prices.map(it => it.packing_id) } }, transaction })

            // const result = []
            // for (let product_index = 0; product_index < products.length; product_index++) {
            //     const product = products[product_index].get()
            //     let found_price = false
            //     for (let price_index = 0; price_index < prices.length; price_index++) {
            //         if (product.id != prices[price_index].product_id) continue;
            //         found_price = true
            //         const price_record = prices[price_index].get()
            //         const packing_record = packings.find(it => it.id === price_record.packing_id)
            //         let found_stock = false
            //         for (let stock_index = 0; stock_index < stocks.length; stock_index++) {
            //             if (product.id != stocks[stock_index].product_id) continue;
            //             if (prices[price_index].series_id != stocks[stock_index].series_id) continue;
            //             found_stock = true
            //             const stock_record = stocks[stock_index].get()
            //             result.push({
            //                 ...product,
            //                 series_id: price_record.series_id,
            //                 packing_id: price_record.packing_id,
            //                 packing: packing_record?.get(),
            //                 price_type_id: price_record.price_type_id,
            //                 warehouse_id: stock_record.warehouse_id,
            //                 assignment_id: stock_record.assignment_id,
            //                 price: price_record.price,
            //                 in_stock: stock_record.in_stock,
            //                 to_ship: stock_record.to_ship,
            //             })
            //         }
            //         if (!found_stock) {
            //             result.push({
            //                 ...product,
            //                 series_id: price_record.series_id,
            //                 packing_id: price_record.packing_id,
            //                 packing: packing_record?.get(),
            //                 price_type_id: price_record.price_type_id,
            //                 warehouse_id: default_warehouse_id,
            //                 assignment_id: null,
            //                 price: price_record.price,
            //                 in_stock: 0,
            //                 to_ship: 0,
            //             })
            //         }
            //     }
            // }
            // await transaction.commit()
            // res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Find products', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async autocomplete(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { text, counterparty_id } = req.query

            const ids_by_counterparty = counterparty_id ? await Database.CounterpartiesProducts.findAll({ attributes: ['product_id'], where: { counterparty_id }, transaction }).then(rows => rows.map(it => it.product_id)) : []

            const products = await Database.Products.findAll({
                attributes: ['id', 'name', 'article', 'uktzed', 'is_enabled'],
                include: [
                    { model: Database.Taxes, as: 'tax', required: false, attributes: ['id', 'name'] },
                    { model: Database.ProductKinds, as: 'kind', required: false, attributes: ['id', 'name'], },
                    { model: Database.ProductCategories, as: 'category', required: false, attributes: ['id', 'name', 'is_enabled'], },
                    { model: Database.ProductProducers, as: 'producer', required: false, attributes: ['id', 'name'], },
                    { model: Database.ProductActiveSubstances, as: 'active_substance', required: false, attributes: ['id', 'name'], },
                ],
                where: {
                    [Op.and]: [
                        text ? Database.instance.literal(`MATCH(products.name) AGAINST(:text IN BOOLEAN MODE)`) : {},
                        counterparty_id ? { id: ids_by_counterparty } : {}
                    ]
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] })
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(products))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete products', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }

        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const products = Array.isArray(req.body) ? req.body : [req.body]
            const created = await Database.Products.bulkCreate(products.map(item => Products.prepare({ ...item, packing_id: null }, item.id)), { transaction })
            for (let j = 0; j < created.length; j++) {
                const { packings, barcodes, packing_id } = products[j]
                const product_id = created[j].id
                for (let i = 0; packings && i < packings.length; i++) {
                    const { id, name, coefficient = 1, is_weight = false } = packings[i]
                    await Database.ProductPacking.create({ id, product_id, name, coefficient: parseFloat(coefficient), is_weight }, { transaction })
                }
                for (let i = 0; barcodes && i < barcodes.length; i++) {
                    const { barcode, packing, characteristic } = barcodes[i]
                    await Database.Barcodes.create({ barcode, product_id, packing_id: packing.id, characteristic_id: characteristic?.id }, { transaction })
                }
                if (packing_id) {
                    await Database.Products.update({ packing_id }, { where: { id: product_id }, transaction })
                }
            }
            const result = await Database.Products.findAll(Products.options({ id: { [Op.in]: created.map(it => it.id) } }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(Array.isArray(req.body) ? result : result.length ? result[0] : null))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const product_id = req.params.id
            const {
                packings_created, packings_updated, packings_deleted,
                barcodes_created, barcodes_updated, barcodes_deleted
            } = req.body

            await Database.Products.update(Products.prepare(req.body, product_id), { where: { id: product_id }, transaction })

            for (let i = 0; packings_created && i < packings_created.length; i++) {
                const { id, name, coefficient = 1, is_weight = false } = packings_created[i]
                await Database.ProductPacking.create({ id, product_id, name, coefficient: parseFloat(coefficient), is_weight }, { transaction })
            }
            for (let i = 0; barcodes_created && i < barcodes_created.length; i++) {
                const { barcode, packing, characteristic } = barcodes_created[i]
                await Database.Barcodes.create({ barcode, product_id, packing_id: packing.id, characteristic_id: characteristic?.id }, { transaction })
            }

            for (let i = 0; packings_updated && i < packings_updated.length; i++) {
                const { id, name, coefficient = 1, is_weight = false } = packings_updated[i]
                await Database.ProductPacking.update({ name, coefficient: parseFloat(coefficient), is_weight }, { where: { id }, transaction })
            }
            for (let i = 0; barcodes_updated && i < barcodes_updated.length; i++) {
                const { barcode, packing, characteristic } = barcodes_updated[i]
                await Database.Barcodes.update({ packing_id: packing.id, characteristic_id: characteristic?.id }, { where: { barcode }, transaction })
            }

            for (let i = 0; barcodes_deleted && i < barcodes_deleted.length; i++) {
                await Database.Barcodes.destroy({ where: { barcode: barcodes_deleted[i].barcode }, transaction })
            }
            for (let i = 0; packings_deleted && i < packings_deleted.length; i++) {
                await Database.ProductPacking.destroy({ where: { id: packings_deleted[i].id }, transaction })
            }

            const result = await Database.Products.findOne(Products.options({ id: product_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case "SequelizeForeignKeyConstraintError":
                    if (error?.sql?.includes('DELETE FROM `products_packings`') && error?.index === 'products_barcodes_ibfk_2') {
                        return res.status(500).send(JSON.stringify({ message: `Неможливо видалити пакування яке використовується для штрихкоду` }))
                    }
                case "SequelizeUniqueConstraintError":
                    if (error?.sql?.includes('INSERT INTO `products_barcodes`') && error?.fields?.PRIMARY) {
                        return res.status(500).send(JSON.stringify({ message: `Штрихкод ${error?.fields?.PRIMARY} вже використовується` }))
                    }
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.Products.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete products', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Товар використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }



    static async import(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { products, counterparties, categories, packings, barcodes, taxes, price_in_register, price_out_register } = req.body
            for (let i = 0; categories && i < categories.length; i++) {
                const color = categories[i].color || COLORS[i % COLORS.length]
                await Database.ProductCategories.upsert({ ...categories[i], color }, { transaction })
            }
            for (let i = 0; taxes && i < taxes.length; i++) {
                await Database.Taxes.upsert(taxes[i], { transaction })
            }
            for (let i = 0; products && i < products.length; i++) {
                await Database.Products.upsert(products[i], { transaction })
            }
            for (let i = 0; packings && i < packings.length; i++) {
                await Database.ProductPacking.upsert(packings[i], { transaction })
            }
            for (let i = 0; barcodes && i < barcodes.length; i++) {
                await Database.Barcodes.upsert(barcodes[i], { transaction })
            }
            for (let i = 0; counterparties && i < counterparties.length; i++) {
                await Database.Counterparties.upsert(counterparties[i], { transaction })
                for (let j = 0; counterparties[i].items && j < counterparties[i].items.length; j++) {
                    await Database.CounterpartiesProducts.upsert({ ...counterparties[i].items[j], counterparty_id: counterparties[i].id }, { transaction })
                }
            }

            const basis_id = ''
            const series_id = ''
            const period = moment(0).toISOString()
            for (let i = 0; price_in_register && i < price_in_register.length; i++) {
                const price_type_id = config.price_type_in.id
                const { product_id, packing_id, price } = price_in_register[i]
                await Database.Registers.ProductPrice.upsert({ basis_id, product_id, series_id, packing_id, price_type_id, price, period }, { transaction })
            }
            for (let i = 0; price_out_register && i < price_out_register.length; i++) {
                const price_type_id = config.price_type_out.id
                const { product_id, packing_id, price } = price_out_register[i]
                await Database.Registers.ProductPrice.upsert({ basis_id, product_id, series_id, packing_id, price_type_id, price, period }, { transaction })
            }
            await ProductPriceRegisterValuesHelper.recalculate({ transaction })

            await transaction.commit()
            res.status(200).send(JSON.stringify({}))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Import products', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


}