import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'


export default class Barcodes {

    static prepare(data) {
        const { barcode, product_id, packing_id, characteristic_id } = data
        return ({
            barcode,
            product_id,
            packing_id,
            characteristic_id
        })
    }

    static options(where = {}, options = {}) {
        return {
            attributes: { exclude: ['createdAt', 'updatedAt'] },
            include: [
                { model: Database.Products, as: 'product', attributes: ['id', 'name'], require: false },
                { model: Database.ProductPacking, as: 'packing', attributes: ['id', 'name'], require: false }
            ],
            where,
            ...options
        }
    }


    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            //const { id } = req.user
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.Barcodes.count({ where, transaction }))
            const result = await Database.Barcodes.findAll(Barcodes.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { barcode } = req.params
            const result = await Database.Barcodes.findOne(Barcodes.options({ barcode }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text, product_id } = req.query
            const result = await Database.Barcodes.findAll({
                attributes: { exclude: ['createdAt', 'updatedAt'] },
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(barcode) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(product_id ? [{ product_id }] : [])
                    ]
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                transaction,
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Пошук пакування товару', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            let created = await Database.Barcodes.create(Barcodes.prepare(req.body), { transaction })
            let result = await Database.Barcodes.findOne(Barcodes.options({ barcode: created.barcode }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { barcode } = req.params
            await Database.Barcodes.update(Database.prepare(req.body, ['barcode', 'product_id', 'packing_id', 'characteristic_id']), { where: { barcode } })
            const result = await Database.Barcodes.findOne(Barcodes.options({ barcode: req.body?.barcode || barcode }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { barcode } = req.params
            await Database.Barcodes.destroy({ where: { barcode } })
            res.status(200).send(JSON.stringify({ barcode }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete products shortcut', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Запис використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}