import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'


export default class Packings {

    static prepare(data, id) {
        const { product_id, name, coefficient, is_weight } = data
        return ({
            id: id || uuid(),
            product_id,
            name,
            coefficient: coefficient || 1.0,
            is_weight: is_weight || false
        })
    }

    static options(where = {}, options = {}) {
        return {
            attributes: { exclude: ['updatedAt'] },
            where,
            ...options,
            order: [['createdAt', 'ASC']],
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            //const { id } = req.user
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.ProductPacking.count({ where, transaction }))
            const result = await Database.ProductPacking.findAll(Packings.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримання списку пакування товару', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.ProductPacking.findOne(Packings.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Отримання пакування товару за ідентифікатором', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text, product_id } = req.query
            const result = await Database.ProductPacking.findAll({
                attributes: ['id', 'name', 'coefficient', 'is_weight'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(product_id ? [{ product_id }] : [])
                    ]
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                transaction,
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Пошук пакування товару', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            let created = await Database.ProductPacking.create(Packings.prepare(req.body), { transaction })
            let result = await Database.ProductPacking.findOne(Packings.options({ id: created.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Оновлення пакування товару', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { barcode } = req.params
            const [row_id, result] = await Database.ProductPacking.update(Database.prepare(req.body, ['barcode', 'product_id', 'packing_id', 'characteristic_id']), { where: { barcode } })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Оновлення пакування товару', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.ProductPacking.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Видалення пакування товару', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Запис використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}