import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'
import { Op } from '@sequelize/core'

export default class Series {


    static prepare(data, id) {
        const { product_id, kind_id, name, expiration_date } = data
        return ({
            id: id ? id : uuid(),
            product_id,
            kind_id,
            name,
            expiration_date
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            where,
            order: [['createdAt', 'DESC']],
            ...params
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Series.count({ where, transaction }))
            const result = await Database.Series.findAll(Series.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All series', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Series.findOne(Series.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get series', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { text, product_id } = req.query
            const result = await Database.Series.findAll({
                attributes: ['id', 'name', 'product_id'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(is_master ? [] : [{ subdivision_id }])
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                order: text ? [] : [['createdAt', 'DESC']],
                having: product_id ? { product_id } : {},
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete series', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const created = await Database.Series.create(Series.prepare({ ...req.body, subdivision_id }, req.body.id), { transaction })
            const result = await Database.Series.findOne(Series.options({ id: created.id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create series', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Series.update(Series.prepare(req.body, id), { where: { id, subdivision_id }, transaction })
            const result = await Database.Series.findOne(Series.options({ id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update series', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Series.destroy({ where: { id, subdivision_id }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Delete series', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Серія використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}