import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import { COLORS } from '../../../constants/index.js'
import Rest from '../../lib/rest.js'


export default class Shortcuts {

    static prepare(data, id) {
        const { product_id, packing_id, subdivision_id, quantity = 1.0, name, color, image, position = 0, is_enabled = true } = data
        return ({
            id: id ? id : uuid(),
            subdivision_id,
            product_id,
            packing_id,
            quantity,
            name,
            color: color || COLORS[parseInt(Math.random() * COLORS.length) % COLORS.length],
            image: image || null,
            position,
            is_enabled,
        })
    }

    static options(where = {}, options = {}) {
        return {
            attributes: { exclude: ['subdivision_id', 'deletion_mark', 'createdAt', 'updatedAt'] },
            include: [
                { model: Database.Products, as: 'product', attributes: ['id', 'name'] },
                { model: Database.ProductPacking, as: 'packing', attributes: ['id', 'name'] }
            ],
            where,
            order: [['position', 'ASC']],
            ...options
        }
    }


    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id, } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.ProductShortcats.count({ where, transaction }))
            const result = await Database.ProductShortcats.findAll(Shortcuts.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id, } = req.auth
            const { id } = req.params
            const result = await Database.ProductShortcats.findOne(Shortcuts.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            let position = await Database.ProductShortcats.max('position', { transaction })
            let created = await Database.ProductShortcats.create(Shortcuts.prepare({ ...req.body, subdivision_id, position: position + 1 }, req.body.id), { transaction })
            let result = await Database.ProductShortcats.findOne(Shortcuts.options({ id: created.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const exists = await Database.ProductShortcats.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(Database.prepare(req.body, ['product_id', 'packing_id', 'quantity', 'name', 'color', 'image', 'is_enabled']), { transaction })
            const result = await Database.ProductShortcats.findOne(Shortcuts.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update products shortcut', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.ProductShortcats.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete products shortcut', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Запис використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async positions(req, res) {
        const transaction = await Database.instance.transaction()
        try {
            const changes = req.body
            for (let i = 0; i < changes.length; i++) {
                const { id, position } = changes[i]
                await Database.ProductShortcats.update({ position }, { where: { id }, transaction })
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(await Database.ProductShortcats.findAll(Shortcuts.options())))
        } catch (error) {
            await transaction.rollback()
            res.status(500).send(JSON.stringify({ message: error.message }))
        }

    }
}