
import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import ScaleFactory from '../../lib/ScaleFactory/index.js'
import Rest from '../../lib/rest.js'

export default class Scales {

    static prepare(data, id) {
        const { subdivision_id, name, vendor, port, is_enabled = false } = data
        return ({
            id: id ? id : uuid(),
            subdivision_id,
            name,
            vendor,
            port,
            is_enabled,
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Scales.count({ where, transaction }))
            const result = await Database.Scales.findAll(Scales.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All scales', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Scales.findOne(Scales.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get scale', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { text } = req.query
            const registers = await Database.Scales.findAll({
                attributes: ['id', 'name', 'is_enabled', 'status', 'error'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                        ...(is_master ? [] : [{ subdivision_id }])
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(registers))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete scales', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async weight(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id } = req.auth
            const workplace = await Database.Workplaces.findOne({ where: { id: workplace_id, subdivision_id }, transaction })
            if (!workplace) throw Error('Неможливо знайти поточний РМК')
            if (!workplace.scale_id) throw Error('Для поточного РМК вага не налаштована')
            var scale = await Database.Scales.findOne({ where: { id: workplace.scale_id }, transaction })
            if (!scale) throw Error('Вагу для цього РМК не знайдено у системі')
            const amount = await ScaleFactory.useScale(scale.get())
            if (typeof amount !== 'number') throw Error('Не вдалось зчитати занчення із ваги')
            await transaction.commit()
            res.status(200).send(JSON.stringify({ amount }))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Product weight draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const result = await Database.Scales.create(Scales.prepare({ ...req.body, subdivision_id }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create scale', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            if (req.body.is_enabled) {
                try {
                    const amount = await ScaleFactory.useScale(req.body)
                    if (amount === null) throw Error("Неможливо підключитись до вагів")
                } catch (error) {
                    throw Error("Неможливо підключитись до вагів")
                }
            }
            await Database.Scales.update(Scales.prepare(req.body, id), { where: { id, subdivision_id }, transaction })
            const result = await Database.Scales.findOne(Scales.options({ id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update scale', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Scales.destroy({ where: { id, subdivision_id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete scale', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Екваєренговий термінал використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}