import { v4 as uuid } from 'uuid'
import md5 from 'md5'
import { Op } from '@sequelize/core'
import Database from '../../database/index.js'
import JsonWebToken from '../../lib/JsonWebToken.js'
import Rest from '../../lib/rest.js'
import config from '../../../config/index.js'

export default class Staff {

    static prepare(data, id) {
        const { subdivision_id, firstname, lastname, fathername, role_id, email = '', phone, password, pin } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            role_id,
            firstname,
            lastname,
            fathername: fathername || null,
            email,
            phone: phone || null,
            pin,
            ...(password && !id ? { password: md5(password) } : {})
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id', 'password', 'createdAt', 'updatedAt'] },
            include: [
                { model: Database.Roles, as: 'role', attributes: { exclude: ['subdivision_id', 'createdAt', 'updatedAt'] } }
            ],
            where,
            order: [['createdAt', 'DESC']],
            ...params
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, { id: { [Op.ne]: config.system_user.id }, ...(is_master ? {} : { subdivision_id }) })
            const pagination = Rest.pagination(req, res, await Database.Staff.count({ where, transaction }))
            const result = await Database.Staff.findAll(Staff.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Staff.findOne(Staff.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const created = await Database.Staff.create(Staff.prepare({ ...req.body, subdivision_id }, req.body.id), { transaction })
            const result = await Database.Staff.findOne(Staff.options({ id: created.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Staff.update(Staff.prepare(req.body, id), { where: { id, subdivision_id }, transaction })
            const result = await Database.Staff.findOne(Staff.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Staff.destroy({ where: { id, subdivision_id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete staff', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Користувач використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async signin(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, user_id, workplace_id } = req.auth
            const { pin, email, password } = req.query
            if (pin && !workplace_id) return res.status(400).send(JSON.stringify({ message: "Потрібно спочатку авторизувати термінал" }))
            if (pin) {
                const result = await Database.Staff.findOne({ where: { pin, subdivision_id } })
                if (!result) return res.status(401).send({ message: "Невірні дані авторизації" })
                res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id, workplace_id, staff_id: result.id }) }))
            } else {
                const result = await Database.Staff.findOne({ where: { email, password: md5(password).toString() } })
                if (!result) return res.status(401).send({ message: "Невірні дані авторизації" })
                res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id: result.id, workplace_id }) }))
            }
        } catch (error) {
            await Database.onCatchError(error, { name: 'Signin staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async adminpanel(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            if (!staff_id) throw Error('Касир не авторизований в системі')
            return res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id: staff_id, workplace_id, staff_id }) }))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async signout(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            let { subdivision_id, workplace_id } = req.auth
            res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, workplace_id }) }))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async signinManager(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, user_id, workplace_id, staff_id } = req.auth
            const { pin } = req.query
            if (pin && !workplace_id) return res.status(400).send(JSON.stringify({ message: "Потрібно спочатку авторизувати термінал" }))
            if (!pin) return res.status(400).send(JSON.stringify({ message: "Очікується коректний пін-код" }))
            const result = await Database.Staff.findOne({ where: { pin, subdivision_id } })
            if (!result) return res.status(401).send({ message: "Невірні дані авторизації" })
            res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id, workplace_id, staff_id, manager_id: result.id }) }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Signin staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async signoutManager(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            let { subdivision_id, user_id, workplace_id, staff_id } = req.auth
            res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id, workplace_id, staff_id }) }))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async profile(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, user_id } = req.auth
            if (!user_id) {
                await transaction.rollback()
                return res.status(401).send(JSON.stringify({ message: "Користувача не авторизовано" }))
            }
            const result = await Database.Staff.findOne(Staff.options({ id: user_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get profile staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async cashier(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, staff_id } = req.auth
            const result = await Database.Staff.findOne(Staff.options({ id: staff_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get cashier staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async reset(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const { password } = req.body
            await Database.Staff.update({ password: md5(password).toString() }, { where: { id, subdivision_id }, transaction })
            const result = await Database.Staff.findOne(Staff.options({ id: staff_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Reset password staff', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}