import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import Rest from '../../lib/rest.js'

export default class Taxes {

    static prepare(data, id) {
        const { name, value, code, letter } = data
        return ({
            id: id ? id : uuid(),
            name: name,
            value: typeof value === "number" ? value : parseInt(value),
            code: code || null,
            letter: letter || null
        })
    }

    static options(where = {}, params = {}) {
        return {
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.Taxes.count({ where, transaction }))
            const result = await Database.Taxes.findAll(Taxes.options(where, { ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All taxes', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const result = await Database.Taxes.findOne(Taxes.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get taxes', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text } = req.query
            const taxes = await Database.Taxes.findAll({
                attributes: ['id', 'name'],
                where: {
                    [Op.and]: [
                        ...(text ? [Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`)] : []),
                    ],
                },
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                limit: 100,
                offset: 0,
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(taxes))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete taxes', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            if (Array.isArray(req.body)) {
                const created = await Database.Taxes.bulkCreate(req.body.map(item => Taxes.prepare(item, item.id)), { transaction })
                const result = await Database.Taxes.findAll(Taxes.options({ id: { [Op.in]: created.map(it => it.id) } }, { transaction }))
                await transaction.commit()
                res.status(200).send(JSON.stringify(result))
            } else {
                const created = await Database.Taxes.create(Taxes.prepare(req.body, req.body.id), { transaction })
                const result = await Database.Taxes.findOne(Taxes.options({ id: created.id }, { transaction }))
                await transaction.commit()
                res.status(200).send(JSON.stringify(result))
            }
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create taxes', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            await Database.Taxes.update(Taxes.prepare(req.body, id), { where: { id }, transaction })
            const result = await Database.Taxes.findOne(Taxes.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update taxes', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { id } = req.params
            await Database.Taxes.destroy({ where: { id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete taxes', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Податкова ставка використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

}