
import { Op } from '@sequelize/core'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import JsonWebToken from '../../lib/JsonWebToken.js'
import Rest from '../../lib/rest.js'
import Shifts from '../shifts/index.js'

export default class Workplaces {

    static prepare(data, id) {
        const { subdivision_id, name, login, password, register_id, scale_id, message = "", printing = true } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            name,
            login,
            password,
            register_id,
            scale_id,
            message,
            printing
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.FiscalRegisters, as: 'register', required: false, attributes: ['id', 'name', 'is_enabled', 'status', 'error'] },
                { model: Database.Scales, as: 'scale', required: false, attributes: ['id', 'name', 'is_enabled', 'status', 'error'] }
            ],
            where,
            order: [['createdAt', 'DESC']],
            ...params
        }
    }


    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Workplaces.count({ where, transaction }))
            const result = await Database.Workplaces.findAll(Workplaces.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Workplaces.findOne(Workplaces.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { register_id, scale_id } = req.body
            if (register_id) {
                const register = await Database.FiscalRegisters.findOne({ where: { id: register_id, subdivision_id }, transaction })
                if (!register) throw Error("Обраний фіскальний реєстратор не знайдено в системі")
                if (register.used) {
                    const conflict_workplace = await Database.Workplaces.findOne({ where: { register_id, subdivision_id }, transaction })
                    if (conflict_workplace) throw Error("Фіскальний реєстратор уже використовується для іншого РМК: " + conflict_workplace.name)
                }
                await Database.FiscalRegisters.update({ used: 1 }, { where: { id: register_id, subdivision_id }, transaction })
            }
            if (scale_id) {
                const scale = await Database.Scales.findOne({ where: { id: scale_id, subdivision_id }, transaction })
                if (!scale) throw Error("Обрані ваги не знайдено в системі")
                const conflict_workplace = await Database.Workplaces.findOne({ where: { scale_id, subdivision_id }, transaction })
                if (conflict_workplace) throw Error("Вага уже використовується для іншого РМК: " + conflict_workplace.name)
            }
            const created = await Database.Workplaces.create(Workplaces.prepare({ ...req.body, subdivision_id }), { transaction })
            const result = await Database.Workplaces.findOne(Workplaces.options({ id: created.id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const { register_id, scale_id } = req.body
            const current = await Database.Workplaces.findOne(Workplaces.options({ id, subdivision_id }, { transaction }))
            if (register_id && register_id != current.getDataValue("register_id")) {
                const register = await Database.FiscalRegisters.findOne({ where: { id: register_id, subdivision_id }, transaction })
                if (!register) throw Error("Обраний фіскальний реєстратор не знайдено в системі")
                if (register.used) {
                    const conflict_workplace = await Database.Workplaces.findOne({ where: { register_id, subdivision_id }, transaction })
                    throw Error("Фіскальний реєстратор уже використовується для іншого РМК: " + conflict_workplace.name)
                }
                const shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id: id, subdivision_id }))
                if (shift.status != "CLOSED") throw Error("Неможливо змінити фіскальний реєстратор при відкритій зміні")

                await Database.FiscalRegisters.update({ used: 1 }, { where: { id: register_id, subdivision_id }, transaction })
                await Database.FiscalRegisters.update({ used: 0 }, { where: { id: current.getDataValue("register_id"), subdivision_id }, transaction })
            }

            if (scale_id && scale_id != current.getDataValue("scale_id")) {
                const scale = await Database.Scales.findOne({ where: { id: scale_id, subdivision_id }, transaction })
                if (!scale) throw Error("Обрані ваги не знайдено в системі")
                const conflict_workplace = await Database.Workplaces.findOne({ where: { id: { [Op.ne]: id }, scale_id, subdivision_id }, transaction })
                if (conflict_workplace) throw Error("Вага уже використовується для іншого РМК: " + conflict_workplace.name)
            }

            await Database.Workplaces.update(req.body, { where: { id, subdivision_id }, transaction })
            const result = await Database.Workplaces.findOne(Workplaces.options({ id: id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Workplaces.destroy({ where: { id, subdivision_id } })
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Delete workplaces', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Робоче місце касира використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }

    static async login(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, user_id } = req.auth
            const { login, password } = req.query
            const result = await Database.Workplaces.findOne({ where: { login, password, subdivision_id } })
            if (!result) return res.status(401).send({ message: "Невірні дані авторизації" })
            const token = JsonWebToken.sign({ subdivision_id, user_id, workplace_id: result.id })
            res.status(200).send(JSON.stringify({ token }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Login workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async logout(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, user_id } = req.auth
            res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id, user_id }) }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Logout workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async current(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id } = req.auth
            if (!workplace_id) {
                await transaction.rollback()
                return res.status(401).send(JSON.stringify({ message: "Спочатку проведіть авторизацію терміналу" }))
            }
            const result = await Database.Workplaces.findOne(Workplaces.options({ id: workplace_id, subdivision_id }, { transaction }))
            if (!result) {
                await transaction.rollback()
                return res.status(404).send({ message: "Не знайдено РМК для поточного ідентифікатора" })
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Current workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}