import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import ProductPriceRegisterValuesHelper from '../../../database/schemes/registers/product/product_price_values_helper.js'
import ProductStockRegisterValuesHelper from '../../../database/schemes/registers/product/product_stock_values_helper.js'
import Rest from '../../../lib/rest.js'
import ArrivalHelper from '../../../helpers/ArrivalHelper.js'

export default class ArrivalDocuments {

    static prepare(data, id = uuid()) {
        const { subdivision_id, number, name, warehouse_id, price_type_id, counterparty_id = null, createdAt } = data
        const payload = {
            id,
            subdivision_id,
            name,
            number,
            warehouse_id,
            price_type_id,
            counterparty_id
        }
        if (createdAt) payload['createdAt'] = createdAt
        return payload
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.Warehouses, as: 'warehouse', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                { model: Database.PriceTypes, as: 'price_type', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                { model: Database.Counterparties, as: 'counterparty', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                {
                    model: Database.Documents.ArrivalGoods, as: 'items', attributes: {
                        exclude: ['subdivision_id', 'document_id', 'series_id', 'product_id', 'packing_id', 'deletion_mark', 'position', 'updatedAt']
                    },
                    include: [
                        { model: Database.Products, attributes: ['id', 'name'], as: 'product' },
                        { model: Database.Series, attributes: ['id', 'name'], as: 'series' },
                        { model: Database.ProductPacking, attributes: ['id', 'name'], as: 'packing' }
                    ],
                    required: false,
                    where: { 'deletion_mark': { [Op.or]: { [Op.is]: null, [Op.eq]: false } } }
                }
            ],
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.ArrivalGoods, as: 'items' }, 'position', 'ASC']
            ],
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Arrival.count({ where, transaction }))
            const result = await Database.Documents.Arrival.findAll(ArrivalDocuments.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async pricing(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const arrival = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id }, { transaction }))
            await transaction.commit()
            const result = {
                items: arrival.items.map(item => {
                    return {
                        product: item.product,
                        series: item.series,
                        packing: item.packing,
                        price: item.price,
                        pricing_type: 'PERCANTAGE',
                        pricing_value: 0,
                        pricing_result: item.price
                    }
                })
            }
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async return(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const arrival = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id }, { transaction }))
            await transaction.commit()
            const result = {
                price_type_id: arrival.price_type_id,
                price_type: arrival.price_type,
                warehouse_id: arrival.warehouse_id,
                warehouse: arrival.warehouse,
                items: arrival.items
            }
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { warehouse_id, price_type_id, createdAt, items } = req.body

            const number = req.body.number ?? (((await Database.Documents.Arrival.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)

            const { id } = await Database.Documents.Arrival.create(ArrivalDocuments.prepare({ ...req.body, subdivision_id, number }), { transaction })
            if (items && Array.isArray(items) && items.length) {
                for (let i = 0; i < items.length; i++) {
                    await Database.Documents.ArrivalGoods.create({ ...items[i], document_id: id, subdivision_id, position: i }, { transaction })
                }
                await ArrivalHelper.ApplyProductsStockMoves({ basis_id: id, warehouse_id, period: createdAt, items }, transaction)
                await ArrivalHelper.ApplyProductsPriceMoves({ basis_id: id, price_type_id, period: createdAt, items }, transaction)
            }
            const result = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const { warehouse_id, price_type_id, createdAt, items } = req.body
            let now = new Date().toISOString()
            const exists = await Database.Documents.Arrival.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) } })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(ArrivalDocuments.prepare(req.body, id), { transaction })
            const document_subdivision_id = exists.getDataValue('subdivision_id')
            if (items && Array.isArray(items) && items.length) {
                // помічаємо попередні позиції на вилучення
                await Database.Documents.ArrivalGoods.update({ deletion_mark: true }, { where: { document_id: id }, transaction })
                for (let i = 0; i < items.length; i++) {
                    // оновлюємо нові позиції і знімаємо мітку про вилучення
                    await Database.Documents.ArrivalGoods.upsert({ ...items[i], document_id: id, subdivision_id: document_subdivision_id, position: i, deletion_mark: false, updatedAt: now }, { transaction })
                }
                await ArrivalHelper.ApplyProductsStockMoves({ basis_id: id, warehouse_id, period: createdAt, items }, transaction)
                await ArrivalHelper.ApplyProductsPriceMoves({ basis_id: id, price_type_id, period: createdAt, items }, transaction)
            }
            const result = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }




    static async add(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { document_id, product_id, series_id } = req.params
            const { amount, price } = req.body
            const document = await Database.Documents.ArrivalGoods.findOne({ where: { subdivision_id, document_id, product_id, series_id }, transaction })
            if (document) {
                await document.update({ amount, price }, { transaction })
            } else {
                await Database.Documents.ArrivalGoods.create({ subdivision_id, document_id, product_id, series_id, amount, price }, { transaction })
            }
            const result = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id: document_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
            await Database.onCatchError(error, { name: 'Add arrival documents', subdivision_id: req.auth.subdivision_id })
            await transaction.rollback()
        }
    }

    static async remove(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { document_id, product_id, series_id } = req.params
            await Database.Documents.ArrivalGoods.destroy({ where: { subdivision_id, document_id, product_id, series_id }, transaction })
            const result = await Database.Documents.Arrival.findOne(ArrivalDocuments.options({ id: document_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Remove arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async price(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { price_type_id, product_id, packing_id } = req.query
            const result = await Database.Documents.Arrival.findOne({
                include: [
                    { model: Database.Documents.ArrivalGoods, as: 'items', where: { product_id, packing_id }, attributes: ['price'], required: true }
                ],
                where: { subdivision_id, price_type_id },
                order: [['createdAt', 'DESC']],
                raw: true
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result ? { price: result['items.price'] } : null))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Remove arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async history(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { price_type_id, counterparty_id = null, items } = req.body
            const result = []
            const arrivals = await Promise.all(items.map(({ product_id, packing_id }) => Database.Documents.Arrival.findOne({
                include: [{ model: Database.Documents.ArrivalGoods, as: 'items', where: { product_id, packing_id }, attributes: ['product_id', 'packing_id', 'price', 'amount'], required: true }],
                where: {
                    price_type_id,
                    ...(counterparty_id ? { counterparty_id } : {})
                },
                order: [['createdAt', 'DESC']],
            })))
            for (let i = 0; i < arrivals.length; i++) {
                for (let j = 0; j < items.length; j++) {
                    if (!arrivals[i]) break
                    if (items[j].product_id !== arrivals[i].items[0].product_id) continue
                    if (items[j].packing_id !== arrivals[i].items[0].packing_id) continue
                    result.push({ ...items[j], price: arrivals[i].items[0].price, amount: arrivals[i].items[0].amount })
                }
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Пошук товарів в історії прихідних ордерів', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

}