import { v4 as uuid } from 'uuid'
import { Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import Helper from './helper.js'
import ProductPriceRegisterValuesHelper from '../../../database/schemes/registers/product/product_price_values_helper.js'
import SettingsHelper from '../../../database/schemes/dictionaries/system/settings_helper.js'
import Staffs from '../../../models/dictionaries/staff.js'
import DraftHelper from '../../../helpers/DraftHelper.js'
import Round from '../../../lib/Round.js'



export default class Drafts {


    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id, staff_id } = req.auth
            if (is_master) throw Error('Із центральної бази заборонений вхід на робоче місце касира')
            const result = await Helper.getAll({ ...req.query, subdivision_id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id } = req.auth
            if (is_master) throw Error('Із центральної бази заборонений вхід на робоче місце касира')
            const { id } = req.params
            const result = await Helper.getOne({ id, subdivision_id })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            if (!workplace_id) {
                await transaction.rollback()
                return res.status(401).send(JSON.stringify({ message: "РМК не авторизовано" }))
            }
            if (!staff_id) {
                await transaction.rollback()
                return res.status(401).send(JSON.stringify({ message: "Касир не авторизований" }))
            }

            const drafts = await Helper.getAll({ subdivision_id, workplace_id, status: "NEW" }, { transaction, order: [['number', 'ASC']] })
            for (let i = 0; i < drafts.length; i++) {
                if (!drafts[i].goods.length) {
                    await transaction.commit()
                    return res.status(200).send(JSON.stringify(drafts[i]))
                }
            }

            const number = req.body.number ?? (((await Database.Documents.Draft.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)

            const draft = await Database.Documents.Draft.create(Helper.prepare({ ...req.body, number, workplace_id, staff_id, subdivision_id }, req.body.id), { transaction })
            const goods = req.body.goods && Array.isArray(req.body.goods) ? req.body.goods.map((item, position) => Helper.prepareGood({ ...item, document_id: draft.id, subdivision_id, position })) : []
            if (goods.length) await Promise.all(goods.map(item => Database.Documents.DraftGoods.create(item, { transaction })))

            const result = await Helper.getOne({ id: draft.id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.Draft.update(req.body, { where: { id, subdivision_id }, transaction })
            const result = await Helper.getOne({ id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async product(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id, staff_id, manager_id } = req.auth
            const { id: document_id } = req.params

            const warehouse_id = req.body.warehouse_id || await SettingsHelper.get('default_warehouse_id', { subdivision_id, transaction })

            const { barcode } = req.body

            if (barcode && !req.body.product_id) {
                if (!warehouse_id) {
                    await transaction.rollback()
                    return res.status(400).send(JSON.stringify({ message: "В налаштуваннях системи не обрано склад реалізації по замовчуванню" }))
                }

                const product_barcode = await Database.Barcodes.findOne({ where: { barcode }, transaction })
                if (!product_barcode) {
                    await transaction.rollback()
                    return res.status(400).send(JSON.stringify({ message: "Не знайдено товару із штрихкодом " + barcode }))
                } else {
                    req.body.product_id = product_barcode.product_id
                    req.body.packing_id = req.body.packing_id || product_barcode.packing_id || ''
                }
            }

            const warehouse = await Database.Warehouses.findOne({ where: { id: warehouse_id, subdivision_id }, transaction })
            if (!warehouse) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: "Не знайдено склад на якому зберігається товар" }))
            }

            const workplace = await Database.Workplaces.findOne({ where: { id: workplace_id, subdivision_id }, transaction })
            if (!workplace) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: "Не знайдено робочого місця касира" }))
            }


            const {
                product_id,
                series_id = '',
                packing_id = '',
                price_type_id = warehouse.price_type_id,
                assignment_id = ''
            } = req.body

            const { amount } = req.body

            const packing = await Database.ProductPacking.findOne({ attributes: ['coefficient'], where: { id: packing_id }, raw: true, transaction })
            const unpacked_amount = amount * packing.coefficient

            const staff = await Database.Staff.findOne(Staffs.options({ id: staff_id }, { transaction }))
            const staff_manager = await Database.Staff.findOne(Staffs.options({ id: manager_id }, { transaction }))

            const isNegativeCorrectionAllowed = staff?.role?.negative_correction || staff_manager?.role?.negative_correction

            const draft = await Database.Documents.Draft.findOne({ where: { id: document_id, subdivision_id }, transaction })

            let available = Round.money(await Database.Registers.ProductStock.sum("in_stock", {
                where: { product_id, series_id, warehouse_id, assignment_id, basis_id: { [Op.ne]: draft.id } },
                group: ['product_id', 'series_id', 'warehouse_id', 'assignment_id'],
            }))

            if (warehouse.use_advanced_warehousing && (!available || available < unpacked_amount)) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: "Недостатньо кількості в поточній серії" + "\n" + "Доступно " + available + " одиниць" }))
            }

            let good = await Database.Documents.DraftGoods.findOne({ where: { subdivision_id, document_id, product_id, series_id, warehouse_id, packing_id }, transaction })
            let total_amount = amount
            if (good) {
                if (amount > 0) {
                    if (barcode) total_amount += good.amount
                    if (total_amount < good.amount && !isNegativeCorrectionAllowed) throw Error('Користувач немає права коригувати кількість')
                    if (total_amount != good.amount) await good.update({ amount: total_amount }, { transaction })
                } else {
                    if (!isNegativeCorrectionAllowed) throw Error('Користувач немає права коригувати кількість')
                    await good.destroy({ transaction })
                }
            } else {
                const current_position = await Database.Documents.DraftGoods.max('position', { where: { subdivision_id, document_id }, transaction })
                const position = current_position !== null ? (current_position + 1) : 0
                const price = await ProductPriceRegisterValuesHelper.get({ where: { product_id, series_id, packing_id, price_type_id }, transaction })
                if (!price) {
                    await transaction.rollback()
                    return res.status(400).send(JSON.stringify({ message: "Не знайдено ціну для поточного товару" }))
                }
                await Database.Documents.DraftGoods.create({ id: uuid(), subdivision_id, document_id, product_id, series_id, packing_id, warehouse_id, amount, price, position }, { transaction })
            }

            DraftHelper.ApplyProductStockMoves({ basis_id: document_id, warehouse_id, assignment_id, product_id, series_id, in_stock: -total_amount * packing.coefficient })

            const result = await Helper.getOne({ id: document_id, subdivision_id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Product update draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async cancel(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.Draft.update({ status: "CANCELED" }, { where: { id, subdivision_id }, transaction })
            await Database.Registers.ProductStock.update({ deletion_mark: true }, { where: { basis_id: id }, transaction })
            const result = await Helper.getOne({ id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Cancel draft documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}