import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import ProductPriceRegisterValuesHelper from '../../../database/schemes/registers/product/product_price_values_helper.js'
import Rest from '../../../lib/rest.js'

export default class PricingDocuments {

    static prepare(data, id = uuid()) {
        const { subdivision_id, price_type_id, name = null, number, createdAt } = data
        const payload = {
            id,
            subdivision_id,
            name,
            number,
            price_type_id
        }
        if (createdAt) payload['createdAt'] = createdAt
        return payload
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.Documents.Arrival, as: 'basis', attributes: ['id', 'presentation'], require: false },
                { model: Database.PriceTypes, as: 'price_type', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                {
                    model: Database.Documents.PricingGoods, as: 'items', attributes: {
                        exclude: ['subdivision_id', 'document_id', 'series_id', 'product_id', 'packing_id', 'position', 'updatedAt']
                    },
                    include: [
                        { model: Database.Products, attributes: ['id', 'name'], as: 'product', required: false },
                        { model: Database.Series, attributes: ['id', 'name'], as: 'series', required: false },
                        { model: Database.ProductPacking, attributes: ['id', 'name'], as: 'packing', required: false }
                    ],
                    required: false,
                    where: { 'deletion_mark': { [Op.or]: { [Op.is]: null, [Op.eq]: 0 } } },
                }
            ],
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.PricingGoods, as: 'items' }, 'position', 'ASC']
            ],
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Pricing.count({ where, transaction }))
            const result = await Database.Documents.Pricing.findAll(PricingDocuments.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All pricing documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Documents.Pricing.findOne(PricingDocuments.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get pricing documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const { price_type_id, createdAt:period } = req.body

            const number = req.body.number ?? (((await Database.Documents.Pricing.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)

            const document = await Database.Documents.Pricing.create(PricingDocuments.prepare({ ...req.body, number, subdivision_id }, id), { transaction })
            const document_id = document.getDataValue('id')

            const items = req.body.items?.map(it => ({ ...it, series_id: it.series_id || '' }))
            if (items) {
                // записуємо табличну частину документу
                for (let position = 0; position < items.length; position++) {
                    await Database.Documents.PricingGoods.create({ ...items[position], document_id, subdivision_id, position }, { transaction })
                }
                // готуємо записи у формат регістру цін
                const price_counter_items = items.map(it => ({ basis_id: document_id, product_id: it.product_id, series_id: it.series_id || '', packing_id: it.packing_id, price_type_id, price: it.pricing_result, period }))
                for (let i = 0; i < price_counter_items.length; i++) {
                    // оновлюємо нові записи регістру і знімаємо мітку про вилучення
                    await Database.Registers.ProductPrice.create(price_counter_items[i], { transaction })
                }
                // перераховуємо актуальні ціни у службовій табличці
                await Promise.all(price_counter_items.map(({ product_id, series_id, packing_id, price_type_id }) => (
                    ProductPriceRegisterValuesHelper.recalculate({ where: { product_id, series_id, packing_id, price_type_id }, transaction })
                )))
            }
            const result = await Database.Documents.Pricing.findOne(PricingDocuments.options({ id: document_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, is_master } = req.auth
            const { id } = req.params
            const { price_type_id, createdAt:period } = req.body

            const exists = await Database.Documents.Pricing.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) } })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            await exists.update(PricingDocuments.prepare(req.body, id), { transaction })

            const document_id = id
            const document_subdivision_id = exists.getDataValue('subdivision_id')
            const updatedAt = new Date().toISOString()
            const items = req.body.items?.map(it => ({ ...it, series_id: it.series_id || '' }))

            if (items && Array.isArray(items) && items.length) {
                // помічаємо попередні позиції на вилучення
                await Database.Documents.PricingGoods.update({ deletion_mark: true }, { where: { document_id: id }, transaction })
                for (let position = 0; position < items.length; position++) {
                    // оновлюємо нові позиції і знімаємо мітку про вилучення
                    await Database.Documents.PricingGoods.upsert({ ...items[position], document_id: id, subdivision_id: document_subdivision_id, position, deletion_mark: false, updatedAt }, { transaction })
                }
                // помічаємо попередні записи регістру на вилучення
                await Database.Registers.ProductPrice.update({ deletion_mark: true }, { where: { basis_id: document_id } })
                // готуємо записи у формат регістру цін
                const price_counter_items = items.map(it => ({ basis_id: document_id, product_id: it.product_id, series_id: it.series_id || '', packing_id: it.packing_id, price_type_id, price: it.pricing_result, period }))
                for (let i = 0; i < price_counter_items.length; i++) {
                    // оновлюємо нові записи регістру і знімаємо мітку про вилучення
                    await Database.Registers.ProductPrice.upsert({ ...price_counter_items[i], deletion_mark: false }, { transaction })
                }
                // перераховуємо актуальні ціни у службовій табличці
                await Promise.all(price_counter_items.map(({ product_id, series_id, packing_id, price_type_id }) => (
                    ProductPriceRegisterValuesHelper.recalculate({ where: { product_id, series_id, packing_id, price_type_id }, transaction })
                )))
            }
            const result = await Database.Documents.Pricing.findOne(PricingDocuments.options({ id: document_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update pricing documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}