import { v4 as uuid } from 'uuid'
import sequelize from '@sequelize/core'
import Database from '../../../database/index.js'
import Round from '../../../lib/Round.js'


export default class Receipts {

    static prepare(data, id) {
        const { subdivision_id, basis_id, number, type, presentation, staff_id, workplace_id, shift_id, register_id, fiscal_id, fiscal_code, fiscal_date, fiscal_sell_offline } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            basis_id,
            number,
            type: type || "SELL",
            presentation: presentation || "",
            staff_id,
            workplace_id,
            shift_id,
            register_id: register_id || null,
            fiscal_id: fiscal_id || null,
            fiscal_code: fiscal_code || null,
            fiscal_date: fiscal_date || null,
            fiscal_sell_offline: !!fiscal_sell_offline
        })
    }

    static prepareGood(data) {
        const { subdivision_id, document_id, product_id, series_id, packing_id, warehouse_id, tax_id, amount, price } = data
        return ({
            subdivision_id,
            document_id,
            product_id,
            series_id,
            packing_id,
            warehouse_id,
            tax_id,
            amount,
            price
        })
    }

    static preparePayment(data) {
        const { subdivision_id, document_id, method_id, amount, amount_rounded, total, status, invoice_number, reference_number, card_number, card_expiry } = data
        return ({
            subdivision_id,
            document_id,
            method_id,
            amount,
            amount_rounded: amount_rounded ?? null,
            total: total || null,
            status: status || "AWAIT",
            invoice_number: invoice_number || null,
            reference_number: reference_number || null,
            card_number: card_number || null,
            card_expiry: card_expiry || null
        })
    }

    static query(where = {}, params = {}) {
        const options = {
            ...params,
            attributes: {
                exclude: ['subdivision_id', 'workplace_id', 'staff_id', 'shift_id']
            },
            include: [
                { model: Database.Workplaces, as: 'workplace', attributes: ["id", "name", "register_id"] },
                { model: Database.FiscalRegisters, as: 'register', attributes: ["id", "name", "number"] },
                { model: Database.Documents.Shifts, as: 'shift', attributes: ["id", "number", "status", "register_id"] },
                { model: Database.Staff, as: 'staff', attributes: ["id", "firstname", "lastname"] },
                {
                    attributes: { exclude: ['subdivision_id'] },
                    model: Database.Documents.ReceiptGoods, as: 'goods',
                    attributes: {
                        include: [
                            [sequelize.literal('ROUND(`goods`.`amount`*`goods`.`price`, 2)'), 'total'],
                        ],
                        exclude: ['subdivision_id', 'document_id', 'createdAt', 'updatedAt']
                    },
                    include: [
                        // {
                        //     model: Database.Barcodes,
                        //     as: 'barcode',
                        //     attributes: ['barcode'],
                        //     on: { [sequelize.literal("`goods`.`product_id`")]: sequelize.literal("`barcode`.`amount`") }

                        // },
                        {
                            model: Database.Products, as: 'product', attributes: { exclude: ['subdivision_id', 'createdAt', 'updatedAt'] },
                            include: [
                                {
                                    model: Database.Taxes, as: 'tax', attributes: { exclude: ['subdivision_id', 'createdAt', 'updatedAt'] },
                                }
                            ],
                        }
                    ],
                },
                {
                    model: Database.Documents.ReceiptPayment, as: 'payments',
                    attributes: {
                        include: [
                            [sequelize.literal('`payments->method`.`type`'), 'type'],
                            [sequelize.literal('`payments->method`.`name`'), 'name'],
                            [sequelize.literal('`payments->method`.`acquiring_id`'), 'acquiring_id'],
                            [sequelize.literal('`payments`.`createdAt`'), 'createdAt'],
                            [sequelize.literal('`payments`.`updatedAt`'), 'updatedAt'],
                        ],
                        exclude: ['subdivision_id', 'document_id', 'createdAt', 'updatedAt']
                    },
                    include: [
                        {
                            attributes: [],
                            // attributes: {
                            //     exclude: ['subdivision_id', 'id', 'type', 'acquiring_id', 'name', 'is_enabled', 'position', 'createdAt', 'updatedAt']
                            // },
                            model: Database.PaymentMethods, as: 'method',
                        }
                    ]
                },
            ],
            order: [['createdAt', 'DESC']],
        }
        if (where) options['where'] = where
        return options
    }

    static async getOne(where, params = {}) {
        const result = where.id ?
            await Database.Documents.Receipt.findByPk(where.id, Receipts.query({}, params))
            :
            await Database.Documents.Receipt.findOne(Receipts.query(where, params))
        const receipt = result ? result.get({ plain: true }) : null
        if (receipt) {
            receipt['total'] = Round.money([0, receipt.goods.map(it => it.amount * it.price)].reduce((prev, next) => prev + next))
        }
        return receipt
    }

    static async getAll(where = null, params = {}) {
        const result = await Database.Documents.Receipt.findAll(Receipts.query(where, params))
        const receipts = result ? result.map(it => it.get({ plain: true })) : []
        receipts.map(receipt => {
            receipt['total'] = Round.money([0, receipt.goods.map(it => it.amount * it.price)].reduce((prev, next) => prev + next))
        })
        return receipts
    }


}