import { v4 as uuid } from 'uuid'
import sequelize, { HasOne, Transaction } from '@sequelize/core'
import moment from 'moment'
//import printer from '@thiagoelg/node-printer')
import Database from '../../../database/index.js'
import SettingsHepler from '../../../database/schemes/dictionaries/system/settings_helper.js'
import Shifts from '../../shifts/index.js'
import Drafts from '../drafts/helper.js'
import Helper from './helper.js'
import Rest from '../../../lib/rest.js'


import FiscalRegistersFactory from '../../../lib/FiscalRegistersFactory/index.js'
import AquaringTerminalsFactory from '../../../lib/AquaringTerminalsFactory/index.js'
import ReceiptGenerator from '../../../lib/ReceiptGenerator.js'
import ProductStockRegisterValuesHelper from '../../../database/schemes/registers/product/product_stock_values_helper.js'
import SocketFactory from '../../../lib/SocketFactory/index.js'
import config from '../../../../config/index.js'
import Round from '../../../lib/Round.js'
import Time from '../../../lib/Time.js'

const timeout = (ms, data) => new Promise(resolve => setTimeout(() => resolve(data), ms))

export default class Receipts {


    static query(where = {}, params = {}) {
        return {
            ...params,
            attributes: {
                exclude: ['workplace_id', 'staff_id', 'shift_id']
            },
            include: [
                { model: Database.Workplaces, as: 'workplace', attributes: ["id", "name", "register_id"] },
                { model: Database.Documents.Shifts, as: 'shift', attributes: ["id", "number", "status", "register_id"] },
                { model: Database.Staff, as: 'staff', attributes: ["id", "firstname", "lastname"] },
                {
                    model: Database.Documents.ReceiptGoods, as: 'goods',
                    attributes: {
                        include: [
                            [sequelize.literal('ROUND(`goods`.`amount`*`goods`.`price`, 2)'), 'total'],
                        ],
                        exclude: ['document_id', 'product_id', 'updatedAt']
                    },
                    include: [
                        {
                            model: Database.Products, as: 'product', attributes: { exclude: ['createdAt', 'updatedAt'] },
                            include: [
                                {
                                    model: Database.Taxes, as: 'tax', attributes: { exclude: ['createdAt', 'updatedAt'] },
                                }
                            ],
                        }
                    ],
                },
                {
                    model: Database.Documents.ReceiptPayment, as: 'payments',
                    attributes: {
                        include: [
                            [sequelize.literal('`payments->method`.`type`'), 'type'],
                            [sequelize.literal('`payments->method`.`name`'), 'name'],
                            [sequelize.literal('`payments->method`.`acquiring_id`'), 'acquiring_id'],
                            [sequelize.literal('`payments`.`createdAt`'), 'createdAt'],
                            [sequelize.literal('`payments`.`updatedAt`'), 'updatedAt'],
                        ],
                        exclude: ['document_id', 'createdAt', 'updatedAt']
                    },
                    include: [
                        {
                            attributes: {
                                exclude: ['id', 'type', 'acquiring_id', 'name', 'is_enabled', 'position', 'createdAt', 'updatedAt']
                            },
                            model: Database.PaymentMethods, as: 'method',
                        }
                    ]
                },
            ],
            where,
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.ReceiptGoods, as: 'goods' }, 'createdAt', 'ASC']
            ],
        }
    }



    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Receipt.count(where, { transaction }))
            const result = await Helper.getAll(where, { transaction, ...pagination })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Helper.getOne({ id, ...(is_master ? {} : { subdivision_id }) })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { draft_id, payments } = req.body

            var receipt = await Helper.getOne({ subdivision_id, basis_id: draft_id, type: "SELL" }, { transaction })
            if (receipt != null) {
                await Database.Documents.ReceiptPayment.destroy({ where: { document_id: receipt.id, subdivision_id }, transaction })
                await Database.Documents.ReceiptPayment.bulkCreate(payments.map(payment => Helper.preparePayment({
                    subdivision_id,
                    document_id: receipt.id,
                    ...payment
                })), { transaction })
                await transaction.commit()
                return res.status(200).send(JSON.stringify(receipt))
            }

            if (!payments || !payments.length) throw Error('Відсутні дані про оплату')

            const shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift || shift.status !== "OPENED") throw Error('Для оплати замовлення потрібно відкрити касову зміну')
            if (Time.duration(shift.open_transaction.createdAt) > 24 * 60 * 60 * 1000 - 60 * 1000) throw Error("Зміна відкрита більше 24 годин")

            const draft_goods = await Database.Documents.DraftGoods.findAll({
                include: [{ model: Database.Products, as: 'product', attributes: { exclude: ['subdivision_id', 'createdAt', 'updatedAt'] } },],
                where: { document_id: draft_id },
                transaction
            })
            if (!draft_goods.length) throw Error('Не знайдено документа підстави для створення чеку')

            const cash_methods = await Database.PaymentMethods.findAll({ attributes: ['id'], where: { type: 'CASH' }, transaction })
            const cash_methods_ids = cash_methods.map(it => it.id)

            const draft_total = Round.money(draft_goods.map(it => Round.money(it.amount * it.price)).reduce((prev, next) => prev + next))

            const total = Round.money([0, ...payments.map(payment => payment.amount)].reduce((prev, next) => prev + next))
            const total_cash = Round.money([0, ...payments.filter(it => cash_methods_ids.includes(it.method_id)).map(payment => payment.amount)].reduce((prev, next) => prev + next))
            const total_diff = Round.money(total - draft_total)

            if ((total < draft_total && total_cash === 0) || total_diff < -0.05) throw Error('Вартість чеку перевищує суму оплати')
            const rest = Round.money(total - draft_total, 10)
            if (rest > total_cash) throw Error('Сума решти перевищує оплату готівкою')
            if (rest > 0) {
                for (let i = 0; i < payments.length; i++) {
                    if (cash_methods_ids.includes(payments[i].method_id)) {
                        payments[i].total = payments[i].amount
                        payments[i].amount = Round.money(payments[i].total - total_diff)
                        payments[i].amount_rounded = Round.money(payments[i].total - total_diff, 10)
                        break
                    }
                }
            } else {
                if (total_diff > 0) {
                    for (let i = 0; i < payments.length; i++) {
                        if (cash_methods_ids.includes(payments[i].method_id)) {
                            payments[i].amount = Round.money(payments[i].amount - total_diff)
                            payments[i].amount_rounded = Round.money(payments[i].amount, 10)
                            break
                        }
                    }
                }
            }

            const number = req.body.number ?? (((await Database.Documents.Receipt.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)

            const createdAt = moment().toISOString()
            receipt = await Database.Documents.Receipt.create(Helper.prepare({
                subdivision_id,
                number,
                type: "SELL",
                basis_id: draft_id,
                register_id: shift.register_id,
                workplace_id,
                staff_id,
                shift_id: shift.id,
                createdAt
            }), { transaction })

            const goods = draft_goods.map((good, index) => ({
                subdivision_id,
                document_id: receipt.id,
                product_id: good.product_id,
                series_id: good.series_id,
                packing_id: good.packing_id,
                warehouse_id: good.warehouse_id,
                tax_id: good.product.tax_id,
                amount: good.amount,
                price: good.price,
                createdAt: moment(createdAt).add(index, 'second')
            }))
            await Database.Documents.ReceiptGoods.bulkCreate(goods, { transaction })
            await Database.Documents.ReceiptPayment.bulkCreate(payments.map(payment => Helper.preparePayment({
                subdivision_id,
                document_id: receipt.id,
                ...payment
            })), { transaction })
            await Database.Documents.Draft.update({ status: "CLOSED" }, { where: { id: draft_id }, transaction })
            const result = await Helper.getOne({ id: receipt.id, subdivision_id, }, { transaction })
            await transaction.commit()
            req.params = { id: receipt.id }
            //await Receipts.pay(req, res)
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.Receipt.update(req.body, { where: { id, subdivision_id }, transaction })
            const result = await Helper.getOne({ id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.commit()
            await Database.onCatchError(error, { name: 'Update receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async pay(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id } = req.auth
            const { id } = req.params
            const { skip_rro } = req.query

            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (!receipt || !receipt.payments.length) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: 'Відсутні дані про оплату' }))
            }


            let shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього робочого місця")
            if (shift.status !== "OPENED") throw Error("Не знайдено відкритої зміни для цього робочого місця")

            var rro = null
            if (shift.register_id && !skip_rro) {
                rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                await rro.getShift(false, shift.status)
                if (!rro.checkSyncOfShift(shift.id)) throw Error("В реєстраторі і обліковій системі розсинхронізовані зміни")
            }


            const payments_transaction = await Database.instance.transaction() // TODO: await Database.instance.transaction({ transaction }) // can try child transactions
            const payments_unpaid = receipt.payments.filter(it => it.status === "AWAIT")
            const payments_results = await Promise.all(payments_unpaid.map(async payment => {
                try {
                    await Database.Registers.Money.create({ basis_id: receipt.id, subdivision_id, workplace_id, operation: "SELL", type: payment.type, amount: payment.amount_rounded ?? payment.amount }, { transaction: payments_transaction })
                    switch (payment.type) {
                        case 'CARD':
                            if (payment.acquiring_id) {
                                var acquiring = await AquaringTerminalsFactory.getTerminal(payment.acquiring_id)
                                if (!acquiring) {
                                    const terminal = await Database.AcquiringTerminals.findOne({ id: payment.acquiring_id, subdivision_id }, { transaction: payments_transaction })
                                    if (!terminal) throw Error("Не знайдено екваєренгового терміналу")
                                    acquiring = await AquaringTerminalsFactory.addTerminal(terminal.id, terminal.vendor, terminal)
                                }
                                const payment_result = await acquiring.pay(payment.amount)
                                await Database.Documents.ReceiptPayment.update({
                                    invoice_number: payment_result.invoiceNumber,
                                    reference_number: payment_result.rrn,
                                    card_number: payment_result.pan,
                                    card_expiry: payment_result.cardExpiryDate,
                                    status: 'PAYED'
                                }, {
                                    where: {
                                        subdivision_id,
                                        document_id: receipt.id,
                                        method_id: payment.method_id
                                    },
                                    transaction: payments_transaction
                                })
                                return { method_id: payment.method_id, status: true, message: "" }
                            } else {
                                await Database.Documents.ReceiptPayment.update({ status: 'PAYED' }, { where: { subdivision_id, document_id: receipt.id, method_id: payment.method_id }, transaction: payments_transaction })
                                return { method_id: payment.method_id, status: true, message: "" }
                            }
                        case 'CASH':
                            await Database.Documents.ReceiptPayment.update({ status: 'PAYED' }, { where: { subdivision_id, document_id: receipt.id, method_id: payment.method_id }, transaction: payments_transaction })
                            return { method_id: payment.method_id, status: true, message: "" }
                        case 'BONUSES':
                            await Database.Documents.ReceiptPayment.update({ status: 'PAYED' }, { where: { subdivision_id, document_id: receipt.id, method_id: payment.method_id }, transaction: payments_transaction })
                            return { method_id: payment.method_id, status: true, message: "" }
                        default:
                            await Database.Documents.ReceiptPayment.update({ status: 'PAYED' }, { where: { subdivision_id, document_id: receipt.id, method_id: payment.method_id }, transaction: payments_transaction })
                            return { method_id: payment.method_id, status: true, message: "" }
                    }
                } catch (error) {
                    const message = error.message.includes('ECONNREFUSED') ? "Відсутнє зʼєднання з терміналом" : error.message
                    return { method_id: payment.method_id, status: false, message }
                }
            }))
            await payments_transaction.commit()
            // if (payments_unpaid.length) {
            //     await transaction.commit()
            //     transaction = await Database.instance.transaction()
            // }

            const error = payments_results.find(it => !it.status)
            if (error) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: `Помилка оплати: ${error.message}` }))
            }


            if (rro) {
                const taxes = await Database.Taxes.findAll({ transaction })
                const payload = {
                    is_return: false,
                    products: receipt.goods.map(good => ({
                        id: `${good.product.id}-${good.series_id}`,
                        name: good.product.name_in_receipt || good.product.name,
                        uktzed: good.product.uktzed || '',
                        tax: taxes.find(it => it.id == good.tax_id)?.get(),
                        price: good.price,
                        quantity: good.amount,
                    })),
                    payments: receipt.payments.map(payment => {
                        const is_cash = payment.type == 'CASH'
                        const is_card = payment.type == 'CARD'
                        return ({
                            method: {
                                type: payment.type,
                                name: is_cash ? "Оплата готівкою" : is_card ? "Оплата картою" : "Безготівкова оплата"
                            },
                            sum: payment.amount,
                            sum_rounded: is_cash ? payment.amount_rounded ?? Round.money(payment.amount, 10) ?? payment.amount : payment.amount

                        })
                    })
                }
                await rro.getShift(false, true)
                const fiscal_receipt = await rro.sellOnline(payload)
                if (fiscal_receipt) {
                    await Database.Documents.Receipt.update({
                        fiscal_id: fiscal_receipt.id,
                        fiscal_code: fiscal_receipt.fiscal_code,
                        fiscal_date: fiscal_receipt.fiscal_date,
                        fiscal_sell_offline: fiscal_receipt.is_created_offline,
                        fiscal_send: typeof fiscal_receipt.is_sent_dps === 'boolean',
                        status: 'CLOSED'
                    }, { where: { id, subdivision_id }, transaction })
                } else {
                    throw Error("Помилка фіскалізації")
                }
            } else {
                await Database.Documents.Receipt.update({ status: 'CLOSED' }, { where: { id, subdivision_id }, transaction })
            }

            await Promise.all(receipt.goods.map(good => Database.Registers.Sales.create({
                basis_id: id, subdivision_id, workplace_id,
                staff_id: receipt.staff.id,
                product_id: good.product.id,
                category_id: good.product.category_id,
                operation: 'SELL',
                amount: good.amount,
                total: good.total
            }, { transaction })))

            const result = await Helper.getOne({ id, subdivision_id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
            await SocketFactory.sendBalance(workplace_id)
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Pay receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async confirm(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            let error = null
            const { subdivision_id, workplace_id } = req.auth
            const { id } = req.params
            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (receipt.payments.filter(it => it.status !== "PAYED" && (it.type === "BONUSES" || it.type === "CARD")).length) {
                throw Error("Чек ще не оплачений")
            } else {
                await Database.Documents.ReceiptPayment.update({ status: 'PAYED' }, { where: { document_id: id, subdivision_id }, transaction })
            }

            if (receipt.shift.status !== "OPENED") throw Error("Касова зміна закрита")

            if (receipt.shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(receipt.shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: receipt.shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(receipt.shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                if (!rro.checkSyncOfShift(receipt.shift.id)) throw Error("В реєстраторі і обліковій системі розсинхронізовані зміни")

                const taxes = await Database.Taxes.findAll({ transaction })
                const payload = {
                    is_return: false,
                    products: receipt.goods.map(good => ({
                        id: `${good.product.id}-${good.series_id}`,
                        name: good.product.name_in_receipt || good.product.name,
                        uktzed: good.product.uktzed || '',
                        tax: taxes.find(it => it.id == good.product.tax_id)?.get(),
                        price: good.price,
                        quantity: good.amount,
                    })),
                    payments: receipt.payments.map(payment => ({
                        method: {
                            type: payment.type,
                            name: payment.type == 'CASH' ? "Оплата готівкою" : payment.type == 'CARD' ? "Оплата картою" : "Безготівкова оплата"
                        },
                        sum: payment.amount
                    }))
                }

                const fiscal_receipt = await rro.sellOnline(payload)
                if (fiscal_receipt) {
                    await Database.Documents.Receipt.update({
                        fiscal_id: fiscal_receipt.id,
                        fiscal_code: fiscal_receipt.fiscal_code,
                        fiscal_date: fiscal_receipt.fiscal_date,
                        fiscal_sell_offline: fiscal_receipt.is_created_offline,
                        status: 'CLOSED'
                    }, { where: { id, subdivision_id }, transaction })
                } else {
                    throw Error("Помилка фіскалізації")
                }
            } else {
                await Database.Documents.Receipt.update({ status: 'CLOSED' }, { where: { id, subdivision_id }, transaction })
            }
            const result = await Helper.getOne({ id })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Confirm receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async restore(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params

            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (receipt.payments.filter(it => it.status === "PAYED").length) {
                if (receipt.payments.filter(it => it.status === "PAYED").length == receipt.payments.length) {
                    throw Error("Чек уже оплачений")
                } else {
                    throw Error("Чек уже частково оплачений")
                }
            }
            await Database.Documents.Draft.update({ status: "NEW" }, { where: { id: receipt.basis_id, subdivision_id }, transaction })
            await Database.Documents.Receipt.destroy({ where: { id: receipt.id, subdivision_id }, transaction })
            await Database.Documents.ReceiptGoods.destroy({ where: { document_id: receipt.id, subdivision_id }, transaction })
            await Database.Documents.ReceiptPayment.destroy({ where: { document_id: receipt.id, subdivision_id }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify({ id: receipt.id }))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Restore receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async cancel(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params

            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (receipt.payments.filter(it => it.status === "PAYED").length) {
                if (receipt.payments.filter(it => it.status === "PAYED").length == receipt.payments.length) {
                    throw Error("Чек уже оплачений")
                } else {
                    throw Error("Чек уже частково оплачений")
                }
            }
            await Database.Documents.Receipt.update({ status: "CANCELED" }, { where: { id: receipt.id, subdivision_id }, transaction })
            await Database.Registers.ProductStock.destroy({ where: { basis_id: receipt.basis_id }, transaction })

            const result = await Helper.getOne({ id, subdivision_id })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Cancel receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async returnable(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params

            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (!receipt) {
                await transaction.commit()
                return res.status(200).send(JSON.stringify(null))
            }

            const goods = receipt.goods.map(good => ({ ...good, available: good.amount }))

            const returns = await Helper.getAll({ subdivision_id, basis_id: receipt.basis_id, type: "RETURN" }, { transaction })
            returns.map(document => document.goods.map(good_r => goods.map(good_a => {
                if (good_r.product.id == good_a.product.id && good_r.series_id === good_a.series_id) good_a.available -= good_r.amount
            })))

            receipt.goods = goods.filter(good => good.available > 0)

            await transaction.commit()
            res.status(200).send(JSON.stringify(receipt))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get returnable receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async return(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { id } = req.params
            const assignment_id = ''

            const basis_receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            if (basis_receipt.status != "CLOSED") {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: 'Чек по якому створюється повернення не закритий' }))
            }

            const { goods } = req.body
            if (!goods || !goods.length) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: 'Не знайдено товарів для повернення' }))
            }

            const shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift || shift.status !== "OPENED") throw Error('Для повернення товарів потрібно відкрити касову зміну')

            const number = req.body.number ?? (((await Database.Documents.Receipt.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)
            const receipt = await Database.Documents.Receipt.create(Helper.prepare({ number, type: "RETURN", subdivision_id, basis_id: basis_receipt.basis_id, workplace_id, staff_id, shift_id: shift.id, register_id: shift.register_id }), { transaction })
            const receipt_id = receipt.getDataValue('id')

            const return_goods = goods.map(good => Helper.prepareGood({
                subdivision_id,
                document_id: receipt_id,
                product_id: good.product.id,
                series_id: good.series_id,
                packing_id: good.packing_id,
                warehouse_id: good.warehouse_id,
                tax_id: good.tax_id,
                amount: good.amount,
                price: good.price
            }))

            const amount = Round.money(return_goods.map(good => good.amount * good.price).reduce((prev, next) => prev + next))

            const payment = basis_receipt.payments.length === 1 ? basis_receipt.payments[0] : null
            if (!payment?.method_id) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: 'Не знайдено платіжного методу для повернення' }))
            }
            await Database.Documents.ReceiptGoods.bulkCreate(return_goods, { transaction })
            await Database.Documents.ReceiptPayment.create(Helper.preparePayment({
                subdivision_id,
                document_id: receipt_id,
                method_id: payment?.method_id,
                amount,
                status: "PAYED"
            }), { transaction })
            await Database.Registers.Money.create({ basis_id: receipt_id, subdivision_id, workplace_id, operation: "RETURN", type: payment.type, amount: -amount }, { transaction })

            const register_goods = goods.map(it => ({ basis_id: receipt.id, product_id: it.product.id, series_id: it.series_id, warehouse_id: it.warehouse_id, assignment_id, in_stock: it.amount, to_ship: 0, period: moment().toISOString() }))
            await Database.Registers.ProductStock.bulkCreate(register_goods, { transaction })
            // TODO: if need return payment for CARD type
            //
            // const acquiring = await AquaringTerminalsFactory.getTerminal(method.acquiring_id)
            // if (!acquiring) {
            //     throw Error("Немає пікдючення до екваєренгового терміналу")
            // }
            // const payment_result = await acquiring.pay(payment.amount)

            if (shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")

                if (!rro.isShiftOpened) throw Error("Касова зміна фіскального реєстратора закрита")
                if (rro.isShiftExpire) throw Error("Касова зміна фіскального реєстратора відкрита більше 24х годин")
                if (!rro.checkSyncOfShift(shift.id)) throw Error("В реєстраторі і обліковій системі розсинхронізовані зміни")

                const taxes = await Database.Taxes.findAll({ transaction })
                const payload = {
                    is_return: true,
                    products: goods.map(good => ({
                        id: `${good.product.id}-${good.series_id}`,
                        name: good.product.name_in_receipt || good.product.name,
                        uktzed: good.product.uktzed || '',
                        tax: taxes.find(it => it.id == good.tax_id)?.get(),
                        price: good.price,
                        quantity: good.amount,
                    })),
                    // TODO: потрібно зробити вибір платіжного методу для повернення
                    payments: [{
                        method: {
                            type: 'CASH',
                            name: 'Повернення готівкою',
                        },
                        sum: amount
                    }]
                }

                const fiscal_receipt = await rro.sellOnline(payload)
                if (fiscal_receipt) {
                    await Database.Documents.Receipt.update({
                        fiscal_id: fiscal_receipt.id,
                        fiscal_code: fiscal_receipt.fiscal_code,
                        fiscal_date: fiscal_receipt.fiscal_date,
                        fiscal_sell_offline: fiscal_receipt.is_created_offline,
                        status: 'CLOSED'
                    }, { where: { id: receipt.id, subdivision_id }, transaction })
                } else {
                    throw Error("Помилка фіскалізації")
                }
            } else {
                await Database.Documents.Receipt.update({ status: 'CLOSED' }, { where: { id: receipt.id, subdivision_id }, transaction })
            }

            await Promise.all(goods.map(good => Database.Registers.Sales.create({
                basis_id: receipt_id, subdivision_id, workplace_id, staff_id,
                product_id: good.product.id,
                category_id: good.product.category_id,
                operation: 'RETURN',
                amount: good.amount,
                total: good.total
            }, { transaction })))

            await transaction.commit()
            res.status(200).send(JSON.stringify(await Helper.getOne({ id: receipt.id })))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Return receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async withdrawal(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const receipt = await Helper.getOne({ id, subdivision_id }, { transaction })
            const payments = receipt.payments.filter(it => it.status === 'PAYED' && it.type === 'CARD' && it.amount > 0)

            if (!payments.length) {
                await transaction.rollback()
                return res.status(200).send(JSON.stringify(await Helper.getOne({ id })))
            }

            await Promise.all(payments.map(async payment => {
                const acquiring = await AquaringTerminalsFactory.getTerminal(payment.acquiring_id)
                if (!acquiring) {
                    throw Error("Немає пікдючення до екваєренгового терміналу")
                }
                await acquiring.withdrawal(payment.invoice_number)
                return await Database.Documents.ReceiptPayment.update({
                    status: 'AWAIT',
                    invoice_number: null,
                    reference_number: null,
                    card_number: null,
                    card_expiry: null,
                }, {
                    where: {
                        subdivision_id,
                        document_id: receipt.id,
                        method_id: payment.method_id
                    },
                    transaction
                })
            }))

            const result = await Helper.getOne({ id, subdivision_id }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Withdrawal receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }



    static async generate(receipt, { width = 38, font_size = 13, font_weight = 700 }, { name, address, tax_number, cash_register_name = "", cashier_name = "" }) {

        const taxes = await Database.Taxes.findAll({})
        const builder = new ReceiptGenerator({ width, font_size, font_weight })
        builder.separator(true)
        builder.text(name, ReceiptGenerator.Align.Center)
        builder.text(address, ReceiptGenerator.Align.Center)
        if (tax_number) {
            builder.text(`ПН ${tax_number}`, ReceiptGenerator.Align.Center)
        }

        builder.separator()
        switch (receipt.type) {
            case 'SELL': builder.text(`ЧЕК ПРОДАЖУ № ${receipt.number.toString().padStart(2, '0')}`, ReceiptGenerator.Align.Center); break;
            case 'RETURN': builder.text(`ЧЕК ПОВЕРНЕННЯ № ${receipt.number.toString().padStart(2, '0')}`, ReceiptGenerator.Align.Center); break;
        }
        if (cash_register_name || cashier_name) {
            builder.justify({ left: cash_register_name, right: cashier_name })
        }
        builder.separator(true)
        receipt.goods.map(good => {
            const name = good.product.name_in_receipt || good.product.name
            const price = good.price
            const quantity = good.amount
            const total = Round.money(good.price * good.amount)
            const tax = taxes.find(it => it.id == good.product.tax_id) ? taxes.find(it => it.id == good.product.tax_id)?.letter || " " : " "
            builder.product({ name, price, quantity, total, tax })
        })
        builder.separator()
        const total = [0, ...receipt.payments.map(it => it.amount_rounded ?? it.amount)].reduce((prev, next) => prev + next)
        builder.summ({ name: 'СУМА', total, currency: 'грн' })

        const action = receipt.type === "SELL" ? "Оплата" : "Повернення"
        receipt.payments.map(payment => {
            switch (payment.type) {
                case 'CASH':
                    builder.summ({ name: action + ' готівкою', total: payment.amount_rounded ?? payment.amount, currency: 'грн' })
                    if (payment.amount_rounded && payment.amount !== payment.amount_rounded) {
                        builder.summ({ name: 'Заокруглення готівки', total: Round.money(payment.amount_rounded - payment.amount), currency: 'грн' })
                    }
                    return
                case 'CARD':
                    return builder.summ({ name: action + ' картою', total: payment.amount, currency: 'грн' })
                default:
                    return builder.summ({ name: action + ' безготівкою', total: payment.amount, currency: 'грн' })
            }
        })
        taxes.map(tax => {
            let total = 0
            receipt.goods.map(good => {
                if (good.product.tax_id === tax.id) total += good.total
            })
            if (total && tax.letter) {
                const tax_value = tax.value === "string" ? parseFloat(tax.value) : tax.value || 0
                const tax_leter = tax.letter ? tax.letter : ' '
                const str_value = `ПДВ ${tax_leter} (${tax.value}%)`
                const total_value = Round.money(total / (tax_value + 100) * tax_value)
                builder.summ({ name: str_value, total: total_value, currency: 'грн' })
            }
        })
        builder.separator(true)

        if (receipt.fiscal_code) {
            builder.justify({ left: 'ФН ПРРО', right: receipt.register?.number || 'Не Вказано' })
            builder.justify({ left: new Date(receipt.fiscal_date).toLocaleDateString(), right: new Date(receipt.fiscal_date).toLocaleTimeString() })
            builder.text(receipt.fiscal_sell_offline ? "ОФЛАЙН" : "ОНЛАЙН", ReceiptGenerator.Align.Center)
            builder.text(`№ ${receipt.fiscal_code}`, ReceiptGenerator.Align.Center)
            builder.qr({ text: `https://cabinet.tax.gov.ua/cashregs/check?id=${receipt.fiscal_code}` }) // &date=20221119 &fn=4000094976
        } else {
            builder.justify({ left: new Date(receipt.createdAt).toLocaleDateString(), right: new Date(receipt.createdAt).toLocaleTimeString() })
            builder.text("НЕ ФІСКАЛЬНИЙ ЧЕК", ReceiptGenerator.Align.Center)
        }
        builder.separator(true)
        return builder.toHTML()
    }


    static async print(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'text/plain; charset=utf-8')
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            const { width, font_size, font_weight } = req.query

            const test_receipt = {
                number: 12345,
                type: "SELL",
                goods: [
                    { product: { name: 'Печево вівсяне із родзинками' }, price: 13.20, amount: 1 },
                    { product: { name: 'Йогурт із персиковим смаком' }, price: 24.56, amount: 2 },
                    { product: { name: 'Вафлі "АРТЕК" 400гр' }, price: 52.12, amount: 1 },
                ],
                payments: [{ type: 'CASH', amount: 114.44 }],
                workplace: { name: 'Каса №1' },
                staff: { lastname: 'Васьків', firstname: 'Назар' },
                total: 114.44,
                fiscal_code: 'u98yht40pkws',
                fiscal_date: new Date().toISOString(),
                fiscal_sell_offline: false
            }

            const receipt = id === 'test' ? test_receipt : await Helper.getOne({ id, subdivision_id }, { transaction })
            if (!receipt) throw Error("Не знайдено чеку із цим ідентифікатором")

            const html = await Receipts.generate(receipt, { width, font_size, font_weight }, {
                name: await SettingsHepler.get('receipt_subdivision_name', { subdivision_id, transaction }) || "Назву організації не вказано у налаштуваннях",
                address: await SettingsHepler.get('receipt_subdivision_address', { subdivision_id, transaction }) || "Адресу організації не вказано у налаштуваннях",
                tax_number: await SettingsHepler.get('receipt_subdivision_tax_number', { subdivision_id, transaction }),
                cash_register_name: receipt.workplace.name,
                cashier_name: `${receipt.staff.lastname}${(receipt.staff.firstname ? ` ${receipt.staff.firstname.slice(0, 1)}.` : '')}`,
            })
            await transaction.commit()
            res.status(200).send(html)
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Print receipt documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }



    static async sync(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id } = req.auth
            const workplace = await Database.Workplaces.findOne({ where: { id: workplace_id } })
            if (!workplace) throw Error("Не знайдено РМК")
            await Receipts.syncOfflineReceipts(workplace.register_id)
            await SocketFactory.sendBalance(workplace_id)
            res.status(200).send(JSON.stringify([]))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Sync receipts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async hasOfflineReceipts(register_id) {
        try {
            return await Database.Documents.Receipt.count(Receipts.query({ register_id, fiscal_sell_offline: true, fiscal_send: false })) > 0
        } catch (error) {
            await Database.onCatchError(error, { name: 'Receipts.hasOfflineReceipts', subdivision_id: config.subdivision.id })
            return false
        }
    }

    static async syncOfflineReceipts(register_id) {
        const transaction = await Database.instance.transaction()
        const subdivision_id = config.subdivision.id
        try {

            let shift = await Database.Documents.Shifts.findOne(Shifts.options({ register_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього РРО")
            if (shift.status !== "OPENED") throw Error("Не знайдено відкритої зміни для цього РРО")

            const receipts = await Database.Documents.Receipt.findAll(Receipts.query({ register_id, fiscal_sell_offline: true, fiscal_send: false }, { transaction, order: ['createdAt', 'ASC'] }))
            if (!receipts.length) {
                await transaction.commit()
                return true
            }

            var rro = null
            if (shift.register_id) {
                rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                if (!rro.isShiftOpened) throw Error("Касова зміна фіскального реєстратора закрита")
                if (rro.isShiftExpire) throw Error("Касова зміна фіскального реєстратора відкрита більше 24х годин")
                if (!rro.checkSyncOfShift(shift.id)) throw Error("В реєстраторі і обліковій системі розсинхронізовані зміни")
            }

            if (!rro.isOffline) {
                const fiscal_date = receipts[0].createdAt
                const offline_code = await rro.getOfflineCode()
                if (offline_code) {
                    const fiscal_code = offline_code.fiscal_code
                    const error = await rro.goOffline(fiscal_date, fiscal_code)
                    if (!rro.isOffline && error) throw Error(error)
                    await Database.Documents.ShiftsTransactions.create({ id: uuid(), subdivision_id, shift_id: shift.id, type: 'GO_OFFLINE', fiscal_code, fiscal_date }, { transaction })
                    if (!rro.isOffline) throw Error("Не вдалось перевести РРО у офлайн режим")
                } else {
                    throw Error("Не вдалось перевести РРО у офлайн режим через нестачу офлайн кодів")
                }
            }

            const taxes = await Database.Taxes.findAll({ transaction })
            const errors = []

            for (let i = 0; i < receipts.length; i++) {
                const receipt = receipts[i]
                const payload = {
                    products: receipt.goods.map(good => ({
                        id: `${good.product.id}-${good.series_id}`,
                        name: good.product.name_in_receipt || good.product.name,
                        uktzed: good.product.uktzed || '',
                        taxes: taxes.find(it => it.id === good.tax_id) ? [{ code: taxes.find(it => it.id == good.tax_id).code }].filter(it => typeof it.code === 'number') : [],
                        price: good.price,
                        quantity: good.amount,
                        is_return: false,
                    })),
                    payments: receipt.payments.map(payment_row => {
                        const payment = payment_row.get()
                        return ({
                            method: {
                                type: payment.type,
                                name: payment.type == 'CASH' ? "Оплата готівкою" : payment.type == 'CARD' ? "Оплата картою" : "Безготівкова оплата"
                            },
                            sum: payment.amount
                        })
                    })
                }
                console.log(`[${new Date().toISOString()}] `, "Надсилання офлайн чеку у ДПС", receipt.fiscal_code, receipt.fiscal_date, JSON.stringify(payload, null, 2))
                const fiscal_receipt = await rro.sellOffline(payload, receipt.fiscal_code, receipt.fiscal_date)
                if (fiscal_receipt) {
                    if (fiscal_receipt.status === "ERROR") {
                        errors.push({ id: receipt.id, message: fiscal_receipt.message })
                    } else {
                        await Database.Documents.Receipt.update({ fiscal_id: fiscal_receipt.id, fiscal_send: true, }, { where: { id: receipt.id, subdivision_id } })
                    }
                }
            }

            if (rro.isOffline) {
                const error = await rro.goOnline()
                if (!error) {
                    await Database.Documents.ShiftsTransactions.create({ id: uuid(), subdivision_id, shift_id: shift.id, type: 'GO_ONLINE' })
                }
            }
            if (errors.length) throw Error(JSON.stringify(errors))
            await transaction.commit()
            return true
        } catch (error) {
            await transaction.rollback()
            console.log(`[${new Date().toISOString()}] `, 'Помилка надсилання офлайн чеків у ДПС', error.message)
            await Database.onCatchError(error, { name: 'Помилка надсилання офлайн чеків у ДПС', subdivision_id })
            return false
        }
    }
}




// moment(fiscal_receipt.fiscal_date,'YYYY-MM-DDTHH:mm:ss.SSSSSSZ').toISOString();