import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import ProductPriceRegisterValuesHelper from '../../../database/schemes/registers/product/product_price_values_helper.js'
import ProductStockRegisterValuesHelper from '../../../database/schemes/registers/product/product_stock_values_helper.js'
import Rest from '../../../lib/rest.js'
import ReturnHelper from '../../../helpers/ReturnHelper.js'

export default class ReturnDocuments {

    static prepare(data, id = uuid()) {
        const { subdivision_id, number, name, counterparty_id, warehouse_id, price_type_id, createdAt } = data
        const payload = {
            id,
            subdivision_id,
            name,
            number,
            counterparty_id,
            warehouse_id,
            price_type_id
        }
        if (createdAt) payload['createdAt'] = createdAt
        return payload
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.Warehouses, as: 'warehouse', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                { model: Database.PriceTypes, as: 'price_type', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                { model: Database.Counterparties, as: 'counterparty', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                {
                    model: Database.Documents.ReturnGoods, as: 'items', attributes: {
                        exclude: ['subdivision_id', 'document_id', 'series_id', 'product_id', 'packing_id', 'deletion_mark', 'position', 'updatedAt']
                    },
                    include: [
                        { model: Database.Products, attributes: ['id', 'name'], as: 'product' },
                        { model: Database.Series, attributes: ['id', 'name'], as: 'series' },
                        { model: Database.ProductPacking, attributes: ['id', 'name'], as: 'packing' }
                    ],
                    required: false,
                    where: { 'deletion_mark': { [Op.or]: { [Op.is]: null, [Op.eq]: false } } }
                }
            ],
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.ReturnGoods, as: 'items' }, 'position', 'ASC']
            ],
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Return.count({ where, transaction }))
            const result = await Database.Documents.Return.findAll(ReturnDocuments.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Documents.Return.findOne(ReturnDocuments.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { warehouse_id, items } = req.body
            const number = req.body.number ?? (((await Database.Documents.Return.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)
            const document = await Database.Documents.Return.create(ReturnDocuments.prepare({ ...req.body, number, subdivision_id }), { transaction })
            const document_id = document.getDataValue('id')
            const period = document.getDataValue('createdAt')
            if (items) {
                await Database.Documents.ReturnGoods.bulkCreate(items.map((item, position) => Object.assign({}, item, { document_id, subdivision_id, position })), { transaction })
                await ReturnHelper.ApplyProductsStockMoves({ basis_id: document_id, warehouse_id, period, items }, transaction)
            }
            const result = await Database.Documents.Return.findOne(ReturnDocuments.options({ id: document_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const { warehouse_id, createdAt, items } = req.body
            const updatedAt = new Date().toISOString()
            const exists = await Database.Documents.Return.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            exists.update(ReturnDocuments.prepare(req.body, id), { where: { id }, transaction })
            const document_subdivision_id = exists.getDataValue('subdivision_id')
            if (items) {
                // помічаємо попередні позиції на вилучення
                await Database.Documents.ReturnGoods.update({ deletion_mark: true }, { where: { document_id: id }, transaction })
                for (let position = 0; position < items.length; position++) {
                    // оновлюємо нові позиції і знімаємо мітку про вилучення
                    await Database.Documents.ReturnGoods.upsert({ ...items[position], document_id: id, subdivision_id: document_subdivision_id, position, deletion_mark: false, updatedAt }, { transaction })
                }
                // записуємо рухи документа в регістр залишків
                await ReturnHelper.ApplyProductsStockMoves({ basis_id: id, warehouse_id, period: createdAt, items }, transaction)
            }
            const result = await Database.Documents.Return.findOne(ReturnDocuments.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }



    static async add(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { document_id, product_id, series_id } = req.params
            const { amount, price } = req.body
            const document = await Database.Documents.ReturnGoods.findOne({ where: { subdivision_id, document_id, product_id, series_id }, transaction })
            if (document) {
                await document.update({ amount, price }, { transaction })
            } else {
                await Database.Documents.ReturnGoods.create({ subdivision_id, document_id, product_id, series_id, amount, price }, { transaction })
            }
            const result = await Database.Documents.Return.findOne(ReturnDocuments.options({ id: document_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
            await Database.onCatchError(error, { name: 'Add return documents', subdivision_id: req.auth.subdivision_id })
            await transaction.rollback()
        }
    }

    static async remove(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { document_id, product_id, series_id } = req.params
            await Database.Documents.ReturnGoods.destroy({ where: { subdivision_id, document_id, product_id, series_id }, transaction })
            const result = await Database.Documents.Return.findOne(ReturnDocuments.options({ id: document_id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Remove return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async price(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { price_type_id, product_id, packing_id } = req.query
            const result = await Database.Documents.Arrival.findOne({
                include: [
                    { model: Database.Documents.ArrivalGoods, as: 'items', where: { product_id, packing_id }, attributes: ['price'], required: true }
                ],
                where: { subdivision_id, price_type_id },
                order: [['createdAt', 'DESC']],
                raw: true
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(result ? { price: result['items.price'] } : null))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Remove return documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async history(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { price_type_id, counterparty_id = null, items } = req.body
            const result = []
            const returns = await Promise.all(items.map(({ product_id, packing_id }) => Database.Documents.Return.findOne({
                include: [{ model: Database.Documents.ReturnGoods, as: 'items', where: { product_id, packing_id }, attributes: ['product_id', 'packing_id', 'price', 'amount'], required: true }],
                where: {
                    price_type_id,
                    ...(counterparty_id ? { counterparty_id } : {})
                },
                order: [['createdAt', 'DESC']],
            })))
            for (let i = 0; i < returns.length; i++) {
                for (let j = 0; j < items.length; j++) {
                    if (!returns[i]) break
                    if (items[j].product_id !== returns[i].items[0].product_id) continue
                    if (items[j].packing_id !== returns[i].items[0].packing_id) continue
                    result.push({ ...items[j], price: returns[i].items[0].price, amount: returns[i].items[0].amount })
                }
            }
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Пошук товарів в історії повернень', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

}