import { v4 as uuid } from 'uuid'
import { Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import Rest from '../../../lib/rest.js'
import WriteoffHelper from '../../../helpers/WriteoffHelper.js'

export default class WriteoffDocuments {

    static prepare(data, id = uuid()) {
        const { subdivision_id, number, name, warehouse_id, createdAt } = data
        const payload = {
            id,
            subdivision_id,
            name,
            number,
            warehouse_id
        }
        if (createdAt) payload['createdAt'] = createdAt
        return payload
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.Warehouses, as: 'warehouse', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                {
                    model: Database.Documents.WriteoffGoods, as: 'items', attributes: {
                        exclude: ['subdivision_id', 'document_id', 'series_id', 'product_id', 'packing_id', 'deletion_mark', 'position', 'updatedAt']
                    },
                    include: [
                        { model: Database.Products, attributes: ['id', 'name'], as: 'product' },
                        { model: Database.Series, attributes: ['id', 'name'], as: 'series' },
                        { model: Database.ProductPacking, attributes: ['id', 'name'], as: 'packing' }
                    ],
                    required: false,
                    where: { 'deletion_mark': { [Op.or]: { [Op.is]: null, [Op.eq]: false } } }
                }
            ],
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.WriteoffGoods, as: 'items' }, 'position', 'ASC']
            ],
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Writeoff.count({ where, transaction }))
            const result = await Database.Documents.Writeoff.findAll(WriteoffDocuments.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All writeoff documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Documents.Writeoff.findOne(WriteoffDocuments.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get writeoff documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { warehouse_id, items } = req.body
            const number = req.body.number ?? (((await Database.Documents.Writeoff.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)
            const document = await Database.Documents.Writeoff.create(WriteoffDocuments.prepare({ ...req.body, number, subdivision_id }), { transaction })
            const document_id = document.getDataValue('id')
            const period = document.getDataValue('createdAt')
            if (items) {
                await Database.Documents.WriteoffGoods.bulkCreate(items.map((item, position) => ({ ...item, document_id, subdivision_id, position })), { transaction })
                await WriteoffHelper.ApplyProductsStockMoves({ basis_id: document_id, warehouse_id, period, items }, transaction)
            }
            const result = await Database.Documents.Writeoff.findOne(WriteoffDocuments.options({ id: document_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update writeoff documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const { warehouse_id, createdAt, items } = req.body
            const updatedAt = new Date().toISOString()
            const exists = await Database.Documents.Writeoff.findOne({ where: { id, ...(is_master ? {} : { subdivision_id }) }, transaction })
            if (!exists) throw Error('Обʼєкт для редагування не знайдено або недостатньо прав для його модифікації')
            exists.update(WriteoffDocuments.prepare(req.body, id), { where: { id }, transaction })
            const document_subdivision_id = exists.getDataValue('subdivision_id')
            if (items) {
                // помічаємо попередні позиції на вилучення
                await Database.Documents.WriteoffGoods.update({ deletion_mark: true }, { where: { document_id: id }, transaction })
                for (let position = 0; position < items.length; position++) {
                    // оновлюємо нові позиції і знімаємо мітку про вилучення
                    await Database.Documents.WriteoffGoods.upsert({ ...items[position], document_id: id, subdivision_id: document_subdivision_id, position, deletion_mark: false, updatedAt }, { transaction })
                }
                // записуємо рухи документа в регістр залишків
                await WriteoffHelper.ApplyProductsStockMoves({ basis_id: id, warehouse_id, period: createdAt, items }, transaction)
            }
            const result = await Database.Documents.Writeoff.findOne(WriteoffDocuments.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update writeoff documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }



}