
import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { col, fn, literal, cast, Op, where } from '@sequelize/core'
import Database from '../../../database/index.js'
import Round from '../../../lib/Round.js'

const HOUR_PERIOD_FORMAT = 'HH'
const DAY_PERIOD_FORMAT = 'DD MMM'

export default class DashboardReport {

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { period } = req.query
            const isDayRange = period === 'day'

            const range = { after: moment().startOf('day'), before: moment().endOf('day').add(1, 'hour').startOf('hour'), }
            switch (period) {
                case 'day':
                    range.after = moment().startOf('day').add(1, 'hour').startOf('hour')
                    range.before = moment().endOf('day').add(1, 'hour').startOf('hour')
                    break;
                case 'month':
                    range.after = moment().subtract(1, 'month').startOf('day')
                    break;
                case 'week':
                    range.after = moment().subtract(1, 'week').startOf('day')
                    break;
            }
            const prev_range = { after: moment(range.after).subtract(1, period), before: moment(range.before).subtract(1, period), }
            const result = { id: 'TOTAL', title: 'Разом', color: '#FF8888', data: [] }
            const result_cash = { id: 'CASH', title: 'Готівкою', color: '#43CCB4', data: [] }
            const result_card = { id: 'CARD', title: 'Картою', color: '#48BDD4', data: [] }

            let step = null
            const target = moment(range.after).format(isDayRange ? HOUR_PERIOD_FORMAT : DAY_PERIOD_FORMAT)
            do {
                step = moment(range.before).subtract(result.data.length, isDayRange ? 'hour' : 'day').format(isDayRange ? HOUR_PERIOD_FORMAT : DAY_PERIOD_FORMAT)
                result.data.unshift({ period: step, value: 0 })
                result_cash.data.unshift({ period: step, value: 0 })
                result_card.data.unshift({ period: step, value: 0 })
            } while (step !== target)

            const payments = await Database.Documents.ReceiptPayment.findAll({
                attributes: ['amount', [col('method.type'), 'type'], ['createdAt', 'period']],
                include: [{ model: Database.PaymentMethods, as: 'method', attributes: [] }],
                where: { status: "PAYED", createdAt: { [Op.between]: [range.after, range.before] }, ...(is_master ? {} : { subdivision_id }) },
                order: [['createdAt', 'ASC']],
                transaction,
            })

            payments.map(item => item.get()).map((item) => {
                const date = moment(item.period).format(DAY_PERIOD_FORMAT)
                const time = moment(item.period).add(1, 'hour').startOf('hour').format(HOUR_PERIOD_FORMAT)
                const exist = result.data.find(it => it.period === (isDayRange ? time : date))
                if (exist) {
                    exist.value += item.amount
                } else {
                    result.data.push({ period: isDayRange ? time : date, value: item.amount })
                }
                switch (item.type) {
                    case 'CASH':
                        const exist_cash = result_cash.data.find(it => it.period === (isDayRange ? time : date))
                        if (exist_cash) {
                            exist_cash.value += item.amount
                        } else {
                            result_cash.data.push({ period: isDayRange ? time : date, value: item.amount })
                        }
                        break;
                    case 'CARD':
                        const exist_card = result_card.data.find(it => it.period === (isDayRange ? time : date))
                        if (exist_card) {
                            exist_card.value += item.amount
                        } else {
                            result_card.data.push({ period: isDayRange ? time : date, value: item.amount })
                        }
                        break;
                }
            })
            result.data.map(it => { it.value = Round.money(it.value); if (isDayRange) it.period = `${it.period}:00` })
            result_cash.data.map(it => { it.value = Round.money(it.value); if (isDayRange) it.period = `${it.period}:00` })
            result_card.data.map(it => { it.value = Round.money(it.value); if (isDayRange) it.period = `${it.period}:00` })
            const total_sales = Round.money(result.data.map(it => it.value).reduce((a, b) => a + b))
            const prev_total_sales = await Database.Documents.ReceiptPayment.sum('amount', { where: { status: "PAYED", createdAt: { [Op.between]: [prev_range.after, prev_range.before] }, ...(is_master ? {} : { subdivision_id }) }, transaction })
            const total_sales_changes = Round.money(((total_sales - prev_total_sales) / prev_total_sales) * 100)

            const prev_total_sell_receipts = await Database.Documents.Receipt.count({ where: { type: "SELL", status: "CLOSED", createdAt: { [Op.between]: [prev_range.after, prev_range.before] }, ...(is_master ? {} : { subdivision_id }) }, transaction })
            const total_sell_receipts = await Database.Documents.Receipt.count({ where: { type: "SELL", status: "CLOSED", createdAt: { [Op.between]: [range.after, range.before] }, ...(is_master ? {} : { subdivision_id }) }, transaction })
            const total_sell_receipts_changes = Round.money(((total_sell_receipts - prev_total_sell_receipts) / prev_total_sell_receipts) * 100)

            const prev_total_return_receipts = await Database.Documents.Receipt.count({ where: { type: "RETURN", status: "CLOSED", createdAt: { [Op.between]: [prev_range.after, prev_range.before] }, ...(is_master ? {} : { subdivision_id }) }, transaction })
            const total_return_receipts = await Database.Documents.Receipt.count({ where: { type: "RETURN", status: "CLOSED", createdAt: { [Op.between]: [range.after, range.before] }, ...(is_master ? {} : { subdivision_id }) }, transaction })
            const total_return_receipts_changes = Round.money(((total_return_receipts - prev_total_return_receipts) / prev_total_return_receipts) * 100)

            const all_categories = await Database.Documents.ReceiptGoods.findAll({
                attributes: [
                    [col('product.category.id'), 'id'],
                    [col('product.category.name'), 'label'],
                    [col('product.category.color'), 'color'],
                    [fn('sum', literal('`amount`*`price`')), 'value'],
                    [fn('count', 'id'), 'receipts'],
                ],
                include: [
                    {
                        model: Database.Documents.Receipt, as: 'document', attributes: [], required: true, where: {
                            type: "SELL",
                            status: "CLOSED",
                            createdAt: { [Op.between]: [range.after, range.before] },
                            ...(is_master ? {} : { subdivision_id })
                        }
                    },
                    {
                        model: Database.Products, as: 'product', attributes: [], include: [
                            { model: Database.ProductCategories, as: 'category', attributes: [] }
                        ]
                    }
                ],
                group: ['id'],
                transaction
            })
            const top_categories = all_categories.map(it => it.get()).sort((a, b) => a.value < b.value ? 1 : -1).slice(0, 5).map(category => {
                return { ...category, name: category.name || 'Без категорії', value: Round.money(category.value) }
            })

            const all_products = await Database.Documents.ReceiptGoods.findAll({
                attributes: [
                    [col('product.id'), 'id'],
                    [col('product.name'), 'label'],
                    [fn('sum', literal('`amount`*`price`')), 'value'],
                    [fn('count', 'id'), 'receipts'],
                ],
                include: [
                    {
                        model: Database.Documents.Receipt, as: 'document', attributes: [], required: true, where: {
                            type: "SELL",
                            status: "CLOSED",
                            createdAt: { [Op.between]: [range.after, range.before] },
                            ...(is_master ? {} : { subdivision_id })
                        }
                    },
                    { model: Database.Products, as: 'product', attributes: [], include: [] }
                ],
                group: ['id'],
                transaction
            })

            const top_products = all_products.map(it => it.get()).sort((a, b) => a.value < b.value ? 1 : -1).slice(0, 8).map(product => {
                return { ...product, value: Round.money(product.value) }
            })


            const prev_avg_receipt = await Database.Documents.ReceiptGoods.findOne({
                attributes: [
                    [fn('sum', literal('`amount`*`price`')), 'value'],
                    [fn('count', col('document.id')), 'receipts'],
                ],
                include: [
                    {
                        model: Database.Documents.Receipt, as: 'document', attributes: [], required: true, where: {
                            type: "SELL",
                            status: "CLOSED",
                            createdAt: { [Op.between]: [prev_range.after, prev_range.before] },
                            ...(is_master ? {} : { subdivision_id })
                        }
                    }
                ],
                transaction
            }).then(row => row ? row.get() : null)

            const avg_receipt = await Database.Documents.ReceiptGoods.findOne({
                attributes: [
                    [fn('sum', literal('`amount`*`price`')), 'value'],
                    [fn('count', col('document.id')), 'receipts'],
                ],
                include: [
                    {
                        model: Database.Documents.Receipt, as: 'document', attributes: [], required: true, where: {
                            type: "SELL",
                            status: "CLOSED",
                            createdAt: { [Op.between]: [range.after, range.before] },
                            ...(is_master ? {} : { subdivision_id })
                        }
                    }
                ],
                transaction
            }).then(row => row ? row.get() : null)
            const prev_avg_receipt_value = prev_avg_receipt ? Round.money(prev_avg_receipt.value / prev_avg_receipt.receipts) : 0
            const avg_receipt_value = avg_receipt ? Round.money(avg_receipt.value / avg_receipt.receipts) : 0
            const avg_receipt_changes = prev_avg_receipt_value ? Round.money(((avg_receipt_value - prev_avg_receipt_value) / prev_avg_receipt_value) * 100) : 0


            await transaction.commit()
            res.status(200).send(JSON.stringify({
                total_sales, total_sales_changes,
                total_sell_receipts, total_sell_receipts_changes,
                total_return_receipts, total_return_receipts_changes,
                sales: [result, result_cash, result_card],
                categories: top_categories,
                products: top_products,
                avg_receipt_value,
                avg_receipt_changes,
            }))


        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get money report', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}