
import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { col, fn, and, literal, cast, Op, where } from '@sequelize/core'
import Database from '../../../database/index.js'
import Round from '../../../lib/Round.js'

function compare(item, subdivision_id, workplace_id, type, period) {
    return item.subdivision_id === subdivision_id && item.workplace_id === workplace_id && item.type === type && item.period === period
}

const types = ["CARD", "CASH"]
const type_titles = {
    "CASH": "Готівкою",
    "CARD": "Картою"
}
const operations_by_type = {
    "CASH": ["CASH_IN_AUTO", "CASH_IN", "SELL", "RETURN", "CASH_OUT", "CASH_OUT_AUTO"],
    "CARD": ["SELL", "RETURN"]
}
const operations_titles = {
    "CASH_IN_AUTO": "Автоматичне внесення готівки",
    "CASH_IN": "Внесення готівки",
    "SELL": "Продаж",
    "RETURN": "Повернення",
    "CASH_OUT": "Вилучення готівки",
    "CASH_OUT_AUTO": "Автоматичне вилучення готівки"
}

export default class MoneyReport {

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, type, period_start, period_end } = req.query


            if (!period_start || !moment(period_start).isValid()) throw Error("Некоректний початок періоду звіту")
            if (!period_end || !moment(period_end).isValid()) throw Error("Некоректний кінець періоду звіту")
            if (moment(period_start).isAfter(moment(period_end))) throw Error("Некоректний період звіту")

            const subdivisions = await Database.Subdivisions.findAll({
                attributes: ['id', 'name'],
                where: subdivision_id && !workplace_id ? { id: subdivision_id } : {},
                order: [['name', 'ASC']],
                transaction
            })

            const workplaces = await Database.Workplaces.findAll({
                attributes: ['id', 'name', 'subdivision_id'],
                where: workplace_id ? { id: workplace_id } : subdivision_id ? { subdivision_id } : {},
                order: [['name', 'ASC']],
                transaction
            })


            const on_start_period = await Database.Registers.Money.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'type', [fn('sum', col('amount')), 'amount']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    period: { [Op.lt]: moment(period_start).toDate() }
                },
                group: ['subdivision_id', 'workplace_id', 'type'],
                transaction
            })
            const income_on_period = await Database.Registers.Money.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'type', [fn('sum', col('amount')), 'amount'], [cast(col('period'), 'DATE'), 'period']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    period: { [Op.lt]: moment(period_end).toDate(), [Op.gt]: moment(period_start).toDate() },
                    amount: { [Op.gt]: 0 }
                },
                group: ['subdivision_id', 'workplace_id', 'type', cast(col('period'), 'DATE')],
                transaction
            })
            const outcome_on_period = await Database.Registers.Money.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'type', [fn('sum', col('amount')), 'amount'], [cast(col('period'), 'DATE'), 'period']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    period: { [Op.lt]: moment(period_end).toDate(), [Op.gt]: moment(period_start).toDate() },
                    amount: { [Op.lt]: 0 }
                },
                group: ['subdivision_id', 'workplace_id', 'type', cast(col('period'), 'DATE')],
                transaction
            })

            const on_end_period = await Database.Registers.Money.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'type', [fn('sum', col('amount')), 'amount']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    period: { [Op.lt]: moment(period_end).toDate() }
                },
                group: ['subdivision_id', 'workplace_id', 'type'],
                transaction
            })

            const cash_period_moves = await Database.Registers.Money.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'operation', 'type', [fn('sum', col('amount')), 'amount'], [cast(col('period'), 'DATE'), 'period']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    period: { [Op.lt]: moment(period_end).toDate(), [Op.gt]: moment(period_start).toDate() },
                    type: 'CASH'
                },
                group: ['subdivision_id', 'workplace_id', 'operation', 'type', cast(col('period'), 'DATE')],
                transaction
            })

            const periods = new Array(moment(period_end).local().diff(moment(period_start), 'day') + 1).fill(0).map((it, index) => moment(period_start).local().add(index, 'day').format('YYYY-MM-DD'))

            const group_by_subdivision = [];
            subdivisions.filter(it => !subdivision_id || subdivision_id === it.id).map((subdivision) => {
                // Групування по підрозділах
                const payload = { id: subdivision.id, name: subdivision.name }
                const group_by_workplace = [];
                workplaces.filter(it => it.subdivision_id == subdivision.id).map((workplace) => {
                    // Групування по касах
                    const payload = { id: workplace.id, name: workplace.name }
                    const group_by_period = [];
                    periods.map((period, index) => {
                        // Групування по днях

                        const has_moves = income_on_period.find(it => it.period === period) || outcome_on_period.find(it => it.period === period)
                        if (!has_moves) return

                        const isFirst = index === 0
                        const isLast = index === (periods.length - 1)
                        const payload = { name: moment(period).format('DD.MM.YYYY'), period, expanded: false }
                        const group_by_types = []
                        types.filter(it => !type || it === type).map(type => {
                            // Групування типу оплати 
                            const prev = !isFirst ? group_by_period[group_by_period.length - 1]?.items?.find(it => it.type === type) : null
                            const on_start = isFirst ? (on_start_period.find(it => compare(it, subdivision.id, workplace.id, type))?.amount || 0.0) : (prev?.on_end || 0)
                            const income = income_on_period.find(it => compare(it, subdivision.id, workplace.id, type, period))?.amount || 0.0
                            const outcome = -outcome_on_period.find(it => compare(it, subdivision.id, workplace.id, type, period))?.amount || 0.0
                            const on_end = isLast ? (on_end_period.find(it => compare(it, subdivision.id, workplace.id, type))?.amount || 0.0) : (on_start + income - outcome)

                            group_by_types.push({
                                name: type_titles[type],
                                type,
                                expanded: false,
                                on_start: Round.money(on_start),
                                income: Round.money(income),
                                outcome: Round.money(outcome),
                                on_end: Round.money(on_end),
                                items: operations_by_type[type]?.map(operation => {
                                    const record = cash_period_moves.find(it => it.operation === operation && it.period === period)
                                    switch (operation) {
                                        case "CASH_IN_AUTO":
                                        case "CASH_IN":
                                        case "SELL":
                                            return {
                                                name: operations_titles[operation],
                                                operation,
                                                income: record?.amount || 0
                                            }
                                        case "CASH_OUT_AUTO":
                                        case "CASH_OUT":
                                        case "RETURN":
                                            return {
                                                name: operations_titles[operation],
                                                operation,
                                                outcome: record?.amount || 0
                                            }
                                    }
                                })
                            })
                        })
                        payload['on_start'] = group_by_types.length ? group_by_types.map(it => it.on_start).reduce((prev, next) => prev + next) : 0
                        payload['income'] = group_by_types.length ? group_by_types.map(it => it.income).reduce((prev, next) => prev + next) : 0
                        payload['outcome'] = group_by_types.length ? group_by_types.map(it => it.outcome).reduce((prev, next) => prev + next) : 0
                        payload['on_end'] = group_by_types.length ? group_by_types.map(it => it.on_end).reduce((prev, next) => prev + next) : 0
                        group_by_period.push({ ...payload, items: group_by_types.length === 1 ? group_by_types[0].items : group_by_types, })
                    })
                    payload['on_start'] = group_by_period.length ? group_by_period.map(it => it.on_start).reduce((prev, next) => prev + next) : 0
                    payload['income'] = group_by_period.length ? group_by_period.map(it => it.income).reduce((prev, next) => prev + next) : 0
                    payload['outcome'] = group_by_period.length ? group_by_period.map(it => it.outcome).reduce((prev, next) => prev + next) : 0
                    payload['on_end'] = group_by_period.length ? group_by_period.map(it => it.on_end).reduce((prev, next) => prev + next) : 0
                    group_by_workplace.push({ ...payload, items: group_by_period.length === 1 ? group_by_period[0].items : group_by_period })
                })
                payload['on_start'] = group_by_workplace.length ? group_by_workplace.map(it => it.on_start).reduce((prev, next) => prev + next) : 0
                payload['income'] = group_by_workplace.length ? group_by_workplace.map(it => it.income).reduce((prev, next) => prev + next) : 0
                payload['outcome'] = group_by_workplace.length ? group_by_workplace.map(it => it.outcome).reduce((prev, next) => prev + next) : 0
                payload['on_end'] = group_by_workplace.length ? group_by_workplace.map(it => it.on_end).reduce((prev, next) => prev + next) : 0
                group_by_subdivision.push({ ...payload, items: group_by_workplace.length === 1 ? group_by_workplace[0].items : group_by_workplace })
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(group_by_subdivision.filter(it => !!it.items.length)))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get money report', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}