
import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { col, fn, and, literal, cast, Op, where } from '@sequelize/core'
import Database from '../../../database/index.js'

export default class SalesReport {

    static groupByPeriod(moves, { periods }, options) {
        const result = []
        for (let i = 0; i < periods.length; i++) {
            const period = periods[i]
            const current_moves = moves.filter(it => moment(it.period).isBetween(moment(period).startOf('day'), moment(period).endOf('day')))
            if (!current_moves.length) continue
            const current_sales = current_moves.filter(it => it.operation === "SELL")
            const current_returns = current_moves.filter(it => it.operation === "RETURN")
            const amount_moves = [...current_sales.map(it => it.amount), ...current_returns.map(it => -it.amount)]
            const amount = amount_moves.length ? amount_moves.reduce((prev, next) => prev + next) : 0
            const total_moves = [...current_sales.map(it => it.total), ...current_returns.map(it => -it.total)]
            const total = total_moves.length ? total_moves.reduce((prev, next) => prev + next) : 0
            result.push({
                name: moment(period).format('DD.MM.YYYY'),
                period,
                amount,
                total
            })
        }
        return result
    }

    static groupByProduct(moves, { products, periods }, options) {
        const result = []
        for (let i = 0; i < products.length; i++) {
            const product = products[i]
            if (!product) continue
            const current_moves = moves.filter(it => it.product_id === product?.id)
            const payload = { id: product.id, name: product.name, expanded: false }
            const items = SalesReport.groupByPeriod(current_moves, { periods }, options)
            if (!items.length) continue
            payload['amount'] = items.length ? items.map(it => it.amount).reduce((prev, next) => prev + next) : 0
            payload['total'] = items.length ? items.map(it => it.total).reduce((prev, next) => prev + next) : 0
            result.push({ ...payload, items })
        }
        return result
    }

    static groupByCategory(moves, { categories, products, periods }, options) {
        if (!options.group_by_category) {
            const products_ids = moves.map(it => it.product_id).filter((item, index, array) => array.indexOf(item) === index)
            const products_list = products.filter(it => products_ids.includes(it.id))
            return SalesReport.groupByProduct(moves, { products: products_list, periods }, options)
        } else {
            const result = []
            for (let i = 0; i < categories.length; i++) {
                const category = categories[i]
                const payload = { id: category?.id, name: category?.name || 'Без категорії' }
                const current_moves = moves.filter(it => it.category_id === category?.id)
                const products_ids = current_moves.map(it => it.product_id).filter((item, index, array) => array.indexOf(item) === index)
                const products_list = products.filter(it => products_ids.includes(it.id))
                const items = SalesReport.groupByProduct(current_moves, { products: products_list, periods }, options)
                if (!items.length) continue
                payload['total'] = items.length ? items.map(it => it.total).reduce((prev, next) => prev + next) : 0
                result.push({ ...payload, items })
            }
            return result
        }
    }

    static groupByStaff(moves, { staffs, categories, products, periods }, options) {
        if (!options.group_by_staff) {
            const categories_ids = moves.map(it => it.category_id).filter((item, index, array) => array.indexOf(item) === index)
            const categories_list = categories.filter(it => categories_ids.includes(it.id))
            return SalesReport.groupByCategory(moves, { categories: categories_list, products, periods }, options)
        } else {
            const result = []
            for (let i = 0; i < staffs.length; i++) {
                const staff = staffs[i]
                const payload = { id: staff?.id, name: `${staff.firstname} ${staff.lastname}` }
                const current_moves = moves.filter(it => it.staff_id === staff?.id)
                const categories_ids = current_moves.map(it => it.category_id).filter((item, index, array) => array.indexOf(item) === index)
                const categories_list = categories.filter(it => categories_ids.includes(it.id))
                const items = SalesReport.groupByCategory(current_moves, { categories: categories_list, products, periods }, options)
                if (!items.length) continue
                payload['total'] = items.length ? items.map(it => it.total).reduce((prev, next) => prev + next) : 0
                result.push({ ...payload, items })
            }
            return result
        }
    }

    static groupByWorkplace(moves, { workplaces, staffs, categories, products, periods }, options) {
        if (!options.group_by_workplace) {
            const staffs_ids = moves.map(it => it.staff_id).filter((item, index, array) => array.indexOf(item) === index)
            const staffs_list = staffs.filter(it => staffs_ids.includes(it.id))
            return SalesReport.groupByStaff(moves, { staffs: staffs_list, categories, products, periods }, options)
        } else {
            const result = []
            for (let i = 0; i < workplaces.length; i++) {
                const workplace = workplaces[i]
                const payload = { id: workplace?.id, name: workplace?.name }
                const current_moves = moves.filter(it => it.workplace_id === workplace?.id)
                const staffs_ids = current_moves.map(it => it.staff_id).filter((item, index, array) => array.indexOf(item) === index)
                const staffs_list = staffs.filter(it => staffs_ids.includes(it.id))
                const items = SalesReport.groupByStaff(current_moves, { staffs: staffs_list, categories, products, periods }, options)
                if (!items.length) continue
                payload['total'] = items.length ? items.map(it => it.total).reduce((prev, next) => prev + next) : 0
                result.push({ ...payload, items })
            }
            return result
        }
    }


    static groupBySubdivision(moves, { subdivisions, workplaces, staffs, categories, products, periods }, options) {
        if (!options.group_by_subdivision) {
            const workplaces_ids = moves.map(it => it.staff_id).filter((item, index, array) => array.indexOf(item) === index)
            const workplaces_list = workplaces.filter(it => workplaces_ids.includes(it.id))
            return SalesReport.groupByWorkplace(moves, { workplaces: workplaces_list, staffs, categories, products, periods }, options)
        } else {
            const result = []
            for (let i = 0; i < subdivisions.length; i++) {
                const subdivision = subdivisions[i]
                const payload = { id: subdivision?.id, name: subdivision?.name }
                const current_moves = moves.filter(it => it.subdivision_id === subdivision?.id)
                const workplaces_ids = current_moves.map(it => it.workplace_id).filter((item, index, array) => array.indexOf(item) === index)
                const workplaces_list = workplaces.filter(it => workplaces_ids.includes(it.id))
                const items = SalesReport.groupByWorkplace(current_moves, { workplaces: workplaces_list, staffs, categories, products, periods }, options)
                if (!items.length) continue
                payload['total'] = items.length ? items.map(it => it.total).reduce((prev, next) => prev + next) : 0
                result.push({ ...payload, items })
            }
            return result
        }
    }


    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const {
                period_start, period_end,
                subdivision_id, group_by_subdivision,
                workplace_id, group_by_workplace,
                staff_id, group_by_staff,
                category_id, group_by_category,
                product_id
            } = req.query


            if (!period_start || !moment(period_start).isValid()) throw Error("Некоректний початок періоду звіту")
            if (!period_end || !moment(period_end).isValid()) throw Error("Некоректний кінець періоду звіту")
            if (moment(period_start).isAfter(moment(period_end))) throw Error("Некоректний період звіту")


            const register_moves = await Database.Registers.Sales.findAll({
                attributes: ['subdivision_id', 'workplace_id', 'staff_id', 'category_id', 'product_id', 'operation', [fn('sum', col('amount')), 'amount'], [fn('sum', col('total')), 'total'], 'period', [cast(col('period'), 'DATE'), 'date']],
                where: {
                    ...(subdivision_id && !workplace_id ? { subdivision_id } : {}),
                    ...(workplace_id ? { workplace_id } : {}),
                    ...(staff_id ? { staff_id } : {}),
                    ...(category_id ? { category_id } : {}),
                    ...(product_id ? { product_id } : {}),
                    period: { [Op.lt]: moment(period_end).toDate(), [Op.gt]: moment(period_start).toDate() },
                },
                group: ['subdivision_id', 'workplace_id', 'staff_id', 'category_id', 'product_id', 'operation', 'date'],
                transaction
            })

            const subdivisions = await Database.Subdivisions.findAll({
                attributes: ['id', 'name'],
                where: { id: { [Op.in]: register_moves.map(it => it.subdivision_id).filter((item, index, array) => !!item && array.indexOf(item) === index) } },
                order: [['name', 'ASC']],
                transaction
            })


            const workplaces = await Database.Workplaces.findAll({
                attributes: ['id', 'name', 'subdivision_id'],
                where: { id: { [Op.in]: register_moves.map(it => it.workplace_id).filter((item, index, array) => !!item && array.indexOf(item) === index) } },
                order: [['name', 'ASC']],
                transaction
            })

            const staffs = await Database.Staff.findAll({
                where: { id: { [Op.in]: register_moves.map(it => it.staff_id).filter((item, index, array) => !!item && array.indexOf(item) === index) } },
                transaction
            })

            const categories = await Database.ProductCategories.findAll({
                attributes: ['id', 'name'],
                where: { id: { [Op.in]: register_moves.map(it => it.category_id).filter((item, index, array) => !!item && array.indexOf(item) === index) } }
            })

            const products = await Database.Products.findAll({
                attributes: ['id', 'name'],
                where: { id: { [Op.in]: register_moves.map(it => it.product_id).filter((item, index, array) => !!item && array.indexOf(item) === index) } }
            })


            const periods = new Array(moment(period_end).local().diff(moment(period_start), 'day') + 1).fill(0).map((it, index) => moment(period_start).local().add(index, 'day').format('YYYY-MM-DD'))

            let result = SalesReport.groupBySubdivision(register_moves, { subdivisions, workplaces, staffs, categories, products, periods }, {
                group_by_subdivision: group_by_subdivision && group_by_subdivision !== '0',
                group_by_workplace: group_by_workplace && group_by_workplace !== '0',
                group_by_staff: group_by_staff && group_by_staff !== '0',
                group_by_category: group_by_category && group_by_category !== '0',
            })

            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get money report', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}