
import { v4 as uuid } from 'uuid'
import moment from 'moment'
import { col, fn, Op } from '@sequelize/core'
import Database from '../../../database/index.js'
import ProductStockRegisterValuesHelper from '../../../database/schemes/registers/product/product_stock_values_helper.js'
import ProductPriceRegisterValuesHelper from '../../../database/schemes/registers/product/product_price_values_helper.js'
import Round from '../../../lib/Round.js'

export default class StockReport {

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id'] },
            include: [
                { model: Database.Warehouses, as: 'warehouse', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                { model: Database.PriceTypes, as: 'price_type', attributes: { exclude: ['createdAt', 'updatedAt'] } },
                {
                    model: Database.Documents.ArrivalGoods, as: 'items', attributes: {
                        exclude: ['subdivision_id', 'document_id', 'series_id', 'product_id', 'packing_id', 'deletion_mark', 'position', 'updatedAt']
                    },
                    include: [
                        { model: Database.Products, attributes: ['id', 'name'], as: 'product' },
                        { model: Database.Series, attributes: ['id', 'name'], as: 'series' },
                        { model: Database.ProductPacking, attributes: ['id', 'name'], as: 'packing' }
                    ],
                    required: false,
                    where: { 'deletion_mark': { [Op.or]: { [Op.is]: null, [Op.eq]: false } } }
                }
            ],
            order: [
                ['createdAt', 'DESC'],
                [{ model: Database.Documents.ArrivalGoods, as: 'items' }, 'position', 'ASC']
            ],
            where,
            ...params
        }
    }


    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const {
                subdivision_id,
                warehouse_id,
                price_type_id,
                category_id,
                product_id,
                period_start, period_end
            } = req.query

            const group_by_subdivision = req.query.group_by_subdivision === '1'
            const group_by_warehouse = req.query.group_by_warehouse === '1'


            if (!period_start || !moment(period_start).isValid()) throw Error("Некоректний початок періоду звіту")
            if (!period_end || !moment(period_end).isValid()) throw Error("Некоректний кінець періоду звіту")
            if (moment(period_start).isAfter(moment(period_end))) throw Error("Некоректний період звіту")

            const use_products_filter = !!product_id || !!category_id
            const use_warehouse_filter = !!warehouse_id

            const subdivisions = await Database.Subdivisions.findAll({
                attributes: ['id', 'name'],
                where: subdivision_id && !warehouse_id ? { id: subdivision_id } : {},
                order: [['name', 'ASC']],
                transaction
            })
            const warehouses = await Database.Warehouses.findAll({
                attributes: ['id', 'subdivision_id', 'name'],
                where: warehouse_id ? { id: warehouse_id } : subdivision_id ? { subdivision_id } : {},
                order: [['name', 'ASC']],
                transaction
            })
            const products = await Database.Products.findAll({
                attributes: ['id', 'name', 'packing_id'],
                where: {
                    ...(product_id ? { id: product_id } : {})
                },
                include: [
                    ...(category_id ? [{ model: Database.ProductCategories, as: 'category', attributes: [], where: { id: category_id } }] : [])
                ],
                transaction,
                raw: true
            })
            const product_ids = products.map(it => it.id)
            const packing_ids = products.map(it => it.packing_id)

            const packings = await Database.ProductPacking.findAll({
                attributes: ['id', 'name', 'product_id', 'coefficient'],
                where: { [Op.or]: { id: { [Op.in]: packing_ids }, product_id: { [Op.in]: product_ids } } },
                order: [['coefficient', 'ASC']],
                transaction
            })

            const prices = !price_type_id ? [] : await ProductPriceRegisterValuesHelper.getAll({
                where: {
                    product_id: { [Op.in]: product_ids },
                    packing_id: { [Op.in]: packing_ids },
                    price_type_id
                },
                transaction
            })

            const stock_on_period_start = await ProductStockRegisterValuesHelper.getAll({
                where: {
                    ...(use_products_filter ? { product_id: { [Op.in]: product_ids } } : {}),
                    ...(use_warehouse_filter ? { warehouse_id } : {}),
                    period: { [Op.lte]: moment(period_start).toDate() }
                },
                transaction
            })

            const income_on_period = await ProductStockRegisterValuesHelper.getAll({
                where: {
                    ...(use_products_filter ? { product_id: { [Op.in]: product_ids } } : {}),
                    ...(use_warehouse_filter ? { warehouse_id } : {}),
                    period: {
                        [Op.and]: {
                            [Op.gte]: moment(period_start).toDate(),
                            [Op.lte]: moment(period_end).toDate()
                        }
                    },
                    in_stock: { [Op.gt]: 0 }
                },
                transaction
            })

            const outcome_on_period = await ProductStockRegisterValuesHelper.getAll({
                where: {
                    ...(use_products_filter ? { product_id: { [Op.in]: product_ids } } : {}),
                    ...(use_warehouse_filter ? { warehouse_id } : {}),
                    period: {
                        [Op.and]: {
                            [Op.gte]: moment(period_start).toDate(),
                            [Op.lte]: moment(period_end).toDate()
                        }
                    },
                    in_stock: { [Op.lt]: 0 }
                },
                transaction
            })

            const stock_on_period_end = await ProductStockRegisterValuesHelper.getAll({
                where: {
                    ...(use_products_filter ? { product_id: { [Op.in]: product_ids } } : {}),
                    ...(use_warehouse_filter ? { warehouse_id } : {}),
                    period: { [Op.lte]: moment(period_end).toDate() }
                },
                transaction
            })

            let data = []

            stock_on_period_start.map(item => {
                const existed = data.find(it => it.product_id === item.product_id && it.warehouse_id == item.warehouse_id)
                if (existed) {
                    existed['on_start'] = item.in_stock
                } else {
                    const { name } = products.find(it => it.id === item.product_id)
                    const { name: unit, id: packing_id, coefficient } = packings.find(it => it.product_id === item.product_id) || { name: null, coefficient: 1 }
                    data.push({ name, unit, product_id: item.product_id, packing_id, warehouse_id: item.warehouse_id, on_start: item.in_stock / coefficient })
                }
            })

            income_on_period.map(item => {
                const existed = data.find(it => it.product_id === item.product_id && it.warehouse_id == item.warehouse_id)
                if (existed) {
                    existed['income'] = item.in_stock
                } else {
                    const { name } = products.find(it => it.id === item.product_id)
                    const { name: unit, id: packing_id, coefficient } = packings.find(it => it.product_id === item.product_id) || { name: null, coefficient: 1 }
                    data.push({ name, unit, product_id: item.product_id, packing_id, warehouse_id: item.warehouse_id, income: item.in_stock / coefficient, })
                }
            })

            outcome_on_period.map(item => {
                const existed = data.find(it => it.product_id === item.product_id && it.warehouse_id == item.warehouse_id)
                if (existed) {
                    existed['outcome'] = -item.in_stock
                } else {
                    const { name } = products.find(it => it.id === item.product_id)
                    const { name: unit, id: packing_id, coefficient } = packings.find(it => it.product_id === item.product_id) || { name: null, coefficient: 1 }
                    data.push({ name, unit, product_id: item.product_id, packing_id, warehouse_id: item.warehouse_id, outcome: -item.in_stock / coefficient })
                }
            })

            stock_on_period_end.map(item => {
                const existed = data.find(it => it.product_id === item.product_id && it.warehouse_id == item.warehouse_id)
                if (existed) {
                    const { price } = prices.find(it => it.product_id === existed.product_id && it.packing_id === existed.packing_id) || { price: null }
                    existed['on_end'] = item.in_stock
                    existed['on_end_total'] = price ? Round.money(item.in_stock * price) : null
                } else {
                    const { name, id: product_id } = products.find(it => it.id === item.product_id)
                    const { name: unit, id: packing_id, coefficient } = packings.find(it => it.product_id === item.product_id) || { name: null, packing_id: null, coefficient: 1 }
                    const { price } = prices.find(it => it.product_id === product_id && it.packing_id === packing_id) || { price: null }
                    const on_end_total = price ? Round.money(item.in_stock / coefficient * price) : null
                    data.push({ name, unit, product_id: item.product_id, packing_id, warehouse_id: item.warehouse_id, on_end: item.in_stock / coefficient, on_end_total })
                }
            })

            data.map(item => {
                if (!item.on_start) item.on_start = 0.0
                if (!item.income) item.income = 0.0
                if (!item.outcome) item.outcome = 0.0
                if (!item.on_end) item.on_end = 0.0
            })
            data = data.sort((a, b) => a.name > b.name ? 1 : -1)

            var result = []

            if (group_by_subdivision) {
                subdivisions.map(subdivision_row => {
                    const subdivision = subdivision_row.get()
                    const subdivision_warehouses = warehouses.filter(it => it.subdivision_id === subdivision.id)
                    if (group_by_warehouse) {
                        subdivision['items'] = []
                        subdivision_warehouses.map(warehouse_row => {
                            const warehouse = warehouse_row.get()
                            warehouse['items'] = data.filter(it => it.warehouse_id === warehouse.id)
                            if (warehouse.items.length) {
                                warehouse['on_end_total'] = price_type_id ? Round.money(warehouse.items.map(it => it.on_end_total).reduce((a, b) => a + b)) : null
                                subdivision['items'].push(warehouse)
                            }
                        })
                    } else {
                        subdivision['items'] = []

                        data.filter(row => subdivision_warehouses.map(it => it.id).includes(row.warehouse_id)).map(row => {
                            let exists = subdivision['items'].find(it => it.product_id === row.product_id)
                            if (exists) {
                                exists.on_start += row.on_start
                                exists.income += row.income
                                exists.outcome += row.outcome
                                exists.on_end += row.on_end
                            } else {
                                subdivision['items'].push(row)
                            }
                        })
                    }
                    if (subdivision.items.length) {
                        subdivision['on_end_total'] = price_type_id ? Round.money(subdivision.items.map(it => it.on_end_total).reduce((a, b) => a + b)) : null
                        result.push(subdivision)
                    }
                })
            } else {
                data.map(row => {
                    let exists = result.find(it => it.product_id === row.product_id)
                    if (exists) {
                        exists.on_start += row.on_start
                        exists.income += row.income
                        exists.outcome += row.outcome
                        exists.on_end += row.on_end
                    } else {
                        result.push(row)
                    }
                })
            }

            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get arrival documents', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


}