import { v4 as uuid } from 'uuid'
import sequelize, { Op } from '@sequelize/core'
import moment from 'moment'
import Database from '../../database/index.js'
import Receipts from '../documents/receipts/helper.js'
import Rest from '../../lib/rest.js'
import FiscalRegistersFactory from '../../lib/FiscalRegistersFactory/index.js'
import SettingsHepler from '../../database/schemes/dictionaries/system/settings_helper.js'
import config from '../../../config/index.js'
import ReceiptGenerator from '../../lib/ReceiptGenerator.js'
import Round from '../../lib/Round.js'
import Replication from '../../replication/index.js'



export default class Shifts {

    static prepare(data, id) {
        const { subdivision_id, number, status, workplace_id, staff_id, register_id, fiscal_shift_id, created_at, updated_at } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            status: status || "CREATED",
            number,
            workplace_id,
            staff_id,
            register_id,
            fiscal_shift_id,
            ...(created_at ? { createdAt: created_at } : {}),
            ...(updated_at ? { updatedAt: updated_at } : {})
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id', 'workplace_id'] },
            include: [
                { model: Database.Workplaces, as: 'workplace', attributes: ['id', 'name', 'register_id', 'price_type_id', 'message'] },
                {
                    model: Database.Staff, as: 'staff', attributes: ['id', 'firstname', 'lastname'],
                    include: [{ model: Database.Roles, as: 'role', attributes: ['id', 'name'] }]
                },
                { model: Database.Documents.ShiftsTransactions, as: 'open_transaction', where: { type: "SHIFT_OPEN" }, required: false },
                { model: Database.Documents.ShiftsTransactions, as: 'close_transaction', where: { type: "SHIFT_CLOSE" }, required: false },
                { model: Database.Documents.ShiftsTransactions, as: 'transactions' },
            ],
            where,
            ...params,
            order: [
                ...(params.order && Array.isArray(params.order) ? params.order : [['createdAt', 'DESC']]),
                [Database.Documents.ShiftsTransactions, 'createdAt', 'ASC'],
            ],
        }
    }


    static async all(req, res) {
        const transaction = await Database.instance.transaction()
        res.setHeader('Content-Type', 'application/json')
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.Shifts.count({ where, transaction }))
            const shifts = await Database.Documents.Shifts.findAll(Shifts.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(shifts))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All shifts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const shift = await Database.Documents.Shifts.findOne(Shifts.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(shift))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get shifts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async open(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth

            const number = req.body.number ?? (((await Database.Documents.Shifts.max('number', { where: { subdivision_id }, transaction })) ?? 0) + 1)

            var current_shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (current_shift && (current_shift.status == "CREATED" || current_shift.status == "OPENING" || current_shift.status == "OPENED")) {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: "Уже існує відкрита зміна для цього терміналу" }))
            }
            if (current_shift && current_shift.status == "CLOSING") {
                await transaction.rollback()
                return res.status(400).send(JSON.stringify({ message: "Дочекайтесь закриття попередньої зміни" }))
            }
            const prev_shift_cash_amount = current_shift ? [...current_shift.transactions.filter(it => it.type === "CASH_OUT_AUTO").map(it => it.amount), 0, 0].reduce((prev, next) => prev + next) : 0
            const prev_shift_cash_amount_in_register = current_shift ? [...current_shift.transactions.filter(it => it.type === "CASH_OUT_AUTO" && !!it.fiscal_id).map(it => it.amount), 0, 0].reduce((prev, next) => prev + next) : 0


            const workplace = await Database.Workplaces.findOne({ where: { id: workplace_id, subdivision_id }, transaction })
            if (!workplace) throw new Database.Error("Не знайдено поточне РМК", 400)

            var shift = null
            if (workplace.register_id) {
                // Отримуємо ПРРО із колекції
                var rro = await FiscalRegistersFactory.getRegister(workplace.register_id)
                var error = null

                // Якщо ПРРО не доступний - намагаємось його ініціалізувати
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: workplace.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(workplace.register_id, register.get())
                }
                // Якщо ПРРО не доступний - повертаємо помилку
                if (!rro) throw new Database.Error("Немає доступу до обраного РРО", 400)
                // Відкриваємо зміну в ПРРО
                error = await rro.getShift(true)
                if (error) throw new Database.Error("Не вдалось отримати актуальну інформацію про касову зміну: \n" + error)
                // Створюємо в базі запис про відкриття зміни
                shift = await Database.Documents.Shifts.create(Shifts.prepare({ subdivision_id, number, status: rro.shift.status, workplace_id, staff_id, register_id: workplace.register_id, created_at: rro.shift.created_at, updated_at: rro.shift.updated_at }, rro.shift.id), { transaction })
                // Створюємо в базі запис транзакції відкриття зміни
                await Database.Documents.ShiftsTransactions.create({ subdivision_id, id: uuid(), shift_id: shift.id, type: "SHIFT_OPEN", date: Date(), amount: 0, comment: null, createdAt: rro.shift.opened_at || shift.createdAt }, { transaction })

                // Створюємо в базі ті внесення які уже є в ПРРО
                if (rro.shift.balance.service_in > 0) {
                    const receipts = await rro.getReceipts("SERVICE_IN")
                    if (receipts === null) throw Error('Не вдалось отримати інформацію про внесення готівки в ПРРО')
                    for (let i = 0; i < receipts.length; i++) {
                        const receipt = receipts[i]
                        const createdAt = new Date(receipt.created_at)
                        createdAt.setSeconds(createdAt.getSeconds() + 1)
                        await Database.Documents.ShiftsTransactions.upsert({ subdivision_id, id: receipt.id, shift_id: shift.id, type: "CASH_IN_AUTO", amount: receipt.total_sum / 100, fiscal_id: receipt.id, comment: null, createdAt }, { transaction })
                        await Database.Registers.Money.upsert({ basis_id: receipt.id, subdivision_id, workplace_id, operation: "CASH_IN_AUTO", type: "CASH", amount: receipt.total_sum / 100, period: createdAt, createdAt }, { transaction })
                    }
                }
                // Робимо ще внесення якщо в ПРРО внесена недостатня сума
                const prev_shift_cash_amount_in_register_aviable = prev_shift_cash_amount_in_register - (rro.shift.balance.service_in / 100)
                if (prev_shift_cash_amount_in_register_aviable > 0) {
                    const receipt = await rro.cashIn(prev_shift_cash_amount_in_register)
                    await Database.Documents.ShiftsTransactions.create({
                        subdivision_id,
                        id: receipt.id,
                        shift_id: shift.id,
                        type: "CASH_IN_AUTO",
                        amount: prev_shift_cash_amount_in_register,
                        fiscal_id: receipt.id,
                        fiscal_date: receipt.fiscal_date,
                        comment: null
                    }, { transaction })
                    await Database.Registers.Money.create({
                        basis_id: receipt.id,
                        subdivision_id,
                        workplace_id,
                        operation: "CASH_IN_AUTO",
                        type: "CASH",
                        amount: prev_shift_cash_amount_in_register
                    }, { transaction })
                }
                // Робимо внесення готівки поза межами ПРРО
                if (prev_shift_cash_amount > prev_shift_cash_amount_in_register) {
                    const available_cash = prev_shift_cash_amount - prev_shift_cash_amount_in_register
                    const transaction_id = uuid()
                    await Database.Documents.ShiftsTransactions.create({ subdivision_id, id: transaction_id, shift_id: shift.id, type: "CASH_IN_AUTO", amount: available_cash, comment: null }, { transaction })
                    await Database.Registers.Money.create({ basis_id: transaction_id, subdivision_id, workplace_id, operation: "CASH_IN_AUTO", type: "CASH", amount: available_cash }, { transaction })
                }
            } else {
                shift = await Database.Documents.Shifts.create(Shifts.prepare({ subdivision_id, number, status: "OPENED", workplace_id, staff_id, register_id: null }), { transaction })
                await Database.Documents.ShiftsTransactions.create({ subdivision_id, id: uuid(), shift_id: shift.id, type: "SHIFT_OPEN", date: Date(), amount: 0, comment: null }, { transaction })
                if (prev_shift_cash_amount > 0) {
                    const transaction_id = uuid()
                    await Database.Documents.ShiftsTransactions.create({ subdivision_id, id: transaction_id, shift_id: shift.id, type: "CASH_IN_AUTO", amount: prev_shift_cash_amount, comment: null }, { transaction })
                    await Database.Registers.Money.create({ basis_id: transaction_id, subdivision_id, workplace_id, operation: "CASH_IN_AUTO", type: "CASH", amount: prev_shift_cash_amount }, { transaction })
                }
            }
            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id: shift.id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Open shifts', subdivision_id: req.auth.subdivision_id })
            res.status(error.status || 500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async active(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id } = req.auth
            const { cashier_id } = req.params
            let error = null

            let shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift) {
                await transaction.rollback()
                return res.status(200).send(JSON.stringify(null))
            }
            if (shift.register_id && shift.register_id === shift.workplace.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id, is_enabled: true }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (rro) {
                    const current = await rro.getShiftById(shift.id)
                    if (current && current.status != shift.status) {
                        await Database.Documents.Shifts.update({ status: current.status }, { where: { id: shift.id, subdivision_id }, transaction })
                    }
                }
            }
            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id: shift.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Active shifts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async close(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { amount } = req.body

            let shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього робочого місця")
            if (shift.status === "CLOSED") {
                await transaction.rollback()
                return res.status(200).send(JSON.stringify(shift))
            }
            if (shift.status !== "OPENED") throw Error("Не знайдено відкритої зміни для цього робочого місця")


            const balance = await Shifts.prepareBalance(subdivision_id, workplace_id)

            const reports = []
            const reports_goods = []
            const receipts = await Database.Documents.Receipt.findAll({
                where: { shift_id: shift.id, subdivision_id },
                include: {
                    model: Database.Documents.ReceiptGoods, as: 'goods',
                },
                transaction
            })
            receipts.map((receipt) => {
                receipt.goods.map(item => {
                    var report = reports.find(it => it.warehouse_id == item.warehouse_id)
                    if (!report) {
                        const presentation = `Звіт про роздрібні продажі від ${moment().format('DD.MM.YYYY HH:mm:ss')}`
                        report = { subdivision_id, id: uuid(), presentation, status: 'NEW', workplace_id, staff_id, shift_id: shift.id, warehouse_id: item.warehouse_id }
                        reports.push(report)
                    }
                    var good = reports_goods.find(it =>
                        it.document_id == report.id &&
                        it.product_id == item.product_id &&
                        it.series_id == item.series_id &&
                        it.packing_id == item.packing_id
                    )
                    if (!good) {
                        good = {
                            subdivision_id,
                            document_id: report.id,
                            product_id: item.product_id,
                            series_id: item.series_id,
                            packing_id: item.packing_id,
                            tax_id: item.tax_id,
                            amount: item.amount,
                            price: item.price,
                        }
                        reports_goods.push(good)
                    } else {
                        good.amount += item.amount
                    }
                })
            })

            await Database.Documents.Report.bulkCreate(reports, { transaction })
            await Database.Documents.ReportGoods.bulkCreate(reports_goods, { transaction })

            var rro = null
            var error = null
            if (shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) {
                    await transaction.rollback()
                    return res.status(400).send(JSON.stringify({ message: "Немає доступу до обраного РРО" }))
                }

                const has_offline_receipts = !! await Database.Documents.Receipt.count({
                    where: {
                        shift_id: shift.id,
                        register_id: shift.register_id,
                        fiscal_sell_offline: true,
                        fiscal_send: false
                    }
                })

                if (has_offline_receipts) {
                    await transaction.rollback()
                    return res.status(400).send(JSON.stringify({ message: "В поточній зміні наявні офлайн чеки не надіслані в ДПС" }))
                }
            }

            const available_balance = balance.current_balance
            if (shift.register_id) {
                if (rro.isShiftOpened) {
                    // Оновлюємо дані про поточну зміну
                    error = await rro.getShift(false, true)
                    if (error) throw Error("Не вдалось отримати поточну зміну із фіскального реєстратора" + "\n" + error)
                    // Перевіряємо чи ми не намагаємось закрити зміну яка уже закрита у фіскальному реєстраторі
                    if (!rro.checkSyncOfShift(shift.id)) {
                        const old_shift = await rro.getShiftById(shift.id)
                        if (old_shift.status === 'CLOSED') {
                            await Database.Documents.ShiftsTransactions.update({ createdAt: old_shift.closed_at }, { where: { subdivision_id, shift_id: shift.id, type: "SHIFT_CLOSE" }, transaction })
                            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id: shift.id }, { transaction }))
                            await transaction.commit()
                            return res.status(200).send(JSON.stringify(result))
                        } else {
                            throw Error("В РРО і обліковій системі розсинхронізовані зміни")
                        }
                    }

                    const register_auto_cashout_amount = rro.shift.balance.balance / 100
                    if (available_balance < register_auto_cashout_amount) throw Error("Готівковий залишок в системі менший ніж у фіскальному реєстраторі")
                    // Перевіряємо чи не робилось автоматичне вилучення раніше
                    //const register_auto_cashout_transactions = await Database.Documents.ShiftsTransactions.findOne({ where: { shift_id: shift.id, type: 'CASH_OUT_AUTO', fiscal_id: { [Op.not]: null } }, transaction })
                    if (register_auto_cashout_amount > 0) { // && register_auto_cashout_transactions
                        // Створюємо перше вилучення суми яка є у фіскальному реєстраторі
                        const secure_transaction = await Database.instance.transaction()
                        try {
                            const transaction_id = uuid()
                            const register_auto_cashout_receipt = await rro.cashOut(register_auto_cashout_amount)
                            await Database.Documents.ShiftsTransactions.create({
                                id: transaction_id,
                                subdivision_id,
                                shift_id: shift.id,
                                type: "CASH_OUT_AUTO",
                                amount: register_auto_cashout_amount,
                                fiscal_id: register_auto_cashout_receipt.id,
                                fiscal_date: register_auto_cashout_receipt.fiscal_date
                            }, { transaction: secure_transaction })
                            await Database.Registers.Money.create({
                                basis_id: transaction_id,
                                subdivision_id,
                                workplace_id,
                                operation: "CASH_OUT_AUTO",
                                type: "CASH",
                                amount: -register_auto_cashout_amount
                            }, { transaction: secure_transaction })
                            // Зберігаємо виведення коштів поза межами транзакції запиту
                            secure_transaction.commit()
                        } catch (error) {
                            secure_transaction.rollback()
                            throw error
                        }
                    }
                    if (available_balance > register_auto_cashout_amount) {
                        // Створюємо друге виведення для залишку коштів які не входять у облік фіскального реєстратора
                        const transaction_id = uuid();
                        await Database.Documents.ShiftsTransactions.create({
                            id: transaction_id,
                            subdivision_id,
                            shift_id: shift.id,
                            type: "CASH_OUT_AUTO",
                            amount: available_balance - register_auto_cashout_amount,
                        }, { transaction })
                        await Database.Registers.Money.create({
                            basis_id: transaction_id,
                            subdivision_id,
                            workplace_id,
                            operation: "CASH_OUT_AUTO",
                            type: "CASH",
                            amount: -(available_balance - register_auto_cashout_amount)
                        }, { transaction })
                    }



                    await Database.Documents.ShiftsTransactions.create({ subdivision_id, id: uuid(), shift_id: shift.id, type: "SHIFT_CLOSE", comment: null }, { transaction })
                    await Database.Documents.Shifts.update({ status: "CLOSED" }, { where: { id: shift.id, subdivision_id }, transaction })

                    error = await rro.closeShift()
                    if (error) throw Error("Помилка закриття зміни у фіскальному реєстраторі" + "\n" + error)
                } else {
                    if (rro.shift?.id === shift.id) {
                        await Database.Documents.ShiftsTransactions.update({ createdAt: old_shift.closed_at }, { where: { subdivision_id, shift_id: shift.id, type: "SHIFT_CLOSE" }, transaction })
                    }
                }
            } else {
                const transaction_id = uuid();
                await Database.Documents.ShiftsTransactions.create({
                    id: transaction_id,
                    subdivision_id,
                    shift_id: shift.id,
                    type: "CASH_OUT_AUTO",
                    amount: balance.current_balance,
                }, { transaction })
                await Database.Registers.Money.create({
                    basis_id: transaction_id,
                    subdivision_id,
                    workplace_id,
                    operation: "CASH_OUT_AUTO",
                    type: "CASH",
                    amount: -balance.current_balance
                }, { transaction })
                await Database.Documents.ShiftsTransactions.create({
                    id: uuid(),
                    subdivision_id,
                    shift_id: shift.id,
                    type: "SHIFT_CLOSE"
                }, { transaction })
                await Database.Documents.Shifts.update({ status: "CLOSED" }, { where: { id: shift.id, subdivision_id }, transaction })
            }
            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id: shift.id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
            await Replication.iteration()
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Close shifts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.Shifts.update(Shifts.prepare(req.body, id), { where: { id, subdivision_id }, transaction })
            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update shifts', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.Shifts.destroy({ where: { id, subdivision_id }, transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Delete shifts', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Зміна використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }


    static async prepareBalance(subdivision_id, workplace_id, shift_id = null) {
        const transaction = await Database.instance.transaction()
        try {
            const opened_drafts = await Database.Documents.Draft.findAll({
                include: [{ model: Database.Documents.DraftGoods, as: 'goods', required: true }],
                where: { subdivision_id, workplace_id, status: "NEW" },
                transaction
            })

            const shift = await Database.Documents.Shifts.findOne(Shifts.options(shift_id ? { id: shift_id } : { workplace_id, subdivision_id }, { transaction }))
            if (shift == null) throw Error("Ще не вдкрито жодної зміни")

            var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
            if (!rro && shift.register_id) {
                const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id, is_enabled: true }, transaction })
                rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
            }
            if (rro) await rro.getShift()

            const methods = await Database.PaymentMethods.findAll({ attributes: ['id', 'name', 'type'], transaction })
            const taxes = await Database.Taxes.findAll({ attributes: ['id', 'name', 'letter'], transaction })
            const receipts = await Receipts.getAll({ shift_id: shift.id, subdivision_id }, { transaction })
            const receipts_offline = receipts.filter(item => item.fiscal_sell_offline && !item.fiscal_send)

            const receipts_sales = receipts.filter(it => it.type === "SELL")
            let sales_group_by_method = []
            let sales_group_by_method_type = []
            let sales_group_by_tax = []
            receipts_sales.map(receipt => {
                receipt.payments.map(payment => {
                    const amount = payment.amount_rounded ?? payment.amount
                    const group = sales_group_by_method.find(it => it.id === payment.method_id)
                    if (group) {
                        group.total += amount
                    } else {
                        const new_group = methods.find(it => it.id === payment.method_id)?.get()
                        sales_group_by_method.push({ ...new_group, total: amount })
                    }
                })
                receipt.payments.map(payment => {
                    const amount = payment.amount_rounded ?? payment.amount
                    const type = methods.find(it => it.id === payment.method_id)?.getDataValue('type')
                    const group = sales_group_by_method_type.find(it => it.type === type)
                    if (group) {
                        group.total += amount
                    } else {
                        sales_group_by_method_type.push({ type, total: amount })
                    }
                })
                receipt.goods.map(good => {
                    const group = sales_group_by_tax.find(it => it.id === good.tax_id)
                    if (group) {
                        group.total += Round.money(good.price * good.amount)
                    } else {
                        const new_group = taxes.find(it => it.id === good.tax_id)?.get()
                        sales_group_by_tax.push({ ...new_group, total: good.total })
                    }
                })
            })
            sales_group_by_method.map(item => item.total = Round.money(item.total))
            sales_group_by_method_type.map(item => item.total = Round.money(item.total))
            sales_group_by_tax.map(item => item.total = Round.money(item.total))


            const receipts_returns = receipts.filter(it => it.type === "RETURN")
            let returns_group_by_method = []
            let returns_group_by_method_type = []
            let returns_group_by_tax = []
            receipts_returns.map(receipt => {
                receipt.payments.map(payment => {
                    const amount = payment.amount_rounded ?? payment.amount
                    const group = returns_group_by_method.find(it => it.id === payment.method_id)
                    if (group) {
                        group.total += amount
                    } else {
                        const new_group = methods.find(it => it.id === payment.method_id)?.get()
                        returns_group_by_method.push({ ...new_group, total: amount })
                    }
                })
                receipt.payments.map(payment => {
                    const amount = payment.amount_rounded ?? payment.amount
                    const type = methods.find(it => it.id === payment.method_id)?.getDataValue('type')
                    const group = returns_group_by_method_type.find(it => it.type === type)
                    if (group) {
                        group.total += amount
                    } else {
                        returns_group_by_method_type.push({ type, total: amount })
                    }
                })
                receipt.goods.map(good => {
                    const group = returns_group_by_tax.find(it => it.id === good.tax_id)
                    if (group) {
                        group.total += Round.money(good.price * good.amount)
                    } else {
                        const new_group = taxes.find(it => it.id === good.tax_id)?.get()
                        returns_group_by_tax.push({ ...new_group, total: good.total })
                    }
                })
            })
            returns_group_by_method.map(item => item.total = Round.money(item.total))
            returns_group_by_method_type.map(item => item.total = Round.money(item.total))
            returns_group_by_tax.map(item => item.total = Round.money(item.total))

            let service_in = 0
            let service_out = 0
            shift.transactions.map(transaction => {
                switch (transaction.type) {
                    case "CASH_IN":
                        service_in += transaction.amount; break;
                    case "CASH_OUT":
                        service_out += transaction.amount; break;
                }
            })
            let sales_total = Round.money([0, ...sales_group_by_method_type.map(it => it.total)].reduce((prev, next) => prev + next))
            let sales_total_cash = [0, ...sales_group_by_method_type.filter(it => it.type === "CASH").map(it => it.total)].reduce((prev, next) => prev + next)
            let returns_total = Round.money([0, ...returns_group_by_method_type.map(it => it.total)].reduce((prev, next) => prev + next))
            let returns_total_cash = [0, ...returns_group_by_method_type.filter(it => it.type === "CASH").map(it => it.total)].reduce((prev, next) => prev + next)

            const start_shift_amount = [...shift.transactions.filter(it => it.type === 'CASH_IN_AUTO').map(it => it.amount), 0].reduce((prev, next) => prev + next) || 0
            const close_shift_amount = [...shift.transactions.filter(it => it.type === 'CASH_OUT_AUTO').map(it => it.amount), 0].reduce((prev, next) => prev + next) || 0


            await transaction.commit()
            return {
                initial_balance: Round.money(start_shift_amount),
                current_balance: Round.money(start_shift_amount + sales_total_cash - returns_total_cash + service_in - service_out - close_shift_amount),
                final_balance: Round.money(close_shift_amount),
                offline_receipts: receipts_offline.length,
                is_register_used: !!shift.register_id,
                is_register_online: rro?.isOnline && !rro?.hasOfflineReceipts || false,
                register_balance: {
                    balance: (rro?.shift?.balance?.balance || 0) / 100,
                    card_returns: (rro?.shift?.balance?.card_returns || 0) / 100,
                    card_sales: (rro?.shift?.balance?.card_sales || 0) / 100,
                    cash_returns: (rro?.shift?.balance?.cash_returns || 0) / 100,
                    cash_sales: (rro?.shift?.balance?.cash_sales || 0) / 100,
                    discounts_sum: (rro?.shift?.balance?.discounts_sum || 0) / 100,
                    extra_charge_sum: (rro?.shift?.balance?.extra_charge_sum || 0) / 100,
                    initial: (rro?.shift?.balance?.initial || 0) / 100,
                    service_in: (rro?.shift?.balance?.service_in || 0) / 100,
                    service_out: (rro?.shift?.balance?.service_out || 0) / 100,
                },
                has_opened_drafts: !!opened_drafts.length,
                sales_group_by_method_type,
                sales_group_by_method,
                sales_group_by_tax,
                sales_total,
                sales_count: receipts_sales.length,
                returns_group_by_method_type,
                returns_group_by_method,
                returns_group_by_tax,
                returns_total,
                returns_count: receipts_returns.length,
                service_in: Round.money(service_in),
                service_out: Round.money(service_out - close_shift_amount),
            }
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Balance prepare shifts', subdivision_id: config.subdivision.id })
            return null
        }
    }

    static async balance(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { subdivision_id, workplace_id } = req.auth
            res.status(200).send(JSON.stringify(await Shifts.prepareBalance(subdivision_id, workplace_id)))
        } catch (error) {
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async report(req, res) {
        res.setHeader('Content-Type', 'text/plain')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { id, width = 30, font_size, font_weight } = req.query

            let shift = await Database.Documents.Shifts.findOne(Shifts.options(id ? { id } : { workplace_id, subdivision_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього РМК")
            if (!shift.close_transaction) throw Error("Поточна зміна ще не закрита")

            const name = await SettingsHepler.get('receipt_subdivision_name', { subdivision_id, transaction }) || "Назву організації не вказано у налаштуваннях"
            const address = await SettingsHepler.get('receipt_subdivision_address', { subdivision_id, transaction }) || "Адресу організації не вказано у налаштуваннях"
            const tax_number = await SettingsHepler.get('receipt_subdivision_tax_number', { subdivision_id, transaction })

            const taxes = await Database.Taxes.findAll({ transaction })
            if (!staff_id) throw Error("Для генерування звіту потрібно авторизувати касира")
            const staff = await Database.Staff.findOne({ where: { id: staff_id }, transaction })

            const balance = await Shifts.prepareBalance(subdivision_id, workplace_id, shift.id)

            const builder = new ReceiptGenerator({ width, font_size, font_weight })
            builder.separator(true)
            builder.text(name, ReceiptGenerator.Align.Center)
            builder.text(address, ReceiptGenerator.Align.Center)
            if (tax_number) builder.text(`ПН ${tax_number}`, ReceiptGenerator.Align.Center)

            builder.separator()
            builder.text(`Z-ЗВІТ № ${shift.number}`, ReceiptGenerator.Align.Center)

            const cashier = `${staff.lastname}${staff.firstname ? ` ${staff.firstname.slice(0, 1)}.` : ''}`
            builder.justify({ left: shift?.workplace?.name, right: cashier })

            builder.separator(true)
            builder.justify({ left: 'СУМА ПРОДАЖ', right: balance.sales_total.toFixed(2) })
            builder.justify({ left: 'КІЛЬКІСТЬ ПРОДАЖ', right: balance.sales_count })
            if (balance.sales_group_by_method.length) builder.separator()
            balance.sales_group_by_method.map(method => {
                builder.justify({ left: method.name, right: method.total.toFixed(2) })
            })
            builder.separator(true)
            builder.justify({ left: 'СУМА ПОВЕРНЕНЬ', right: balance.returns_total.toFixed(2) })
            builder.justify({ left: 'КІЛЬКІСТЬ ПОВЕРНЕНЬ', right: balance.returns_count })
            if (balance.returns_group_by_method.length) builder.separator()
            balance.returns_group_by_method.map(method => {
                builder.justify({ left: method.name, right: method.total.toFixed(2) })
            })
            builder.separator(true)
            builder.text('ГОТІВКА')
            builder.justify({ left: 'Службове внесення', right: balance.service_in.toFixed(2) })
            builder.justify({ left: 'Службова видача', right: balance.service_out.toFixed(2) })
            builder.separator()
            builder.text('В КАСІ')
            builder.justify({ left: 'Готівка', right: balance.current_balance.toFixed(2) })
            balance.sales_group_by_method.filter(it => it.type !== "CASH").map(method => {
                builder.justify({ left: method.name, right: method.total.toFixed(2) })
            })
            builder.separator(true)
            builder.text('ПОДАТКОВІ СТАВКИ')
            taxes.map(tax => {
                const by_tax = balance.sales_group_by_tax.find(it => it.id === tax.id)
                builder.justify({ left: tax.name, right: by_tax?.total?.toFixed(2) || '0.00' })
            })
            builder.separator(true)
            builder.justify({ left: new Date().toLocaleDateString(), right: new Date().toLocaleTimeString() })
            builder.separator(true)
            await transaction.commit()
            res.status(200).send(builder.toHTML())
        } catch (error) {
            await transaction.rollback()
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async regsiterReportX(req, res) {
        res.setHeader('Content-Type', 'text/plain')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { id, width = 30, font_size, font_weight } = req.query

            let shift = await Database.Documents.Shifts.findOne(Shifts.options(id ? { id } : { workplace_id, subdivision_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього РМК")
            if (shift.close_transaction) throw Error("Поточна зміна вже закрита")

            var rro = null
            var error = null
            var content = null
            if (shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                // Оновлюємо дані про поточну зміну
                error = await rro.getShift(false, shift.status)
                if (error) throw Error("Не вдалось отримати поточну зміну із фіскального реєстратора" + "\n" + error)
                if (rro.isShiftOpened) {
                    // Перевіряємо чи ми не намагаємось закрити зміну яка уже закрита у фіскальному реєстраторі
                    if (!rro.checkSyncOfShift(shift.id)) {
                        const old_shift = await rro.getShiftById(shift.id)
                        if (old_shift.status === 'CLOSED') {
                            await Database.Documents.ShiftsTransactions.update({ createdAt: old_shift.closed_at }, { where: { subdivision_id, shift_id: shift.id, type: "SHIFT_CLOSE" }, transaction })
                            const result = await Database.Documents.Shifts.findOne(Shifts.options({ id: shift.id }, { transaction }))
                            await transaction.commit()
                            return res.status(200).send(JSON.stringify(result))
                        } else {
                            throw Error("В РРО і обліковій системі розсинхронізовані зміни")
                        }
                    }

                    const report = await rro.createReportX()
                    if (report === null) throw Error("Не вдалось створити X звіт")
                    content = await rro.printReport(report.id, width)
                    if (content === null) throw Error("Не вдалось отримати X звіт для друку")
                } else {
                    throw Error('В поточному фіскальному реєстраторі зміна вже закрита')
                }
            } else {
                throw Error('В поточному РМК не використовується фіскальний реєстратор')
            }
            await transaction.commit()
            res.status(200).send(content)
        } catch (error) {
            await transaction.rollback()
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async regsiterReportZ(req, res) {
        res.setHeader('Content-Type', 'text/plain')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { id, width = 30, font_size, font_weight } = req.query

            let shift = await Database.Documents.Shifts.findOne(Shifts.options(id ? { id } : { workplace_id, subdivision_id }, { transaction }))
            if (!shift) throw Error("Не знайдено жодної зміни для цього РМК")
            if (!shift.close_transaction) throw Error("Поточна зміна ще не закрита")

            var rro = null
            var error = null
            var content = null
            if (shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                const rro_shift = await rro.getLastShift()
                if (!rro_shift) throw Error("Не вдалось отримати остальню зміну із фіскального реєстратора")
                if (!rro_shift?.z_report?.id) throw Error('Поточна зміна ще не має Z звіту')
                content = await rro.printReport(rro_shift.z_report.id, width)
                if (content === null) throw Error("Не вдалось отримати Z звіт для друку")
            } else {
                throw Error('В поточному РМК не використовується фіскальний реєстратор')
            }
            await transaction.commit()
            res.status(200).send(content)
        } catch (error) {
            await transaction.rollback()
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }



    static async regsiterReportPeriodical(req, res) {
        res.setHeader('Content-Type', 'text/plain')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id, staff_id } = req.auth
            const { from_date, to_date, width = 30, font_size, font_weight } = req.query
            const workplace = await Database.Workplaces.findOne({ where: { id: workplace_id }, transaction })
            const register_id = workplace.register_id
            var rro = null
            var content = null
            if (register_id) {
                var rro = await FiscalRegistersFactory.getRegister(register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(register_id, register.get())
                }
                if (!rro) throw Error("Немає доступу до обраного РРО")
                content = await rro.getPeriodicalReportText(from_date, to_date, width)
                if (content === null) throw Error("Не вдалось отримати періодичний звіт для друку")
            } else {
                throw Error('В поточному РМК не використовується фіскальний реєстратор')
            }
            await transaction.commit()
            res.status(200).send(content)
        } catch (error) {
            await transaction.rollback()
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }


}