import { col, fn } from '@sequelize/core'
import Checkbox from 'checkbox-prro'
import { v4 as uuid } from 'uuid'
import Database from '../../database/index.js'
import FiscalRegistersFactory from '../../lib/FiscalRegistersFactory/index.js'
import Rest from '../../lib/rest.js'
import Shifts from './index.js'


export default class Transactions {

    static prepare(data, id) {
        const { subdivision_id, shift_id, article_id, type, amount, comment, fiscal_id } = data
        return ({
            subdivision_id,
            id: id ? id : uuid(),
            shift_id,
            article_id,
            type,
            amount: amount || null,
            comment: comment || null,
            fiscal_id: fiscal_id || null,
        })
    }

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['subdivision_id', 'shift_id'] },
            include: [
                { model: Database.Documents.Shifts, as: 'shift', attributes: ['id', 'number'] },
                { model: Database.TransactionsArticles, as: 'article', attributes: ['id', 'name'] }
            ],
            where,
            ...params,
            order: [
                ...(params.order && Array.isArray(params.order) ? params.order : [['createdAt', 'DESC']]),
                [fn('FIELD', col('type'), `SHIFT_OPEN`, `CASH_IN_AUTO`, `CASH_IN`, `CASH_OUT`, `CASH_OUT_AUTO`, `SHIFT_CLOSE`), 'DESC'],
            ],
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const where = Rest.where(req, is_master ? {} : { subdivision_id })
            const pagination = Rest.pagination(req, res, await Database.Documents.ShiftsTransactions.count(where, { transaction }))
            const result = await Database.Documents.ShiftsTransactions.findAll(Transactions.options(where, { transaction, ...pagination }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async get(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { is_master, subdivision_id } = req.auth
            const { id } = req.params
            const result = await Database.Documents.ShiftsTransactions.findOne(Transactions.options({ id, ...(is_master ? {} : { subdivision_id }) }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Get workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id, workplace_id } = req.auth
            const { amount, type, article_id, only_register } = req.body

            const shift = await Database.Documents.Shifts.findOne(Shifts.options({ workplace_id, subdivision_id }, { transaction }))
            if (!shift || shift.status !== "OPENED") throw Error('Для створення транзакції потрібно відкрити касову зміну')

            if (!amount) throw Error("Не коректна сума транзакції")
            if (type === 'CASH_OUT') {
                const balance = await Shifts.prepareBalance(subdivision_id, workplace_id, shift.id)
                if (amount > balance?.current_balance) throw Error("Вказана сума транзакції більша за поточний баланс")
            }
            const operation_sign = type.includes('CASH_IN')

            var created = null
            if (shift.register_id) {
                var rro = await FiscalRegistersFactory.getRegister(shift.register_id)
                if (!rro) {
                    const register = await Database.FiscalRegisters.findOne({ where: { id: shift.register_id }, transaction })
                    rro = await FiscalRegistersFactory.addRegister(shift.register_id, register.get())
                }
                if (!rro) throw new Database.Error("Немає доступу до обраного РРО")
                const error = await rro.getShift()
                if (error) throw new Database.Error("Не вдалось отримати актуальну інформацію про касову зміну: \n" + error)
                if (!rro.checkSyncOfShift(shift.id)) throw new Database.Error("В реєстраторі і обліковій системі розсинхронізовані зміни")

                let receipt = null
                switch (type) {
                    case 'CASH_IN':
                        // Проводимо транзакцію внесення в РРО на всю суму
                        created = await Database.Documents.ShiftsTransactions.create(Transactions.prepare({ ...req.body, shift_id: shift.id, subdivision_id, amount, }), { transaction })
                        await Database.Registers.Money.create({ basis_id: created.id, subdivision_id, workplace_id, operation: type, operation_article_id: article_id, type: "CASH", amount: operation_sign ? amount : -amount }, { transaction })
                        receipt = await rro.cashIn(amount)
                        await Database.Documents.ShiftsTransactions.update({ fiscal_id: receipt.id }, { where: { id: created.id }, transaction })
                        break
                    case 'CASH_OUT':
                        // Проводимо транзакцію вилучення із РРО на суму яка є в балансі РРО та зберігаємо дані поза межами транзакції
                        const register_balance = rro.shift.balance.balance / 100
                        if (only_register && register_balance < amount) throw Error('В касі РРО недостатньо коштів для вилучення: ' + register_balance.toFixed(2))
                        let available_amount = Math.min(amount, register_balance)
                        if (available_amount > 0) {
                            receipt = await rro.cashOut(available_amount)
                            created = await Database.Documents.ShiftsTransactions.create(Transactions.prepare({ ...req.body, shift_id: shift.id, subdivision_id, amount: available_amount, fiscal_id: receipt.id }), { transaction })
                            await Database.Registers.Money.create({ basis_id: created.id, subdivision_id, workplace_id, operation: type, operation_article_id: article_id, type: "CASH", amount: operation_sign ? available_amount : -available_amount }, {})
                        }
                        // Проводимо транзакцію вилучення на решту балансу поза РРО
                        if (!only_register && amount > available_amount) {
                            available_amount = amount - available_amount
                            created = await Database.Documents.ShiftsTransactions.create(Transactions.prepare({ ...req.body, shift_id: shift.id, subdivision_id, amount: available_amount, }), { transaction })
                            await Database.Registers.Money.create({ basis_id: created.id, subdivision_id, workplace_id, operation: type, operation_article_id: article_id, type: "CASH", amount: operation_sign ? available_amount : -available_amount }, { transaction })
                        }
                        break
                }
            } else {
                if (only_register) throw Error("В поточному РМК не використовується РРО")
                created = await Database.Documents.ShiftsTransactions.create(Transactions.prepare({ ...req.body, shift_id: shift.id, subdivision_id, amount, }), { transaction })
                await Database.Registers.Money.create({ basis_id: created.id, subdivision_id, workplace_id, operation: type, operation_article_id: article_id, type: "CASH", amount: operation_sign ? amount : -amount }, { transaction })
            }

            const result = await Database.Documents.ShiftsTransactions.findOne(Transactions.options({ id: created?.id, subdivision_id }, { transaction, logging: true }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(error.status || 500).send(JSON.stringify({ message: error.message }))
        }
    }


    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.ShiftsTransactions.update(req.body, { where: { id, subdivision_id }, transaction })
            const result = await Database.Documents.ShiftsTransactions.findOne(Transactions.options({ id, subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update workplaces', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }

    static async delete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const { id } = req.params
            await Database.Documents.ShiftsTransactions.destroy({ where: { id, subdivision_id } }, { transaction })
            await transaction.commit()
            res.status(200).send(JSON.stringify({ id }))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Delete workplaces', subdivision_id: req.auth.subdivision_id })
            switch (error.name) {
                case 'SequelizeForeignKeyConstraintError':
                    return res.status(500).send(JSON.stringify({ message: "Неможливо видалити" + "\n" + "Транзакція використовується" }))
                default:
                    return res.status(500).send(JSON.stringify({ message: error.message }))
            }
        }
    }
}