import { v4 as uuid } from 'uuid'
import config from '../../config/index.js'
import Database from '../database/index.js'
import JsonWebToken from '../lib/JsonWebToken.js'
import Rest from '../lib/rest.js'
import { Op } from '@sequelize/core'

export default class Subdivisions {

    static options(where = {}, params = {}) {
        return {
            attributes: { exclude: ['createdAt', 'updatedAt'] },
            where,
            ...params
        }
    }

    static async all(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const where = Rest.where(req)
            const pagination = Rest.pagination(req, res, await Database.Subdivisions.count({ where, transaction }))
            const result = await Database.Subdivisions.findAll(Subdivisions.options(where, { ...pagination, transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'All subdivisions', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async current(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { subdivision_id } = req.auth
            const result = await Database.Subdivisions.findOne(Subdivisions.options({ id: subdivision_id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Current subdivisions', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async autocomplete(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { text } = req.query
            const subdivisions = await Database.Subdivisions.findAll({
                attributes: ['id', 'name'],
                where: text ? Database.instance.literal(`MATCH(name) AGAINST(:text IN BOOLEAN MODE)`) : {},
                replacements: { text: '*' + decodeURIComponent(text).trim() + '*' },
                limit: 100,
                offset: 0,
                ...(text ? {} : { order: [['createdAt', 'DESC']] }),
                transaction
            })
            await transaction.commit()
            res.status(200).send(JSON.stringify(subdivisions))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Autocomplete subdivisions', subdivision_id: req.auth.subdivision_id })
            if (error.toString().includes('syntax error')) {
                res.status(200).send(JSON.stringify([]))
            } else {
                res.status(500).send(JSON.stringify({ message: error.message }))
            }

        }
    }

    static async change(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            const { user_id, workplace_id } = req.auth
            const { id } = req.params
            if (!config.subdivision.is_master) throw Error('Змінювати підрозділ може лише центральна база даних')
            const subdivision = await Database.Subdivisions.findOne(Subdivisions.options({ id }))
            if (!subdivision) throw Error('Не знайдено обраний підрозділ')
            res.status(200).send(JSON.stringify({ token: JsonWebToken.sign({ subdivision_id: subdivision.id, user_id, }) }))
        } catch (error) {
            await Database.onCatchError(error, { name: 'Change subdivisions', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async create(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { name, number } = req.body
            const created = await Database.Subdivisions.create({ id: uuid(), name, number }, { transaction })
            const result = await Database.Subdivisions.findOne(Subdivisions.options({ id: created.id }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Create subdivisions', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }

    static async update(req, res) {
        res.setHeader('Content-Type', 'application/json')
        const transaction = await Database.instance.transaction()
        try {
            const { id } = req.params
            const { name, number } = req.body
            await Database.Subdivisions.update({ name, number }, { where: { id }, transaction })
            const result = await Database.Subdivisions.findOne(Subdivisions.options({ id }, { transaction }))
            await transaction.commit()
            res.status(200).send(JSON.stringify(result))
        } catch (error) {
            await transaction.rollback()
            await Database.onCatchError(error, { name: 'Update subdivisions', subdivision_id: req.auth.subdivision_id })
            res.status(500).send(JSON.stringify({ message: error.message, canceled: error.canceled }))
        }
    }


}