
import fetch from 'node-fetch'
import { v4 as uuid } from 'uuid'
import moment from 'moment'
import config from '../../config/index.js'
import Database from '../database/index.js'
import SettingsHelper from '../database/schemes/dictionaries/system/settings_helper.js'

import { DB_VERSION_KEY, DB_REPLICATION_SENT_AT_KEY, DB_REPLICATION_RECEIVED_AT_KEY } from '../../constants/index.js'

import PrepareReplicationPayload from './prepare.js'
import SaveReplicationResponse from './save.js'

// Функція виконання HTTP запиту до центральної бази із можливістю переривання по таймауту
async function request(url, options = {}, time = null) {
    return new Promise((resolve, reject) => {
        var controller = new AbortController();
        var timeout = !time ? undefined : setTimeout(() => {
            controller.abort()
            reject(Error(`Вичерпано час очікування на опрацювання транзакції ${time / 1000}с`))
        }, time)
        fetch(url, { ...options, signal: controller.signal }).then(response => {
            resolve(response)
            clearTimeout(timeout)
        }).catch(error => {
            clearTimeout(timeout)
            reject(error)
        })
    })
}
// Функція для визначення кількості записів у пакеті даних
function GetPackageSize(data) {
    return data && Object.keys(data).length ? Object.keys(data).map(key => data[key] && Array.isArray(data[key]) ? data[key].length || null : 0).reduce((prev, next) => prev + next) : 0
}
// Функція для визначення розміру даних у пакеті
function GetPackageMemory(data) {
    return typeof data === 'string' ? (data.length * 2 / 1024 / 1024).toFixed(1) : (JSON.stringify(data).length * 2 / 1024 / 1024).toFixed(1)
}
// Фукнція для визначення останньої дати модифікації даних у пакеті
function GetPackageLastDate(data, min_value = null) {
    var last_date = min_value
    Object.keys(data).map(key => (Array.isArray(data[key]) ? data[key] : []).map((row) => {
        if ((!last_date && row?.updatedAt) || (row?.updatedAt && moment(row.updatedAt).isValid() && moment(row.updatedAt).isAfter(moment(last_date)))) {
            last_date = row.updatedAt
        }
    }))
    return moment(last_date).toISOString()
}

export default class Replication {
    static id = null
    static start_at = null
    static timer = null

    static async iteration() {
        try {
            // Якщо реплікація уже запущена, пропускаємо поточну ітерацію
            if (Replication.id) {
                return console.log(`[${new Date().toISOString()}] `, `Виконання транзакції ${Replication.id} від ${Replication.start_at?.format('HH:mm:ss')}`)
            }
            // Якщо база даних не готова до реплікації перериваємо її
            if (!Database.ready) throw Error(`База даних у процесі ініціалізації`)
            // Присвоюємо ідентифікатор та час поточної ітерації
            Replication.id = uuid()
            Replication.start_at = moment()
            console.log(`[${new Date().toISOString()}] `, `Початок реплікації ${Replication.id} о ${Replication.start_at?.format('HH:mm:ss')}`)
            // Зчитуємо із налаштувань часові мітки останньої успішної реплікації та версію бази даних
            var last_replication_sent_at = await SettingsHelper.get(DB_REPLICATION_SENT_AT_KEY)
            var last_replication_received_at = await SettingsHelper.get(DB_REPLICATION_RECEIVED_AT_KEY)
            var database_version = await SettingsHelper.get(DB_VERSION_KEY) || config.app.version
            // Готуємо мета дані які надсилатимуться у запиті
            const metadata = {
                transaction_id: Replication.id,                 // ідентифікатор реплікації
                subdivision_id: config.subdivision.id,          // ідентифікатор підрозділу
                database_version,                               // версія бази даних
                replicate_from: last_replication_received_at,   // часова мітка останнього успішного отримання даних
            }
            // Вибираємо всі дані із бази які редагувались у часовому діапазоні між останньою успішною реплікацією і поточним моментом
            const sent_records = await PrepareReplicationPayload(config.subdivision.id, last_replication_sent_at)
            // У тіло запиту обєднюємо дані та метадані
            const body = JSON.stringify(Object.assign({}, sent_records, metadata))
            console.log(`[${new Date().toISOString()}] `, `Пакет для надсилання (${GetPackageSize(sent_records)} записів) (${GetPackageMemory(body)} mb)`)
            const response = await request(config.replication.url, { method: "POST", headers: { 'Content-Type': 'application/json' }, body })
            const result = await response.text()
            switch (response.status) {
                case 200:
                    const received_records = JSON.parse(result)
                    console.log(`[${new Date().toISOString()}] `, `Пакет для збереження (${GetPackageSize(received_records)} записів) (${GetPackageMemory(result)} mb)`)
                    // Із надісланих записів визначаємо час останнього редагування даних і зберігаємо як крайній момент успішного надсилання
                    await SettingsHelper.set({ last_replication_sent_at: GetPackageLastDate(sent_records, last_replication_sent_at) })
                    // Зберігаємо дані отримані у відповідь
                    await SaveReplicationResponse(received_records);
                    // Із отриманих записів визначаємо час останнього редагування даних і зберігаємо як крайній момент успішного отримання
                    await SettingsHelper.set({ last_replication_received_at: GetPackageLastDate(received_records, last_replication_received_at) })
                    break;
                case 404:
                    throw Error(`HTTP помилка [${response.status}]: Центральна база не очікує запити на реплікацію. Перевірте шлях до центральної бази у конфігурації`)
                default:
                    const error = JSON.parse(result)
                    throw Error(`HTTP помилка [${response.status}]: ${error.message}`)
            }
            console.log(`[${new Date().toISOString()}] `, `Реплікацію ${Replication.id} виконано успішно`)
        } catch (error) {
            console.log(`[${new Date().toISOString()}] `, `Помилка проведення реплікації ${Replication.id}: `, error.message)
        } finally {
            Replication.id = null
            Replication.start_at = null
        }
    }

    static start() {
        // Перевіряємо чи реплікації увімкнені у конфігурації
        if (!config.replication.enabled) return
        // Якщо поточна база даних є центральною то вимикаємо для неї перевірку зовнішніх ключів
        if (config.subdivision.is_master) {
            switch (config.database.dialect) {
                case "mysql":
                    return Database.instance.query(`SET FOREIGN_KEY_CHECKS = 0;`)
                case "postgres":
                    // TODO: disable constrains
                    break;
                case "mariadb":
                    return Database.instance.query(`SET GLOBAL foreign_key_checks=OFF;`)
                case "sqlite":
                    return Database.instance.query(`PRAGMA foreign_keys = OFF;`)
                case "db2":
                    // TODO: disable constrains
                    break;
                case "mssql":
                    return Database.instance.query(`EXEC sp_msforeachtable "ALTER TABLE ? NOCHECK CONSTRAINT all"`)
            }
        } else {
            // Зупиняємо попередній регламент якщо він був запущений
            clearInterval(Replication.timer)
            // Запускаємо новий регламент із періодом виконання із конфігурації
            Replication.timer = setInterval(Replication.iteration, config.replication.interval * 1000)
            // Запускаємо перше виконання реплікації через 10 секунд після старту серверу
            setTimeout(Replication.iteration, 10000)
            // Вмикаємо перевірку зовнішніх ключів
            switch (config.database.dialect) {
                case "mysql":
                    Database.instance.query(`SET FOREIGN_KEY_CHECKS = 1;`)
                    break;
                case "postgres":
                    // TODO: enable constrains
                    break;
                case "mariadb":
                    Database.instance.query(`SET GLOBAL foreign_key_checks=ON;`)
                    break;
                case "sqlite":
                    Database.instance.query(`PRAGMA foreign_keys = ON;`)
                    break;
                case "db2":
                    // TODO: enable constrains
                    break;
                case "mssql":
                    Database.instance.query(`EXEC sp_msforeachtable @command1="print '?'", @command2="ALTER TABLE ? WITH CHECK CHECK CONSTRAINT all";`)
                    break;
            }
        }
    }

    static stop() {
        clearInterval(Replication.timer)
    }

    // Опрацювання запиту на реплікацію центральною базою
    static async request(req, res) {
        res.setHeader('Content-Type', 'application/json')
        try {
            // Вибираємо метадані із запиту
            const { transaction_id, subdivision_id, replicate_from } = req.body
            // Отримуємо версію поточної бази даних
            const database_version = await SettingsHelper.get('database_version') || config.app.version
            // Якщо версії баз даних не співпадають, перериваємо реплікацію
            if (req.body.database_version !== database_version) throw Error(`Версії бази даних не співпадають ${database_version} != ${req.body.database_version}`)
            // Якщо база даних не готова до реплікації перериваємо її
            if (!Database.ready) throw Error(`База даних у процесі ініціалізації`)
            // Отримуємо інформацію про підрозділ від якого отриманий запит на реплікацію
            const subdivision = await Database.Subdivisions.findOne({ where: { id: subdivision_id } });
            const subdivision_name = subdivision ? `підрозділу ${subdivision.name}` : `невідомого підрозділу`
            console.log(`[${new Date().toISOString()}] `, `Отримано запит на реплікацію ${transaction_id} від ${subdivision_name}:`)
            // Вибираємо всі дані із бази які редагувались у часовому діапазоні між часовою міткою отриманою від підрозділу і поточним моментом
            const payload = await PrepareReplicationPayload(subdivision_id, replicate_from)
            const body = JSON.stringify(payload)
            console.log(`[${new Date().toISOString()}] `, `Пакет для надсилання (${GetPackageSize(payload)} записів) (${GetPackageMemory(body)} мб)`)
            console.log(`[${new Date().toISOString()}] `, `Пакет для збереження (${GetPackageSize(req.body)} записів) (${GetPackageMemory(req.body)} мб)`)
            // Зберігаємо отримані дані від підрозділу
            await SaveReplicationResponse(req.body)
            // Надсилаємо відповідь
            res.status(200).send(body)
            console.log(`[${new Date().toISOString()}] `, `Реплікацію ${transaction_id} виконано успішно`)
        } catch (error) {
            console.log(`[${new Date().toISOString()}] `, `Помилка проведення реплікації ${req.body.transaction_id}: `, error.message)
            res.status(500).send(JSON.stringify({ message: error.message }))
        }
    }
}