import { Op, Transaction } from '@sequelize/core'
import config from '../../config/index.js'
import Database from '../database/index.js'
import { DB_VERSION_KEY } from '../database/migration/index.js'

/*  Перелік даних у пакеті:
    1) Ролі користувачів
    2) Податкові ставки
    3) Статті транзакцій
    4) Категорії товарів
    5) Види товарів
    6) Пакування товарів
    7) Характеристики товарів
    8) Виробники товарів
    9) Активні речовини товарів
    10) Товари
    11) Штрихкоди
    12) Види цін
    13) Підрозділи
    14) Контрагенти
    15) Товари контрагентів
    16) Регістр цін
    17) Регістр продаж
    18) Налаштування
    19) Системні помилки
    20) Системні сповіщення
    21) Користувачі
    22) Склади
    23) Робочі місця
    24) Методи оплати
    25) Екваєрингові термінали
    26) Фіскальні реєстратори
    27) Ваги
    28) Серії товарів
    29) Касові зміни
    30) Касові транзакції
    31) Регістр руху коштів
    32) Регістр продаж
    33) Документ прихідного ордеру (шапка, товари)
    34) Документ повернення товарів (шапка, товари)
    35) Документ ціноутворення (шапка, товари)
    36) Документ підбору товарів у чек (шапка, товари)
    37) Документ касового чеку (шапка, товари, платежі)
    38) Документ звіт касової зміни (шапка, товари)
    39) Документ інвентаризації (шапка, товари)
    40) Документ списання (шапка, товари)
*/

export default async function PrepareReplicationPayload(subdivision_id, replicate_from) {
    const transaction = await Database.instance.transaction({ isolationLevel: Transaction.ISOLATION_LEVELS.SERIALIZABLE })
    try {

        const query = { where: replicate_from ? { updatedAt: { [Op.gt]: replicate_from } } : {}, transaction }

        const subdivision_warehouses_ids = (await Database.Warehouses.findAll({ attributes: ['id'], where: { subdivision_id } })).map(it => it.id)

        const shared_payload = {
            roles: await Database.Roles.findAll(query),
            taxes: await Database.Taxes.findAll(query),
            transactions_articles: await Database.TransactionsArticles.findAll(query),

            products_categories: await Database.ProductCategories.findAll(query),
            products_kinds: await Database.ProductKinds.findAll(query),
            products_packings: await Database.ProductPacking.findAll(query),
            products_characteristic: await Database.ProductCharacteristic.findAll(query),
            products_producers: await Database.ProductProducers.findAll(query),
            products_active_substances: await Database.ProductActiveSubstances.findAll(query),
            products: await Database.Products.findAll(query),
            barcodes: await Database.Barcodes.findAll(query),
            price_types: await Database.PriceTypes.findAll(query),

            subdivisions: await Database.Subdivisions.findAll(query),
            counterparties: await Database.Counterparties.findAll(query),

            counterparties_products: await Database.CounterpartiesProducts.findAll(query),

            registers_product_price: await Database.Registers.ProductPrice.findAll(query),
            registers_product_stock: await Database.Registers.ProductStock.findAll({ ...query, where: { ...query.where, warehouse_id: { [Op.in]: subdivision_warehouses_ids } } }),
        }


        query.where['subdivision_id'] = subdivision_id

        const private_settings = [
            'last_replication_sent_at',
            'last_replication_received_at',
            DB_VERSION_KEY,
        ]

        const subdivision_payload = {
            settings: await Database.Settings.findAll({ ...query, where: { ...query.where, key: { [Op.notIn]: private_settings } } }),
            app_errors: await Database.System.Errors.findAll(query),
            app_notifications: await Database.System.Notifications.findAll(query),

            staff: await Database.Staff.findAll({
                ...query,
                where: {
                    ...query.where,
                    subdivision_id: {
                        [Op.in]: [subdivision_id, config.subdivision.is_master ? config.subdivision.id : null].filter(it => it)
                    }
                }
            }),

            warehouses: await Database.Warehouses.findAll(query),

            workplaces: await Database.Workplaces.findAll(query),
            payment_methods: await Database.PaymentMethods.findAll(query),
            acquiring_terminals: await Database.AcquiringTerminals.findAll(query),
            fiscal_registers: await Database.FiscalRegisters.findAll(query),
            scales: await Database.Scales.findAll(query),

            series: await Database.Series.findAll(query),

            shifts: await Database.Documents.Shifts.findAll(query),
            transactions: await Database.Documents.ShiftsTransactions.findAll(query),

            registers_money: await Database.Registers.Money.findAll(query),
            registers_sales: await Database.Registers.Sales.findAll(query),

            documents_arrival: await Database.Documents.Arrival.findAll(query),
            documents_arrival_goods: await Database.Documents.ArrivalGoods.findAll(query),

            documents_return: config.is_master ? [] : await Database.Documents.Return.findAll(query),
            documents_return_goods: config.is_master ? [] : await Database.Documents.ReturnGoods.findAll(query),

            documents_pricing: await Database.Documents.Pricing.findAll(query),
            documents_pricing_goods: await Database.Documents.PricingGoods.findAll(query),

            documents_draft: config.is_master ? [] : await Database.Documents.Draft.findAll(query),
            documents_draft_goods: config.is_master ? [] : await Database.Documents.DraftGoods.findAll(query),

            documents_receipt: await Database.Documents.Receipt.findAll(query),
            documents_receipt_goods: await Database.Documents.ReceiptGoods.findAll(query),
            documents_receipt_payments: await Database.Documents.ReceiptPayment.findAll(query),

            documents_report: config.is_master ? [] : await Database.Documents.Report.findAll(query),
            documents_report_goods: config.is_master ? [] : await Database.Documents.ReportGoods.findAll(query),

            documents_inventory: await Database.Documents.Inventory.findAll(query),
            documents_inventory_goods: await Database.Documents.InventoryGoods.findAll(query),
            
            documents_writeoff: await Database.Documents.Writeoff.findAll(query),
            documents_writeoff_goods: await Database.Documents.WriteoffGoods.findAll(query),
        }

        await transaction.commit()
        return Object.assign({}, shared_payload, subdivision_payload)
    } catch (error) {
        await transaction.rollback()
        throw error
    }
}